#include "BitsetUtility.h"

#include <cmath>
using std::pow;
#include <iostream>
using std::cout;
using std::endl;
#include <bitset>
#include <exception>
using std::exception;
#include <cstring>

void printBitset(unsigned char* bitset, int byteCount)
{
	for (int i = 0; i < byteCount; i++)
	{
		std::bitset<8> byte(bitset[i]);
		cout << byte << " ";
	}
	cout << endl;
}

void setBytesInBitset(unsigned char* bitset, int byteIndex, int byteCount, const void* bytes)
{
	for (int i = byteCount - 1; i >= 0; i--)
		bitset[byteIndex + (byteCount - i - 1)] = (unsigned char)(*((unsigned long long*)bytes) >> (i * 8));
}

unsigned long long getValueFromBitset(const unsigned char* bitset, int byteIndex, int byteCount)
{
	unsigned long long value = 0;
	for (int i = byteCount - 1; i >= 0; i--)
		value = (value << 8) | bitset[byteIndex + (byteCount - i - 1)];
	return value;
}

/*
Encodes the value using variable byte encoding.
Returns the amount of bytes used.

Example:
This:		00000000 00010110 11001100 00111111
			00000000 000   1011011   0011000   0111111
			             0 1011011 0 0011000 1 0111111
Becomes:                    01011011 00011000 10111111
Most siginificant bit indiciates if this is the last byte.
1 means yes, 0 means more bytes ahead.
*/
int variableByteEncode(unsigned char* bitset, unsigned long long byteIndex, unsigned long long value)
{
	int byteCount = getVariableByteEncodedByteCount(value);

	for (int i = 0; i < byteCount; i++)
		bitset[byteIndex++] = (value >> ((byteCount - i - 1) * 7)) & 0x7F;
	bitset[byteIndex - 1] |= 0x80; // This indicates that this is the last byte.

	return byteCount;
}

/*
For positive and negative values.

Example:
This:		00000000 00010110 11001100 00111111
			00000000 000   1011011   0011000    0111111
			             0 1011011 0 0011000 10 0111111
Becomes:             00000001 01101100 01100010 0111111
											  ^
									This byte indicates if the value is positive or negative.
									0 = positive
									1 = negative
*/
int variableByteEncode(unsigned char* bitset, unsigned long long byteIndex, long long value)
{
	bool isNegative = value < 0;
	if (isNegative)
		value = ~value + 1;
	
	int byteCount = getVariableByteEncodedByteCount(value);

	for (int i = 0; i < byteCount - 1; i++)
		bitset[byteIndex++] = (value >> ((byteCount - i - 1) * 7) - 1) & 0x7F;
	bitset[byteIndex] = value & 0x3F;
	bitset[byteIndex] |= 0x80; // This indicates that this is the last byte.
	if (isNegative)
		bitset[byteIndex] |= 0x40; // If this bit is set to one, it indicates that the value is negative.

	return byteCount;
}

unsigned long long variableByteDecode(const unsigned char* bitset, unsigned long long byteIndex, int* byteCount)
{
	unsigned long long value = 0;

	int i;
	for (i = 0; i < 8; i++)
	{
		value |= (bitset[byteIndex + i] & 0x7F);

		if ((bitset[byteIndex + i] & 0x80) == 0x0)
			value <<= 7;
		else
		{
			i++; // So byteCount will be correct.
			break;
		}
	}
	if (byteCount != NULL)
		*byteCount = i;

	return value;
}

long long variableByteDecodeSigned(const unsigned char* bitset, unsigned long long byteIndex, int* byteCount)
{
	long long value = 0;

	int i;
	for (i = 0; i < 8; i++)
	{
		if ((bitset[byteIndex + i] & 0x80) == 0x0)
		{
			value |= (bitset[byteIndex + i] & 0x7F);
			value <<= 7;
		}
		else
		{
			value >>= 1;
			value |= (bitset[byteIndex + i] & 0x3F);

			if ((bitset[byteIndex + i] & 0x40) != 0)
				value *= -1; // This is a negative value

			i++; // So byteCount will be correct.
			break;
		}
	}
	if (byteCount != NULL)
		*byteCount = i;

	return value;
}

/*
For positive values.
*/
int getVariableByteEncodedByteCount(unsigned long long value)
{
	int byteCount = 1;
	if (value > 0x1FFFFFFFFFFFF)
		byteCount = 8;
	else if (value > 0x3FFFFFFFFFF)
		byteCount = 7;
	else if (value > 0x7FFFFFFFF)
		byteCount = 6;
	else if (value > 0xFFFFFFF)
		byteCount = 5;
	else if (value > 0x1FFFFF)
		byteCount = 4;
	else if (value > 0x3FFF)
		byteCount = 3;
	else if (value > 0x7F)
		byteCount = 2;

	return byteCount;
}

/*
For positive and negative values.
*/
int getVariableByteEncodedByteCount(long long value)
{
	if (value < 0)
		value = ~value + 1;

	int byteCount = 1;
	if (value > 0xFFFFFFFFFFFF)
		byteCount = 8;
	else if (value > 0x1FFFFFFFFFF)
		byteCount = 7;
	else if (value > 0x3FFFFFFFF)
		byteCount = 6;
	else if (value > 0x7FFFFFF)
		byteCount = 5;
	else if (value > 0xFFFFF)
		byteCount = 4;
	else if (value > 0x1FFF)
		byteCount = 3;
	else if (value > 0x3F)
		byteCount = 2;

	return byteCount;
}

bool getBitAt(unsigned char* bitset, int bitIndex)
{
	int byteIndex = bitIndex / 8;
	int localIndex = bitIndex - (bitIndex / 8) * 8;
	return (bitset[byteIndex] & (0x80 >> localIndex)) != 0;
}

void setBitAt(unsigned char* bitset, int bitIndex, bool value)
{
	int byteIndex = bitIndex / 8;
	int localIndex = bitIndex - (bitIndex / 8) * 8;

	if (value)
		bitset[byteIndex] |= (0x80 >> localIndex);
	else
		bitset[byteIndex] &= ~(0x80 >> localIndex);
}

bool getBitAt(Bitset& bitset, int bitIndex)
{
	int byteIndex = bitIndex / 8;
	int localIndex = bitIndex - (bitIndex / 8) * 8;
	return (bitset.bytes[byteIndex] & (0x80 >> localIndex)) != 0;
}

void setBitAt(Bitset& bitset, int bitIndex, bool value)
{
	int byteIndex = bitIndex / 8;
	int localIndex = bitIndex - (bitIndex / 8) * 8;

	if (value)
		bitset.bytes[byteIndex] |= (0x80 >> localIndex);
	else
		bitset.bytes[byteIndex] &= ~(0x80 >> localIndex);
}

void bitwiseAnd(Bitset& byteArray1, Bitset& byteArray2, Bitset* result)
{
	int numBytes = (int)ceil(result->numBits / 8.0);

	while (numBytes % 4 != 0)
	{
		numBytes--;
		result->bytes[numBytes] = byteArray1.bytes[numBytes] & byteArray2.bytes[numBytes];
	}

	uint32_t* array1 = (uint32_t*)byteArray1.bytes;
	uint32_t* array2 = (uint32_t*)byteArray2.bytes;
	uint32_t* r = (uint32_t*)result->bytes;
	numBytes /= 4;
	for (int i = numBytes; i-- > 0;)
		r[i] = array1[i] & array2[i];
}

bool areEqual(Bitset& bitset1, Bitset& bitset2)
{
	return memcmp(bitset1.bytes, bitset2.bytes, (int)ceil(bitset1.numBits / 8.0)) == 0;
}

void setBitsetValues(Bitset& bitset, bool value)
{
	for (unsigned int i = 0; i < bitset.numBits; i++)
		setBitAt(bitset, i, value);
}
