﻿#include "CfpArray.h"
#include "BitsetUtility.h"
#include "Utility.h"
#include "../FimLogger/Logger.h"
#include "CfpTreeUtility.h"

#include <iostream>
using std::cout;
using std::endl;
#include <ctime>
using std::clock;
#include <tuple>
#include <cstdlib>
#include <omp.h>

CfpArray::CfpArray(LogData* logData, DatasetPreprocessor* datasetPreprocessor)
	: logData(logData)
{
	dfsStack = new stack<ElementInfo>();
	dfsCountStack = new stack<unsigned int>();
	minSup = datasetPreprocessor->getMinSup();
	remapToInitialId = datasetPreprocessor->getRemapToInitialIdArray();
	transactionCount = datasetPreprocessor->getTransactionCount();
	subarrayCount = datasetPreprocessor->getNumItems() + 1; // + 1 because item = 0 is not used
	subarrayStartIndices = new unsigned long long[subarrayCount + 1];
}

CfpArray::~CfpArray()
{
	delete dfsStack;
	delete dfsCountStack;
	delete[] cfpArray;
	delete[] subarrayStartIndices;
}

/***** PUBLIC FUNCTIONS *****/

void CfpArray::buildCfpArray(CfpTree* cfpTree)
{
	this->cfpTree = cfpTree;
	arraySize = getCfpArraySize();
	cfpArray = new unsigned char[arraySize];
	transformTreeToArray();
	this->cfpTree = NULL;
}

/*
Returns the new byteIndex after the specified elements are read from the array.
Note: if count == NULL the returned byte index is byteIndex + deltaItemNumBytes + deltaPosNumBytes.
*/
unsigned long long CfpArray::readElementFromArray(unsigned long long byteIndex,
												  unsigned int& deltaItem,
												  long long& deltaPos,
												  unsigned int* count)
{
	int byteCount;
	deltaItem = variableByteDecode(cfpArray, byteIndex, &byteCount);
	byteIndex += byteCount;
	deltaPos = variableByteDecodeSigned(cfpArray, byteIndex, &byteCount);
	byteIndex += byteCount;
	if (count != NULL)
	{
		*count = variableByteDecode(cfpArray, byteIndex, &byteCount);
		byteIndex += byteCount;
	}

	return byteIndex;
}

unsigned long long CfpArray::getArraySize()
{
	return arraySize;
}

int CfpArray::getSubarrayCount()
{
	return subarrayCount;
}

unsigned long long* CfpArray::getSubarrayStartIndices()
{
	return subarrayStartIndices;
}

int* CfpArray::getRemapToInitialIdArray()
{
	return remapToInitialId;
}

double CfpArray::getMinimumSupport()
{
	return minSup;
}

int CfpArray::getTransactionCount()
{
	return transactionCount;
}

/***** PRIVATE FUNCTIONS *****/

/*
Returns the new byteIndex after the element is written to the array.
*/
unsigned long long CfpArray::writeElementToArray(unsigned int byteIndex,
												 unsigned int deltaItem,
												 long long deltaPos,
												 unsigned int count)
{
	byteIndex += variableByteEncode(cfpArray, byteIndex, (unsigned long long)deltaItem);
	byteIndex += variableByteEncode(cfpArray, byteIndex, deltaPos);
	byteIndex += variableByteEncode(cfpArray, byteIndex, (unsigned long long)count);
	return byteIndex;
}

/*
Calculates the CfpArray size and initializes the start position of each subarray,
i.e. the first byte position for each item.
*/
unsigned long long CfpArray::getCfpArraySize()
{
	string startMessage = "Calculating CFP-array size.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	// initialize array
	for (int i = 0; i < subarrayCount + 1; i++)
		subarrayStartIndices[i] = 0;

	unsigned long long arraySize = 0;
	unsigned char* root = cfpTree->getRootNode();
	if (root != NULL)
		arraySize = getCfpArraySize(root);

	// convert from local position to global position
	for (int i = subarrayCount; i > 0; i--)
		subarrayStartIndices[i] = subarrayStartIndices[i - 1];
	subarrayStartIndices[0] = 0;
	for (int i = 1; i <= subarrayCount; i++)
		subarrayStartIndices[i] += subarrayStartIndices[i - 1];

	double totalTime = omp_get_wtime() - time;
	string endMessage = "Array size: " + std::to_string(arraySize) +
		" (" + toString((arraySize / (double)(1024 * 1024))) +
		" MB). [" + toString(totalTime) + "s]\n\n";
	writeToLog(endMessage);

	logData->memoryUsageInMB += arraySize / (double)(1024 * 1024);
	logData->preprocessTime += totalTime;

	return arraySize;
}

unsigned long long CfpArray::getCfpArraySize(unsigned char* startNode)
{
	unsigned long long arraySize = 0;
	stack<unsigned int>& countStack = *dfsCountStack;
	bool lastNodeWasNotVisited = true;
	stack<ElementInfo>& nodes = *dfsStack;
	pushAllSiblingsOnStack(startNode, 0, &nodes);
	while (!nodes.empty())
	{
		ElementInfo& currentNode = nodes.top();

		if (currentNode.visited)
		{
			if (lastNodeWasNotVisited)
			{ // bottom of tree has been reached
				countStack.push(currentNode.count);
				lastNodeWasNotVisited = false;
			}
			else
			{
				for (int i = 0; i < currentNode.numChildren; i++)
				{
					currentNode.count += countStack.top();
					countStack.pop();
				}
				countStack.push(currentNode.count); // If this node has a parent.
			}

			if (currentNode.deltaItem == 0)
			{
				nodes.pop();
				continue; // TODO: deltaItem is not supposed to be zero
			}

			unsigned int elementSize = 0;
			elementSize += getVariableByteEncodedByteCount((unsigned long long)currentNode.deltaItem);
			long long parentLocalPos = (long long)subarrayStartIndices[currentNode.item - currentNode.deltaItem];
			long long localPos = (long long)subarrayStartIndices[currentNode.item];
			elementSize += getVariableByteEncodedByteCount(localPos - parentLocalPos);
			elementSize += getVariableByteEncodedByteCount((unsigned long long)currentNode.count);
			arraySize += elementSize;

			// Only setting total space consumption for each subarray here.
			subarrayStartIndices[currentNode.item] += elementSize;

			nodes.pop();
			continue;
		}
		currentNode.visited = true;
		lastNodeWasNotVisited = true;

		performDfsIteration(currentNode, nodes);
	}

	return arraySize;
}

void CfpArray::transformTreeToArray()
{
	string startMessage = "Transforming CfpTree to CfpArray.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	unsigned long long* currentLocalPosition = new unsigned long long[subarrayCount];
	for (int i = 0; i < subarrayCount; i++)
		currentLocalPosition[i] = 0;

	stack<unsigned int>& countStack = *dfsCountStack;
	bool lastNodeWasNotVisited = true;
	stack<ElementInfo>& nodes = *dfsStack;
	pushAllSiblingsOnStack(cfpTree->getRootNode(), 0, &nodes);
	while (!nodes.empty())
	{
		ElementInfo& currentNode = nodes.top();

		if (currentNode.visited)
		{
			if (lastNodeWasNotVisited)
			{ // bottom of tree has been reached
				countStack.push(currentNode.count);
				lastNodeWasNotVisited = false;
			}
			else
			{
				for (int i = 0; i < currentNode.numChildren; i++)
				{
					currentNode.count += countStack.top();
					countStack.pop();
				}
				countStack.push(currentNode.count); // If this node has a parent.
			}

			if (currentNode.deltaItem == 0)
			{
				nodes.pop();
				continue; // TODO: deltaItem is not supposed to be zero
			}

			long long parentLocalPos = (long long)currentLocalPosition[currentNode.item - currentNode.deltaItem];
			long long deltaPos = (long long)currentLocalPosition[currentNode.item] - parentLocalPos;
			unsigned long long byteIndex = currentLocalPosition[currentNode.item] + subarrayStartIndices[currentNode.item];
			unsigned long long newByteIndex = writeElementToArray(byteIndex, currentNode.deltaItem, deltaPos, currentNode.count);

			currentLocalPosition[currentNode.item] += newByteIndex - byteIndex;

			nodes.pop();
			continue;
		}
		currentNode.visited = true;
		lastNodeWasNotVisited = true;

		performDfsIteration(currentNode, nodes);
	}

	delete[] currentLocalPosition;

	double totalTime = omp_get_wtime() - time;
	string endMessage = "Transformation finished. [" + toString(totalTime) + "s]\n\n";
	writeToLog(endMessage);
	logData->preprocessTime += totalTime;
}

/*
Performs a depth first search iteration, i.e. pushes a node and all its siblings on to stack.
If the given node is a chain node, all suffixes are added and the node it points to with all its
siblings.
*/
void CfpArray::performDfsIteration(ElementInfo& currentNode, stack<ElementInfo>& nodes)
{
	UncompressedNode unode;
	if (isPointerEmbeddedLeafNode(currentNode.node))
	{
		// Nothing to do, this node has already been added.
	}
	else if (isChainNode(currentNode.node[0]))
	{
		currentNode.numChildren = 1; // This is a chain node and hence it has one child.
		int numNodes = getNumNodesInChain(currentNode.node[0]);
		unsigned int parentItem = currentNode.item;
		for (int i = 1; i < numNodes; i++) // i = 1, since currentNode is a chain node and has already been added
		{
			unode.deltaItem = readChainNodeItemDelta(currentNode.node, i);
			ElementInfo info = { currentNode.node, unode.deltaItem, unode.deltaItem + parentItem, 0, true, 1 };
			nodes.push(info);
			parentItem += unode.deltaItem;
		}
		ElementInfo& parent = nodes.top();
		parent.numChildren = pushAllSiblingsOnStack(readChainNodePointer(currentNode.node), parentItem, &nodes);
	}
	else
	{
		readCfpNode(currentNode.node, NULL, &unode.pcount, NULL, &unode.suffix, NULL);

		if (unode.suffix != NULL)
			currentNode.numChildren = pushAllSiblingsOnStack(unode.suffix, currentNode.item, &nodes);
	}
}

/*
Pushes the node and all its siblings ordered into the stack, where the object with smallest deltaItem is at top.
The correct item value is set, count is set to pcount and visited is set to false.

Returns amount of siblings (including the input node).
*/
int CfpArray::pushAllSiblingsOnStack(unsigned char* node, unsigned int parentItem, stack<ElementInfo>* siblings)
{
	int numChildren = 0;
	UncompressedNode unode;
	// TODO: during paralellization this can't be static, create multiple instances and store them as global objects instead
	static stack<std::tuple<unsigned char*, bool> > nodes; // static in order to optimize this function
	nodes.push(std::make_tuple(node, false));
	while (!nodes.empty())
	{
		unsigned char* currentNode = std::get<0>(nodes.top());
		decompressCfpNode(currentNode, &unode);

		bool isVisited = std::get<1>(nodes.top());
		if (unode.right != NULL && !isVisited)
		{
			nodes.pop();
			nodes.push(std::make_tuple(currentNode, true));
			nodes.push(std::make_tuple(unode.right, false));
		}
		else
		{
			ElementInfo info = { currentNode, unode.deltaItem, unode.deltaItem + parentItem, unode.pcount, false, 0 };
			siblings->push(info);
			numChildren++;
			nodes.pop();

			if (unode.left != NULL)
				nodes.push(std::make_tuple(unode.left, false));
		}
	}

	return numChildren;
}

void CfpArray::printArray()
{
	//printBitset(cfpArray, arraySize);
	//cout << endl;
	int numPrintsBeforePause = 50;
	int counter = 0;
	unsigned long long byteIndex = 0;
	while (byteIndex < arraySize)
	{
		unsigned int deltaItem;
		long long deltaPos;
		unsigned int count;
		cout << "byteIndex: " << byteIndex;
		byteIndex = readElementFromArray(byteIndex, deltaItem, deltaPos, &count);
		cout << " deltaItem: " << deltaItem << " deltaPos: " << deltaPos << " count: " << count << endl;

		if (++counter >= numPrintsBeforePause)
		{
			//system("pause");
			counter = 0;
		}
	}
}
