﻿#include "CfpMiner.h"
#include "Utility.h"
#include "BitsetUtility.h"
#include "../FimLogger/Logger.h"

#include <ctime>
#include <omp.h>
#include <cmath>

CfpMiner::CfpMiner(CfpArray* cfpArray, int numThreads, LogData* logData)
	: Miner(cfpArray, numThreads, logData)
{
	headerArray = new int*[numThreads];
	mineStack = new ItemsetStack*[numThreads];
	for (int i = 0; i < numThreads; i++)
	{
		headerArray[i] = new int[cfpArray->getSubarrayCount() + 1];
		mineStack[i] = new ItemsetStack;
	}
}

CfpMiner::~CfpMiner()
{
	for (int i = 0; i < numThreads; i++)
	{
		delete[] headerArray[i];
		delete mineStack[i];
	}
	delete[] headerArray;
	delete[] mineStack;
}

/***** PUBLIC FUNCTIONS *****/

void CfpMiner::startMining()
{
	string startMessage = "Mining procedure started.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	double minSup = cfpArray.getMinimumSupport();
	int numTransactions = cfpArray.getTransactionCount();
	int minNumTransactions = (int)ceil(minSup * numTransactions);
	if (minNumTransactions == 0)
		minNumTransactions = 1; // An itemset must occur in at least one transaction

	#pragma omp parallel for schedule(dynamic, 1) num_threads(numThreads)
	for (int item = cfpArray.getSubarrayCount(); item >= 1; --item) // Smallest item is 1 and not 0
	{
		createLocalHeaderFile(item);
		mine(item);
	}

	double totalTime = omp_get_wtime() - time;
	string endMessage = "Mining procedure finished. [" + toString(totalTime) + "s]\n\n";
	writeToLog(endMessage);
	logData->mineTime += totalTime;
}

/***** PRIVATE FUNCTIONS *****/

void CfpMiner::mine(unsigned int startItem)
{
	int threadNr = omp_get_thread_num();

	unsigned int* item = mineStack[threadNr]->push(1, 0.0);
	item[0] = startItem;
	mine();
	mineStack[threadNr]->pop();
}

void CfpMiner::mine()
{
	int threadNr = omp_get_thread_num();

	Itemset itemset;
	unsigned int* itemCount;
	double* support;
	unsigned int* items;
	mineStack[threadNr]->top(&itemCount, &support, &items);
	itemset.itemCount = *itemCount;
	itemset.support = *support;
	itemset.items = items;

	setSupportCountOf(itemset);
	if (itemset.support < minNumTransactions)
		return;

	if (addToFrequentItemsets(threadNr, (int*)itemset.items, itemset.itemCount, itemset.support))
		return; // Itemset threshold reached!

	int lastItem = itemset.items[itemset.itemCount - 1];
	for (int i = lastItem - 1; i > 0; i--)
	{
		if (!headerArray[threadNr][i])
			continue;

		unsigned int* newItemset = mineStack[threadNr]->push(itemset.itemCount + 1, 0.0);
		for (unsigned int j = 0; j < itemset.itemCount; j++)
			newItemset[j] = itemset.items[j];
		newItemset[itemset.itemCount] = i;

		mine();
		mineStack[threadNr]->pop();
	}
}

/*
All items that are contained within the prefix trees given in the subarray are set to true in the headerArray.
*/
void CfpMiner::createLocalHeaderFile(unsigned int initialItem)
{
	int threadNr = omp_get_thread_num();

	for (int i = 0; i < cfpArray.getSubarrayCount() + 1; i++)
		headerArray[threadNr][i] = 0;
	headerArray[threadNr][initialItem] = initialItem;

	unsigned int item = initialItem;
	unsigned long long byteIndex = cfpArray.getSubarrayStartIndices()[item];
	unsigned long long endByteIndex = item >= (unsigned int)cfpArray.getSubarrayCount() - 1 ?
		cfpArray.getArraySize() : cfpArray.getSubarrayStartIndices()[item + 1];

	while (byteIndex < endByteIndex) // for all leaf nodes in prefix tree
	{
		unsigned int deltaItem;
		long long deltaPos;
		unsigned int count;
		unsigned long long index = byteIndex;
		byteIndex = cfpArray.readElementFromArray(byteIndex, deltaItem, deltaPos, &count);

		while (deltaItem != item) // traverse a path in the prefix tree
		{
			index = (index - cfpArray.getSubarrayStartIndices()[item]) - deltaPos;
			item -= deltaItem;
			index += cfpArray.getSubarrayStartIndices()[item]; // parent byte index*/
			cfpArray.readElementFromArray(index, deltaItem, deltaPos, NULL);

			headerArray[threadNr][item] = true;
		}
		item = initialItem; // reset item
	}
}

/*
Assuming that the items in itemset are given in descending order by frequency (first item is most frequent).
*/
void CfpMiner::setSupportCountOf(Itemset& itemset)
{
	itemset.support = 0;

	if (itemset.itemCount < 1)
		return;

	unsigned int item = itemset.items[0];
	unsigned long long byteIndex = cfpArray.getSubarrayStartIndices()[item];
	unsigned long long endByteIndex = item >= (unsigned int)cfpArray.getSubarrayCount() - 1 ?
		cfpArray.getArraySize() : cfpArray.getSubarrayStartIndices()[item + 1];

	while (byteIndex < endByteIndex) // for all leaf nodes in prefix tree
	{
		unsigned int deltaItem;
		long long deltaPos;
		unsigned int count;
		unsigned long long index = byteIndex;
		byteIndex = cfpArray.readElementFromArray(byteIndex, deltaItem, deltaPos, &count);

		if (itemset.itemCount == 1)
		{
			itemset.support += count;
			continue;
		}

		int currentIndex = 1;
		while (deltaItem != item) // traverse a path in the prefix tree
		{
			index = (index - cfpArray.getSubarrayStartIndices()[item]) - deltaPos;
			item -= deltaItem;
			index += cfpArray.getSubarrayStartIndices()[item]; // parent byte index
			cfpArray.readElementFromArray(index, deltaItem, deltaPos, NULL);

			if (itemset.items[currentIndex] == item)
			{
				if (currentIndex == itemset.itemCount - 1)
				{
					itemset.support += count;
					break; // The itemset has been found.
				}

				currentIndex++;
			}
			else if (itemset.items[currentIndex] > item)
			{
				break; // The itemset does not exist in this path.
			}
		}
		item = itemset.items[0]; // reset item
	}
}
