﻿#include "CfpTree.h"
#include "../FimLogger/Logger.h"
#include "Utility.h"

#include <string>
using std::string;
#include <omp.h>
#include <iostream>
using std::cout;
using std::endl;

CfpTree::CfpTree(MultiBuffer **multiBuffers,
				 DatasetPreprocessor* datasetPreprocessor,
				 LogData* logData)
{
	this->datasetPreprocessor = datasetPreprocessor;
	this->multiBuffers = multiBuffers;
	this->minSup = datasetPreprocessor->getMinSup();
	this->numThreads = datasetPreprocessor->getNumThreads();
	this->logData = logData;

	memoryManager = new CfpTreeMemoryManager*[numThreads];
	for (int i = 0; i < numThreads; i++)
		memoryManager[i] = new CfpTreeMemoryManager();
}

CfpTree::~CfpTree()
{
	for (int i = 0; i < numThreads; i++)
		delete memoryManager[i];
	delete[] memoryManager;
}

unsigned char* CfpTree::getRootNode()
{
	return rootNode;
}

void CfpTree::buildTree()
{
	string startMessage = "Creating CFP-tree.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	// An item must be in at least minNumTransactions transactions in order to be added to the tree.
	int minNumTransactions = datasetPreprocessor->getMinNumTransactions();
	#pragma omp parallel num_threads(numThreads)
	{
		int threadNr = omp_get_thread_num();

		unsigned int* items = new unsigned int[datasetPreprocessor->getLongestTransaction()];
		int itemIndex = 0;
		int item = -1;
		for (int i = 0; i < multiBuffers[threadNr]->getNumBytes(); i++)
		{
			char byte = multiBuffers[threadNr]->getByteAndDelete(&i);

			if (isdigit(byte))
			{
				if (item == -1)
					item = 0;
				else
					item *= 10;

				item += byte - '0';
			}
			else
			{
				if (item != -1) // This is the case if the last byte was a space (in case multiple spaces occur after each other).
					items[itemIndex++] = item;
				item = -1;

				if (byte == '\n')
				{
					qsort(items, itemIndex, sizeof(int), compareItem);

					// Remove infrequent items and remap items.
					for (int j = 0; j < itemIndex; j++)
					{
						if (DatasetPreprocessor::supportCount[items[j]] < minNumTransactions)
						{
							itemIndex = j;
							break;
						}
						items[j] = datasetPreprocessor->getInitialToRemapIdArray()[items[j]];
					}

					if (itemIndex != 0)
					{
						insertTransaction(items, itemIndex);
						itemIndex = 0;
					}
				}
			}
		}

		delete[] items;
	}
	for (int i = 0; i < numThreads; i++)
		delete multiBuffers[i];
	delete[] multiBuffers;

	double memoryUsage = 0.0;
	double totalTime = omp_get_wtime() - time;
	for (int i = 0; i < numThreads; i++)
		memoryUsage += memoryManager[i]->getMemoryUsageInBytes() / (double)(1024 * 1024);
	string endMessage = "CFP-tree created (" + toString(memoryUsage) +
		" MB). [" + toString(totalTime) + "s]\n\n";
	writeToLog(endMessage);
	logData->memoryUsageInMB += memoryUsage;
	logData->preprocessTime += totalTime;
}

void CfpTree::insertTransaction(unsigned int *items, unsigned int itemCount)
{
	int threadNr = omp_get_thread_num();

	// The node that pointed to this element (this element might have been either left, right or suffix pointer).
	// This is needed in case the current node changes size, and hence also changes pointer.
	PointedByInfo pointedByInfo;

	unsigned int delta = 0;
	unsigned char* currentNode = rootNode;
	unsigned int i = 0;

	if (rootNode == NULL)
	{
		rootNode = createChainNodes(items, itemCount, 0, 0, memoryManager[threadNr]);
		return;
	}

	for (; i < itemCount; i++)
	{
		while (true) // Do until the correct node has been found.
		{
			UncompressedNode unode;
			unsigned int currentDeltaItem = (unsigned int)items[i] - delta;

			if (isPointerEmbeddedLeafNode(currentNode))
			{
				bool finished = handleEmbeddedLeafNode(&currentNode,
													   unode,
													   pointedByInfo,
													   currentDeltaItem,
													   itemCount,
													   i,
													   threadNr);
				if (finished)
					return;
			}
			else if (isChainNode(currentNode[0]))
			{
				bool finished = handleChainNode(&currentNode,
												unode,
												pointedByInfo,
												currentDeltaItem,
												delta,
												items,
												itemCount,
												i,
												threadNr);
				if (finished)
					return;

				// All nodes in chain are in items.
				pointedByInfo.wasPointedByLeft = false;
				pointedByInfo.wasPointedBySuffix = true; // The pointer in a chain node is always a suffix pointer.
				pointedByInfo.wasPointedByRight = false;
				pointedByInfo.pointedBy = currentNode;
				currentNode = unode.suffix;
				if (currentNode == NULL)
				{
					currentNode = createChainNodes(items, itemCount, i, delta, memoryManager[threadNr]);
					updatePointedByPointer(currentNode, pointedByInfo, &rootNode);
				}

				continue; // currentNode has changed, continue in order to check if it is chain node or embedded leaf node
				// and also in order to update unode
			}
			else
				readCfpNode(currentNode, &unode.deltaItem, &unode.pcount, &unode.left, &unode.suffix, &unode.right);

			if (currentDeltaItem == unode.deltaItem)
			{ // Update node
				if (i == itemCount - 1)
				{ // Last element to be inserted.
					unsigned int newPcount = unode.pcount + 1;

					bool doesPcountNeedToExpand = false;
					if (unode.pcount == 0 && newPcount > 0)
						doesPcountNeedToExpand = true;
					else if (unode.pcount <= 0xFF && newPcount > 0xFF)
						doesPcountNeedToExpand = true;
					else if (unode.pcount <= 0xFFFF && newPcount > 0xFFFF)
						doesPcountNeedToExpand = true;
					else if (unode.pcount <= 0xFFFFFF && newPcount > 0xFFFFFF)
						doesPcountNeedToExpand = true;

					if (doesPcountNeedToExpand)
					{
						unode.pcount = newPcount;
						expandNode(&unode, currentNode, pointedByInfo, memoryManager[threadNr], &rootNode);
					}
					else
						editNode(currentNode, NULL, &newPcount, NULL, NULL, NULL);
				}
				else if (unode.suffix == NULL)
				{
					// The tree doesn't contain the next item in transaction.
					// Hence, create rest of the nodes here and exit function.

					// i + 1, hence delta + unode.deltaItem
					unode.suffix = createChainNodes(items, itemCount, i + 1, delta + unode.deltaItem, memoryManager[threadNr]);

					// edit current node
					if (unode.left == NULL && unode.right == NULL)
					{ // No expansion of the current node needed.
						editNode(currentNode, NULL, NULL, NULL, &unode.suffix, NULL);
					}
					else
					{ // Expansion of the current node is needed.
						expandNode(&unode, currentNode, pointedByInfo, memoryManager[threadNr], &rootNode);
					}

					return; // The rest of the items have been added in this branch.
				}
				else
				{ // Not the last item, and it already exists in the tree.
					delta += unode.deltaItem;
					pointedByInfo.wasPointedByLeft = false;
					pointedByInfo.wasPointedBySuffix = true;
					pointedByInfo.wasPointedByRight = false;
					pointedByInfo.pointedBy = currentNode;
					currentNode = unode.suffix;
				}

				break;
			}
			else if (currentDeltaItem < unode.deltaItem)
			{ // Insert into left node.
				if (unode.left == NULL)
				{
					unode.left = createChainNodes(items, itemCount, i, delta, memoryManager[threadNr]);

					// edit current node
					if (unode.right == NULL && unode.suffix == NULL)
					{ // No expansion of the current node needed.
						editNode(currentNode, NULL, NULL, &unode.left, NULL, NULL);
					}
					else
					{ // Expansion of the current node is needed.
						expandNode(&unode, currentNode, pointedByInfo, memoryManager[threadNr], &rootNode);
					}

					return; // The rest of the items have been added in this branch.
				}
				else
				{
					pointedByInfo.wasPointedByLeft = true;
					pointedByInfo.wasPointedBySuffix = false;
					pointedByInfo.wasPointedByRight = false;
					pointedByInfo.pointedBy = currentNode;
					currentNode = unode.left;
				}
			}
			else
			{ // Inssert into right node.
				if (unode.right == NULL)
				{
					unode.right = createChainNodes(items, itemCount, i, delta, memoryManager[threadNr]);

					// edit current node
					if (unode.left == NULL && unode.suffix == NULL)
					{ // No expansion of the current node needed.
						editNode(currentNode, NULL, NULL, NULL, NULL, &unode.right);
					}
					else
					{ // Expansion of the current node is needed.
						expandNode(&unode, currentNode, pointedByInfo, memoryManager[threadNr], &rootNode);
					}

					return; // The rest of the items have been added in this branch.
				}
				else
				{
					pointedByInfo.wasPointedByLeft = false;
					pointedByInfo.wasPointedBySuffix = false;
					pointedByInfo.wasPointedByRight = true;
					pointedByInfo.pointedBy = currentNode;
					currentNode = unode.right;
				}
			}
		}
	}
}

/*
Returns true if finished.
*/
bool CfpTree::handleEmbeddedLeafNode(unsigned char** currentNode,
									 UncompressedNode& unode,
									 PointedByInfo& pointedByInfo,
									 unsigned int& currentDeltaItem,
									 unsigned int itemCount,
									 unsigned int currentItemIndex,
									 int threadNr)
{
	unode.left = NULL;
	unode.suffix = NULL;
	unode.right = NULL;
	readEmbeddedLeafNode(*currentNode, &unode.deltaItem, &unode.pcount);

	if (currentDeltaItem == unode.deltaItem)
	{
		if (currentItemIndex == itemCount - 1)
		{
			// If last item check if pcount can be updated, if not create new node.

			unode.pcount++;
			if (unode.pcount <= 0xFFFFFF)
			{
				int pointerByteIndex = 1;
				if (isChainNode(pointedByInfo.pointedBy[0]))
					pointerByteIndex += getNumNodesInChain(pointedByInfo.pointedBy[0]);
				else
				{
					pointerByteIndex +=
							getDeltaItemSize(pointedByInfo.pointedBy[0]) +
							getPcountSize(pointedByInfo.pointedBy[0]
					);

					if (pointedByInfo.wasPointedBySuffix)
					{
						if (!isLeftPointerNull(pointedByInfo.pointedBy[0]))
							pointerByteIndex += POINTER_SIZE_IN_BYTES;
					}
					else if (pointedByInfo.wasPointedByRight)
					{
						if (!isLeftPointerNull(pointedByInfo.pointedBy[0]))
							pointerByteIndex += POINTER_SIZE_IN_BYTES;
						if (!isSuffixPointerNull(pointedByInfo.pointedBy[0]))
							pointerByteIndex += POINTER_SIZE_IN_BYTES;
					}
				}

				createEmbeddedLeafNode(&unode, pointedByInfo.pointedBy, pointerByteIndex);
			}
			else
			{ // The embedded leaf node is too big, hence create new node.
				unsigned char* newCurrentNode = createCfpNode(&unode, memoryManager[threadNr]);
				updatePointedByPointer(
							newCurrentNode,
							pointedByInfo,
							&rootNode
				);
			}

			return true; // Last element updated, hence return.
		}
		else
		{
			// Create new node and replace the embedded leaf node with it.
			*currentNode = createCfpNode(&unode, memoryManager[threadNr]);
			updatePointedByPointer(
				*currentNode, pointedByInfo, &rootNode
			);
		}
	}
	else
	{
		// Item must be added to left or right, hence create new node.
		*currentNode = createCfpNode(&unode, memoryManager[threadNr]);
		updatePointedByPointer(*currentNode, pointedByInfo, &rootNode);
	}

	return false;
}

/*
Returns true if finished adding all items.
*/
bool CfpTree::handleChainNode(unsigned char** currentNode,
							  UncompressedNode& unode,
							  PointedByInfo& pointedByInfo,
							  unsigned int& currentDeltaItem,
							  unsigned int& delta,
							  unsigned int* items,
							  unsigned int itemCount,
							  unsigned int& currentItemIndex,
							  int threadNr)
{
	unode.pcount = 0;
	unode.left = NULL;
	unode.suffix = readChainNodePointer(*currentNode);
	unode.right = NULL;

	int nodeCountInChain = getNumNodesInChain((*currentNode)[0]);
	for (int j = 0; j < nodeCountInChain; j++)
	{
		currentDeltaItem = (unsigned int)items[currentItemIndex] - delta;
		unode.deltaItem = readChainNodeItemDelta(*currentNode, j);
		if (currentDeltaItem == unode.deltaItem)
		{
			if (currentItemIndex == itemCount - 1)
			{ // Last element
				unsigned char* newNode = splitChainNode(*currentNode,
														j, NULL, 0,
														0, false,
														memoryManager[threadNr]);
				updatePointedByPointer(newNode, pointedByInfo, &rootNode);
				unsigned int newPcount = 1;

				// pcount must be set to 1
				// (all intern nodes in a chain node have pcount = 0, hence increment by 1)
				if (j == 0) // newNode == splitNode
				{
					if (getPcountSize(newNode[0]) != 0)
						editNode(newNode, NULL, &newPcount, NULL, NULL, NULL);
					else
					{
						readCfpNode(newNode, &unode.deltaItem, NULL, &unode.left, &unode.suffix, &unode.right);
						unode.pcount = newPcount;
						expandNode(&unode, newNode, pointedByInfo, memoryManager[threadNr], &rootNode);
					}
				}
				else // newNode.suffix == splitNode
				{
					if (isChainNode(newNode[0]))
						unode.suffix = readChainNodePointer(newNode);
					else
						readCfpNode(newNode, NULL, NULL, NULL, &unode.suffix, NULL);

					if (getPcountSize(unode.suffix[0]) != 0)
						editNode(unode.suffix, NULL, &newPcount, NULL, NULL, NULL);
					else
					{
						unsigned char* oldSuffix = unode.suffix;
						readCfpNode(oldSuffix, &unode.deltaItem, NULL, &unode.left, &unode.suffix, &unode.right);
						unode.pcount = newPcount;
						PointedByInfo pbi;
						pbi.pointedBy = newNode;
						pbi.wasPointedByLeft = false;
						pbi.wasPointedBySuffix = true;
						pbi.wasPointedByRight = false;
						expandNode(&unode, oldSuffix, pbi, memoryManager[threadNr], &rootNode);
					}
				}

				return true; // This was the last element in items, hence return.
			}
			else
			{
				currentItemIndex++;
				delta += unode.deltaItem;
			}
		}
		else
		{
			bool isLeftPointer = false;
			if (currentDeltaItem < unode.deltaItem)
				isLeftPointer = true;

			unsigned char* newNode = splitChainNode(*currentNode,
													j,
													&items[currentItemIndex],
													itemCount - currentItemIndex,
													delta,
													isLeftPointer,
													memoryManager[threadNr]);
			updatePointedByPointer(newNode, pointedByInfo, &rootNode);
			return true;
		}
	}

	return false;
}
