#include "CfpTreeMemoryManager.h"
#include "../FimLogger/Logger.h"

#include <exception>
using std::exception;
#include <cstdlib>

CfpTreeMemoryManager::CfpTreeMemoryManager()
	: bufferCount(0), bufferIndex(0), totalMemoryUsage(MEMORY_BUFFER_SIZE)
{
	initializeStacks();
	memoryBuffers = new unsigned char*[BUFFER_COUNT];
	allocateMemoryBuffer(0);
	for (int i = 1; i < BUFFER_COUNT; i++)
		memoryBuffers[i] = NULL;
}

CfpTreeMemoryManager::~CfpTreeMemoryManager()
{
	for (int i = 0; i < BUFFER_COUNT; i++)
		if (memoryBuffers[i] != NULL)
			delete[] memoryBuffers[i];
	delete[] memoryBuffers;

	int stackSize = LARGEST_CFP_NODE_IN_BYTES - SMALLEST_CFP_NODE_IN_BYTES + 1;
	for (int i = 0; i < stackSize; i++)
		delete freeMemoryStack[i];
	delete[] freeMemoryStack;
}

/***** PUBLIC FUNCTIONS *****/

unsigned char* CfpTreeMemoryManager::allocateMemory(int byteCount)
{
	if (byteCount < SMALLEST_CFP_NODE_IN_BYTES || byteCount > LARGEST_CFP_NODE_IN_BYTES)
	{
		writeToLog("Can't allocate memory, byteCount is either too large or too small. CfpTreeMemoryManager::allocateMemory");
		throw exception();
	}

	unsigned char* pointer = allocateMemoryFromStack(byteCount);
	if (pointer == NULL)
	{
		if (bufferIndex + byteCount >= MEMORY_BUFFER_SIZE)
		{
			bufferCount++;
			allocateMemoryBuffer(bufferCount);
			bufferIndex = 0;
			totalMemoryUsage += MEMORY_BUFFER_SIZE;
		}

		pointer = &memoryBuffers[bufferCount][bufferIndex];
		bufferIndex += byteCount;
	}

	return pointer;
}

void CfpTreeMemoryManager::freeMemory(unsigned char* pointer, int byteCount)
{
	addFreeMemoryToStack(pointer, byteCount);
}

unsigned long long CfpTreeMemoryManager::getMemoryUsageInBytes()
{
	return totalMemoryUsage - MEMORY_BUFFER_SIZE + bufferIndex;
}

/***** PRIVATE FUNCTIONS *****/

void CfpTreeMemoryManager::initializeStacks()
{
	int size = LARGEST_CFP_NODE_IN_BYTES - SMALLEST_CFP_NODE_IN_BYTES + 1;
	freeMemoryStack = new stack<unsigned char*>*[size];
	for (int i = 0; i < size; i++)
		freeMemoryStack[i] = new stack<unsigned char*>;
}

unsigned char* CfpTreeMemoryManager::allocateMemoryFromStack(int byteCount)
{
	// TODO: If stack with more bytes than byteCount is available
	// allocate that memory instead, if a stack with size byteCount is empty.

	stack<unsigned char*>* lifo = freeMemoryStack[byteCount - SMALLEST_CFP_NODE_IN_BYTES];
	if ((*lifo).empty())
		return NULL;
	else
	{
		unsigned char* pointer = (*lifo).top();
		(*lifo).pop();
		return pointer;
	}
}

void CfpTreeMemoryManager::addFreeMemoryToStack(unsigned char* pointer, int byteCount)
{
	stack<unsigned char*> *lifo = freeMemoryStack[byteCount - SMALLEST_CFP_NODE_IN_BYTES];
	(*lifo).push(pointer);
}

void CfpTreeMemoryManager::allocateMemoryBuffer(const unsigned int bufferNr)
{
	memoryBuffers[bufferNr] = new unsigned char[MEMORY_BUFFER_SIZE];

	// CFP-Growth highest memory location cannot be or
	// exceed 0xFF00000000, with pointer size set to 5.
	bool memoryError = false;
	while ((unsigned long long)memoryBuffers[bufferCount] >= 0xFF00000000)
	{
		memoryError = true;
		writeToLog("Memory error, allocating new memory.\n");
		delete[] memoryBuffers[bufferCount];
		memoryBuffers[bufferCount] = new unsigned char[MEMORY_BUFFER_SIZE];
	}
	if (memoryError)
		writeToLog("New memory allocation successful.\n");
}
