#include "CfpTreeUtility.h"
#include "BitsetUtility.h"
#include "../FimLogger/Logger.h"
#include "Utility.h"

#include <cstdlib>
#include <cmath>
#include <exception>
using std::exception;

void readCfpNode(const unsigned char* node,
				 unsigned int* deltaItem,
				 unsigned int* pcount,
				 unsigned char** left,
				 unsigned char** suffix,
				 unsigned char** right)
{
	int index = 1;
	unsigned int deltaItemSize = getDeltaItemSize(node[0]);
	if (deltaItem != NULL) // check deltaItem
		*deltaItem = (unsigned int)getValueFromBitset(node, index, deltaItemSize);
	index += deltaItemSize;

	unsigned int pcountSize = getPcountSize(node[0]);
	if (pcount != NULL) // check pcount
		*pcount = (unsigned int)getValueFromBitset(node, index, pcountSize);
	index += pcountSize;

	if (!isLeftPointerNull(node[0])) // check left pointer
	{
		if (left != NULL)
			*left = (unsigned char*)getValueFromBitset(node, index, POINTER_SIZE_IN_BYTES);
		index += POINTER_SIZE_IN_BYTES;
	}
	else if (left != NULL)
		*left = NULL;

	if (!isSuffixPointerNull(node[0])) // check suffix pointer
	{
		if (suffix != NULL)
			*suffix = (unsigned char*)getValueFromBitset(node, index, POINTER_SIZE_IN_BYTES);
		index += POINTER_SIZE_IN_BYTES;
	}
	else if (suffix != NULL)
		*suffix = NULL;

	if (right != NULL && !isRightPointerNull(node[0])) // check right pointer
		*right = (unsigned char*)getValueFromBitset(node, index, POINTER_SIZE_IN_BYTES);
	else if (right != NULL)
		*right = NULL;
}

/*
Outputs the amount of bytes contained in a cfp node, this includes chain nodes.
*/
unsigned int getByteCountOfCfpNode(unsigned char bitmask)
{
	unsigned int byteCount = 1;

	if (isChainNode(bitmask))
	{
		byteCount += getNumNodesInChain(bitmask);
		byteCount += (unsigned char)POINTER_SIZE_IN_BYTES;
	}
	else
	{
		byteCount += getDeltaItemSize(bitmask);
		byteCount += getPcountSize(bitmask);

		if (areAllPointersNull(bitmask))
			byteCount += (unsigned char)POINTER_SIZE_IN_BYTES; // A cfp node must have space to at least one pointer.
		else
		{
			byteCount += isLeftPointerNull(bitmask)   ? (unsigned char)0 : (unsigned char)POINTER_SIZE_IN_BYTES;
			byteCount += isSuffixPointerNull(bitmask) ? (unsigned char)0 : (unsigned char)POINTER_SIZE_IN_BYTES;
			byteCount += isRightPointerNull(bitmask)  ? (unsigned char)0 : (unsigned char)POINTER_SIZE_IN_BYTES;
		}
	}

	return byteCount;
}

unsigned int getDeltaItemSize(unsigned char bitmask)
{
	return (unsigned char)4 - (unsigned char)(bitmask >> 6);
}

unsigned int getPcountSize(unsigned char bitmask)
{
	return (unsigned char)4 - ((unsigned char)(bitmask << 2) >> 5);
}

bool isLeftPointerNull(unsigned char bitmask)
{
	return !(bitmask & (unsigned char)0x4);
}

bool isSuffixPointerNull(unsigned char bitmask)
{
	return !(bitmask & (unsigned char)0x2);
}

bool isRightPointerNull(unsigned char bitmask)
{
	return !(bitmask & (unsigned char)0x1);
}

bool areAllPointersNull(unsigned char bitmask)
{
	return isLeftPointerNull(bitmask) && isSuffixPointerNull(bitmask) && isRightPointerNull(bitmask);
}

/*
Returns true if the given node is a chain node.
*/
bool isChainNode(unsigned char bitmask)
{
	return (bitmask & 0x38) == 0x38; // (xxyy yxxx & 0011 1000) == 0011 1000
}

/*
Outputs the pointer contained in the chain node.
*/
unsigned char* readChainNodePointer(const unsigned char* chainNode)
{
	int firstPointerByteIndex = 1 + getNumNodesInChain(chainNode[0]);
	return (unsigned char*)getValueFromBitset(chainNode, firstPointerByteIndex, POINTER_SIZE_IN_BYTES);
}

/*
Reads the delta item at the specified index in the chain node.
The index indicates which item is specified, i.e. index = 0 is the first item in the node.
*/
unsigned int readChainNodeItemDelta(const unsigned char* chainNode, int index)
{
	return chainNode[index + 1];
}

unsigned int getNumNodesInChain(unsigned char bitmask)
{
	unsigned char p1 = (bitmask >> 3) & 0x18; // (xxyy yyyy >> 3) & 0001 1000 = zzzx xyyy & 0001 1000 = 000x x000
	unsigned char p2 = bitmask & 0x7; // yyyy yxxx & 0000 0111 = 0000 0xxx
	return p1 | p2; // 000x x000 | 0000 0xxx = 000x xxxx
}

void readEmbeddedLeafNode(const unsigned char* leafNode, unsigned int* itemDelta, unsigned int* pcount)
{
	unsigned long long leafValue = (unsigned long long)leafNode;
	*itemDelta = (leafValue >> ((POINTER_SIZE_IN_BYTES - 2) * 8)) & 0xFF;
	*pcount = (unsigned int)leafValue & 0xFFFFFF;
}

bool isPointerEmbeddedLeafNode(unsigned char* pointer)
{
	// if first byte is 255 then this is a leaf node
	return ((unsigned long long)pointer >> ((POINTER_SIZE_IN_BYTES - 1) * 8)) == 0xFF;
}

/*
Decompresses the given node if it is not a chain node.
If the given node is a chain node, the first element of the chain node is returned.
The sufffix pointer of the chain node will however be NULL, even though it is not the last element.
*/
void decompressCfpNode(unsigned char* node, UncompressedNode* unode)
{
	unode->deltaItem = 0;
	unode->pcount = 0;
	unode->left = NULL;
	unode->suffix = NULL;
	unode->right = NULL;

	if (isPointerEmbeddedLeafNode(node))
		readEmbeddedLeafNode(node, &unode->deltaItem, &unode->pcount);
	else if (isChainNode(node[0]))
		unode->deltaItem = readChainNodeItemDelta(node, 0);
	else
		readCfpNode(node, &unode->deltaItem, &unode->pcount, &unode->left, &unode->suffix, &unode->right);
}

/*
compression mask (1B) | deltaItem (max 4B)  | pcount (max 4B) | left (max 5B) | suffix (max 5B) | right (max 5B)

compression mask:
First two bits: Number of leading zeros in deltaItem. (max 3 leading zeros)
Next three bits: Number of leading zeros in pcount. (max 4 leading zeros)
Last three bits: Indicate if left/suffix/right is not null pointer. E.g. left:null|suffix:not null|right:null, bit mask: 010

*/
unsigned char* createCfpNode(UncompressedNode* unode, CfpTreeMemoryManager* memoryManager)
{
	int deltaItemSize = 0;
	int pcountSize = 0;
	int pointerSize = 0;

	// create compression mask
	unsigned char compressionMask = 0;
	if (unode->deltaItem <= 0xFF)
	{
		compressionMask = 0xC0; // 1100 0000
		deltaItemSize = 1;
	}
	else if (unode->deltaItem <= 0xFFFF)
	{
		compressionMask = 0x80; // 1000 0000
		deltaItemSize = 2;
	}
	else if (unode->deltaItem <= 0xFFFFFF)
	{
		compressionMask = 0x40; // 0100 0000
		deltaItemSize = 3;
	}
	else
		deltaItemSize = 4;

	if (unode->pcount == 0)
		compressionMask |= 0x20; // xx10 0000
	else if (unode->pcount <= 0xFF)
	{
		compressionMask |= 0x18; // xx01 1000
		pcountSize = 1;
	}
	else if (unode->pcount <= 0xFFFF)
	{
		compressionMask |= 0x10; // xx01 0000
		pcountSize = 2;
	}
	else if (unode->pcount <= 0xFFFFFF)
	{
		compressionMask |= 0x08; // xx00 1000
		pcountSize = 3;
	}
	else
		pcountSize = 4;

	if (unode->left != NULL)
	{
		compressionMask |= 0x4; // xxyy y100
		pointerSize += POINTER_SIZE_IN_BYTES;
	}
	if (unode->suffix != NULL)
	{
		compressionMask |= 0x2; // xxyy yz10
		pointerSize += POINTER_SIZE_IN_BYTES;
	}
	if (unode->right != NULL)
	{
		compressionMask |= 0x1; // xxyy yzz1
		pointerSize += POINTER_SIZE_IN_BYTES;
	}

	if (pointerSize == 0)
		pointerSize += POINTER_SIZE_IN_BYTES; // There must be at least one pointer. (NULL pointer in this case)

	// create cfpNode bitset
	unsigned char* cfpNode = memoryManager->allocateMemory(1 + deltaItemSize + pcountSize + pointerSize);
	cfpNode[0] = compressionMask;

	int index = 1;
	setBytesInBitset(cfpNode, index, deltaItemSize, &unode->deltaItem);
	index += deltaItemSize;
	setBytesInBitset(cfpNode, index, pcountSize, &unode->pcount);
	index += pcountSize;
	if (unode->left != NULL)
	{
		setBytesInBitset(cfpNode, index, POINTER_SIZE_IN_BYTES, &unode->left);
		index += POINTER_SIZE_IN_BYTES;
	}
	if (unode->suffix != NULL)
	{
		setBytesInBitset(cfpNode, index, POINTER_SIZE_IN_BYTES, &unode->suffix);
		index += POINTER_SIZE_IN_BYTES;
	}
	if (unode->right != NULL)
	{
		setBytesInBitset(cfpNode, index, POINTER_SIZE_IN_BYTES, &unode->right);
		index += POINTER_SIZE_IN_BYTES;
	}

	if (unode->left == NULL && unode->suffix == NULL && unode->right == NULL)
	{
		void* nullPointer = NULL;
		setBytesInBitset(cfpNode, index, POINTER_SIZE_IN_BYTES, &nullPointer);
	}

	return cfpNode;
}

/*
Edits the specified node with the given values.
If an input parameter is NULL, the change to that field is ignored.
IMPORTANT: This algorithm assumes that the node is not going to expand!
*/
void editNode(unsigned char* node,
			  unsigned int* deltaItem,
			  unsigned int* pcount,
			  unsigned char** left,
			  unsigned char** suffix,
			  unsigned char** right)
{
	int index = 1;

	unsigned int deltaItemSize = getDeltaItemSize(node[0]);
	if (deltaItem != NULL)
		setBytesInBitset(node, index, deltaItemSize, deltaItem);
	index += deltaItemSize;

	unsigned int pcountSize = getPcountSize(node[0]);
	if (pcount != NULL)
		setBytesInBitset(node, index, pcountSize, pcount);
	index += pcountSize;

	if (left != NULL)
	{
		setBytesInBitset(node, index, POINTER_SIZE_IN_BYTES, left);
		if (areAllPointersNull(node[0]))
			node[0] |= 0x4; // Left node has been set, hence change the bitmask.
	}
	if (!isLeftPointerNull(node[0]))
		index += POINTER_SIZE_IN_BYTES;

	if (suffix != NULL)
	{
		setBytesInBitset(node, index, POINTER_SIZE_IN_BYTES, suffix);
		if (areAllPointersNull(node[0]))
			node[0] |= 0x2; // Suffix node has been set, hence change the bitmask.
	}
	if (!isSuffixPointerNull(node[0]))
		index += POINTER_SIZE_IN_BYTES;

	if (right != NULL)
	{
		setBytesInBitset(node, index, POINTER_SIZE_IN_BYTES, right);
		if (areAllPointersNull(node[0]))
			node[0] |= 0x1; // Right node has been set, hence change the bitmask.
	}
}

unsigned char* createChainNodes(unsigned int* items,
								unsigned int itemCount,
								unsigned int itemIndex,
								unsigned int delta,
								CfpTreeMemoryManager* memoryManager)
{
	int numItemsRemaining = itemCount - itemIndex;
	UncompressedNode unode;
	unode.pcount = 0;
	unode.left = NULL;
	unode.suffix = NULL;
	unode.right = NULL;

	if (numItemsRemaining == 1)
	{
		unode.deltaItem = items[itemCount - 1] - delta;
		unode.pcount = 1;
		return createCfpNode(&unode, memoryManager); // Only one item, hence return.
	}
	else if (numItemsRemaining == 0)
		return NULL;

	unsigned char** lastSuffixes = new unsigned char*[numItemsRemaining];
	unsigned int* deltaItems = new unsigned int[numItemsRemaining];
	int numChainNodes = 0;
	int numSuffixes = 0;

	// Create nodes
	for (int i = 0; i < numItemsRemaining - 1; i++)
	{
		deltaItems[numChainNodes] = items[itemIndex + i] - delta;

		// deltaItem must be <= 255
		if (deltaItems[numChainNodes] <= 0xFF)
			numChainNodes++;
		else
		{
			addNodesToLastSuffixes(numChainNodes, lastSuffixes, &numSuffixes, deltaItems, unode, memoryManager);

			// Create seperate node for the item with deltaItem > 255
			unode.deltaItem = deltaItems[numChainNodes];
			lastSuffixes[numSuffixes++] = createCfpNode(&unode, memoryManager);
			numChainNodes = 0;
		}

		delta = items[itemIndex + i];
	}

	// Insert the last chain node into last suffixes.
	addNodesToLastSuffixes(numChainNodes, lastSuffixes, &numSuffixes, deltaItems, unode, memoryManager);

	// Create leaf node
	unode.deltaItem = items[itemCount - 1] - items[itemCount - 2];
	unode.pcount = 1;
	if (unode.deltaItem <= 0xFF)
	{
		unsigned int parentFirstPointerByte;
		if (isChainNode(lastSuffixes[numSuffixes - 1][0]))
			parentFirstPointerByte = 1 + getNumNodesInChain(lastSuffixes[numSuffixes - 1][0]);
		else
		{
			parentFirstPointerByte = 1 + getDeltaItemSize(lastSuffixes[numSuffixes - 1][0]);
			// Suffix will not be a NULL pointer anymore, but contain an embedded leaf node.
			// Hence update bitmask so it will say that it contains a suffix node.
			lastSuffixes[numSuffixes - 1][0] |= 0x02;
		}
		createEmbeddedLeafNode(&unode, lastSuffixes[numSuffixes - 1], parentFirstPointerByte);
	}
	else
		lastSuffixes[numSuffixes++] = createCfpNode(&unode, memoryManager);

	// Set suffixes of each node
	for (int i = 0; i < numSuffixes - 1; i++)
	{
		if (isChainNode(lastSuffixes[i][0]))
			editPointerInChainNode(lastSuffixes[i], lastSuffixes[i + 1]);
		else
			editNode(lastSuffixes[i], NULL, NULL, NULL, &lastSuffixes[i + 1], NULL);
	}

	delete[] deltaItems;

	unsigned char* firstNodeInChain = lastSuffixes[0];
	delete[] lastSuffixes;

	return firstNodeInChain;
}

/*
A helper function for createChainNodes.
*/
void addNodesToLastSuffixes(const int numChainNodes,
							unsigned char** lastSuffixes,
							int *numSuffixes,
							unsigned int* deltaItems,
							UncompressedNode &unode,
							CfpTreeMemoryManager* memoryManager)
{
	if (numChainNodes > 1)
	{
		lastSuffixes[(*numSuffixes)++] = createChainNodeWithNullPointer(deltaItems, numChainNodes, memoryManager);

		if (numChainNodes > 18)
		{
			// if this is the case the function createChainNodeWithNullPointer has created multipe chain nodes
			// (last node might be a regular cfp node)

			int numCfpNodes = (int)ceil(numChainNodes / 18.0);
			for (int j = 1; j < numCfpNodes; j++)
			{
				unsigned char* nextNode;
				if (isChainNode(lastSuffixes[*numSuffixes - 1][0]))
					nextNode = readChainNodePointer(lastSuffixes[*numSuffixes - 1]);
				else
					readCfpNode(lastSuffixes[*numSuffixes - 1], NULL, NULL, NULL, &nextNode, NULL);
				lastSuffixes[(*numSuffixes)++] = nextNode;
			}
		}
	}
	else if (numChainNodes == 1)
	{
		unode.deltaItem = deltaItems[numChainNodes - 1];
		lastSuffixes[(*numSuffixes)++] = createCfpNode(&unode, memoryManager);
	}
}

/*
Creates one chain node with a NULL pointer at the end.

Structure:
bitmask | deltaItem1 | deltaItem2 | ... | deltaItemN | NULL pointer
All deltaItems must be smaller than 256.

Bitmask structure:
xx11 1xxx
pcount part of bitmask (111) indicates that this is a chain node.
The concatenation of the x values, i.e. xxxxx indicate how many nodes are
contained in the chain. Max value is 11111, i.e. 31.
But, because the largest node size is 24, max value is actually 24 - 1 - 5 = 18.
A chain node must contain at least two nodes.

If itemCount > 18 this function creates multiple chain nodes with max 18 elements each.
If last item contains 1 element, the last node will be a normal cfp node.
*/
unsigned char* createChainNodeWithNullPointer(unsigned int* deltaItems,
											  unsigned int itemCount,
											  CfpTreeMemoryManager* memoryManager)
{
	if (itemCount < 2)
	{
		writeToLog("itemCount must be greater than 1");
		throw exception();
	}

	unsigned char* lastChainNode = NULL;
	int numCfpNodes = (int)ceil(itemCount / 18.0);
	for (int i = numCfpNodes - 1; i >= 0; i--)
	{
		int currentItemCount = (i == numCfpNodes - 1) ? itemCount - 18 * i : 18;
		unsigned char* currentChainNode = NULL;

		if (currentItemCount >= 2)
		{
			currentChainNode = memoryManager->allocateMemory(1 + currentItemCount + POINTER_SIZE_IN_BYTES);

			// bitmask
			if (currentItemCount <= 7)
				currentChainNode[0] = currentItemCount; // 0000 0xxx, where xxx is itemCount (2^3 - 1 = 7).
			else
				currentChainNode[0] = (currentItemCount << 3) | currentItemCount; // (000a bcde << 3) | 000a bcde = abcd e000 | 000a bcde = abxx xcde
			currentChainNode[0] |= 0x38; // |= 0011 1000, to indicate that this is a chain node.

			// deltaItems
			int start = i * 18;
			int end = start + currentItemCount;
			for (int j = start; j < end; j++)
				setBytesInBitset(&currentChainNode[j - start + 1], 0, 1, &deltaItems[j]);

			// set pointer
			editPointerInChainNode(currentChainNode, lastChainNode);

			lastChainNode = currentChainNode;
		}
		else // currentItemCount == 1
		{
			// This is always the last element if itemCount >= 2, which must be the case.
			UncompressedNode unode;
			unode.deltaItem = deltaItems[itemCount - 1];
			unode.pcount = 0;
			unode.left = NULL;
			unode.suffix = NULL;
			unode.right = NULL;
			lastChainNode = createCfpNode(&unode, memoryManager);
		}
	}

	return lastChainNode;
}

void editPointerInChainNode(unsigned char* chainNode, unsigned char* pointer)
{
	int firstPointerByteIndex = 1 + getNumNodesInChain(chainNode[0]);
	setBytesInBitset(chainNode, firstPointerByteIndex, POINTER_SIZE_IN_BYTES, &pointer);
}

/*
Split chain node and insert the rest of the elements.

chainSplitIndex: first value has index 0. This value tells which index is the split node,
i.e. last element in chain equal to item in items.

This function deletes the old chain node.
*/
unsigned char* splitChainNode(unsigned char* chainNode,
							  unsigned int chainSplitIndex,
							  unsigned int* items,
							  unsigned int itemCount,
							  unsigned int delta,
							  bool isLeftPointer,
							  CfpTreeMemoryManager* memoryManager)
{
	int chainNodeItemCount = getNumNodesInChain(chainNode[0]);
	unsigned int* chainNodeItems = new unsigned int[chainNodeItemCount];
	for (int i = 0; i < chainNodeItemCount; i++)
		chainNodeItems[i] = readChainNodeItemDelta(chainNode, i);

	/*
		newChainNode1
			  |
		  splitNode
			/   \
		   /     \
		  /       \
		 /         \
	newChainNode2	leftOrRightPointer
	*/
	unsigned char* newChainNode1 = NULL;
	unsigned char* newChainNode2 = NULL;
	unsigned char* splitNode = NULL;
	unsigned char* leftOrRightPointer = createChainNodes(items, itemCount, 0, delta, memoryManager);

	int elementCountChainNode1 = chainSplitIndex;
	int elementCountChainNode2 = chainNodeItemCount - chainSplitIndex - 1;
	UncompressedNode unode;
	unode.pcount = 0; // pcount is always zero  in a chain node, (except for in embedded leaf node or suffix node)

	// newChainNode2
	if (elementCountChainNode2 > 1)
	{
		if (elementCountChainNode2 > 18)
		{
			// If this is the case, createChainNodeWithNullPointer will create multiple nodes
			writeToLog("elementCountChainNode2 is greater than 18");
			throw exception();
		}
		newChainNode2 = createChainNodeWithNullPointer(&chainNodeItems[chainSplitIndex + 1], elementCountChainNode2, memoryManager);
		editPointerInChainNode(newChainNode2, readChainNodePointer(chainNode));
	}
	else if (elementCountChainNode2 == 1)
	{
		unode.deltaItem = chainNodeItems[chainSplitIndex + 1];
		unode.left = NULL;
		unode.suffix = readChainNodePointer(chainNode);
		unode.right = NULL;
		newChainNode2 = createCfpNode(&unode, memoryManager);
	}
	else
		newChainNode2 = readChainNodePointer(chainNode);

	// splitNode
	unode.deltaItem = chainNodeItems[chainSplitIndex];
	unode.left = NULL;
	unode.suffix = newChainNode2;
	unode.right = NULL;
	if (isLeftPointer)
		unode.left = leftOrRightPointer;
	else
		unode.right = leftOrRightPointer;
	splitNode = createCfpNode(&unode, memoryManager);


	// newChainNode1
	if (elementCountChainNode1 > 1)
	{
		if (elementCountChainNode1 > 18)
		{
			// If this is the case, createChainNodeWithNullPointer will create multiple nodes
			writeToLog("elementCountChainNode1 is greater than 18");
			throw exception();
		}
		newChainNode1 = createChainNodeWithNullPointer(chainNodeItems, elementCountChainNode1, memoryManager);
		editPointerInChainNode(newChainNode1, splitNode);
	}
	else if (elementCountChainNode1 == 1)
	{
		unode.deltaItem = chainNodeItems[chainSplitIndex - 1];
		unode.left = NULL;
		unode.suffix = splitNode;
		unode.right = NULL;
		newChainNode1 = createCfpNode(&unode, memoryManager);
	}
	else
		newChainNode1 = splitNode;

	delete[] chainNodeItems;
	memoryManager->freeMemory(chainNode, getByteCountOfCfpNode(chainNode[0]));

	return newChainNode1;
}

/*
Expands the node, i.e. creates a new node and deletes the old one.
The node that pointed to this element gets its respective pointer updated.
The old node is deleted.
*/
void expandNode(UncompressedNode* unode,
				unsigned char* oldNode,
				PointedByInfo pointedByInfo,
				CfpTreeMemoryManager* memoryManager,
				unsigned char** rootNode)
{
	unsigned char* newCurrentNode = createCfpNode(unode, memoryManager);
	updatePointedByPointer(newCurrentNode,
						   pointedByInfo,
						   rootNode);
	memoryManager->freeMemory(oldNode, getByteCountOfCfpNode(oldNode[0]));
}

/*
Updates the pointer of pointedBy with the specified pointer, i.e.: newNode.
*/
void updatePointedByPointer(unsigned char* newNode,
							PointedByInfo pointedByInfo,
							unsigned char** rootNode)
{
	if (pointedByInfo.wasPointedByLeft)
		editNode(pointedByInfo.pointedBy, NULL, NULL, &newNode, NULL, NULL);
	else if (pointedByInfo.wasPointedBySuffix)
	{
		if (isChainNode(pointedByInfo.pointedBy[0])) // check if it was pointed by a chain node (this is always from suffix)
			editPointerInChainNode(pointedByInfo.pointedBy, newNode);
		else
			editNode(pointedByInfo.pointedBy, NULL, NULL, NULL, &newNode, NULL);
	}
	else if (pointedByInfo.wasPointedByRight)
		editNode(pointedByInfo.pointedBy, NULL, NULL, NULL, NULL, &newNode);
	else // was pointed by root node
		*rootNode = newNode;
}

/*
deltaItem must be < 256 = 2^8
pcount must be < 16777216 = 2^24

Structure:
First byte is set to 0xFF = 255, in order to indicate that this is a leaf node.
Second byte is the deltaItem.
Last bytes are the pcount values.
*/
void createEmbeddedLeafNode(UncompressedNode* unode, unsigned char* node, int pointerByteIndex)
{
	node[pointerByteIndex] = 0xFF; // To indicate that this node is a leaf node.
	node[pointerByteIndex + 1] = (unsigned char)unode->deltaItem;
	setBytesInBitset(node, pointerByteIndex + 2, POINTER_SIZE_IN_BYTES - 2, &unode->pcount);
}
