#include "FPGrowthMiner.h"

#include <omp.h>
#include <algorithm>

FPGrowthMiner::FPGrowthMiner(CfpArray* cfpArray, int numThreads, LogData* logData)
	: Miner(cfpArray, numThreads, logData)
{
	memoryManagers = new StackMemoryManager*[numThreads];
	for (int i = 0; i < numThreads; i++)
	{
		memoryManagers[i] = new StackMemoryManager();
	}
}

FPGrowthMiner::~FPGrowthMiner()
{
	for (int i = 0; i < numThreads; i++)
	{
		delete memoryManagers[i];
	}
	delete[] memoryManagers;
}

/***** PUBLIC FUNCTIONS *****/

void FPGrowthMiner::startMining()
{
	string startMessage = "Mining procedure started.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	startMineFPTree();

	double totalTime = omp_get_wtime() - time;
	string endMessage = "Mining procedure finished. [" + toString(totalTime) + "s]\n\n";
	writeToLog(endMessage);
	logData->mineTime += totalTime;
}

/***** PRIVATE FUNCTIONS *****/

void FPGrowthMiner::startMineFPTree()
{
	#pragma omp parallel for schedule(dynamic, 1) num_threads(numThreads)
	for (int item = cfpArray.getSubarrayCount(); item >= 1; --item) // Smallest item is 1 and not 0
	{
		int threadNr = omp_get_thread_num();

		int itemSupportCount = DatasetPreprocessor::supportCount[cfpArray.getRemapToInitialIdArray()[item]];
		if (itemSupportCount < minNumTransactions || itemSupportCount == 0) // itemSupportCount == 0 then continue, must be in at least one transaction
			continue; // current item is not frequent

		if (addToFrequentItemsets(threadNr, &item, 1, itemSupportCount))
			continue; // Itemset threshold reached!

		unsigned int previousBufferNr = memoryManagers[threadNr]->getCurrentBufferNr();
		long long previousByteIndex = memoryManagers[threadNr]->getCurrentByteIndexInBuffer();
		FPTree conditionalFPTree(memoryManagers[threadNr]);
		constructConditionalFpTree(item, &conditionalFPTree);
		mineFPTree(&conditionalFPTree, &item, 1);
		memoryManagers[threadNr]->setToPreviousState(previousBufferNr, previousByteIndex);
	}
}

void FPGrowthMiner::mineFPTree(FPTree* tree, int* suffix, int suffixSize)
{
	int threadNr = omp_get_thread_num();
	if (tree->containsSinglePathOnly(suffix[0]))
	{
		CustomVector<int> path(memoryManagers[threadNr]);
		FPNode* node = tree->getRootNode()->firstChild;
		while (node != NULL)
		{
			if (node->item == suffix[0])
				break;
			path.push_back(node->item);
			node = node->firstChild;
		}

		CalculateSupportCountContext context;
		context.hashHeaderTable = tree->getHashHeaderTable();
		context.currentItem = suffix[0];
		addRestOfItemsToFrequentItemsets(
			&path, suffix, suffixSize,
			&FPGrowthMiner::calculateSupportForFpTree,
			&context
		);
	}
	else
	{
		HashTable<HashHeaderTableElement*>* hashHeaderTable = tree->getHashHeaderTable();
		HashTableIterator<HashHeaderTableElement*> it;
		hashHeaderTable->begin(&it);
		while (it.value != NULL)
		{
			unsigned int previousBufferNr = memoryManagers[threadNr]->getCurrentBufferNr();
			long long previousByteIndex = memoryManagers[threadNr]->getCurrentByteIndexInBuffer();

			if (it.value->id == suffix[0]) // Element already added.
			{
				hashHeaderTable->next(&it);
				continue;
			}

			int item = it.value->id;
			FPNode* bottomNode = hashHeaderTable->find(suffix[0])->value->node;
			int supportCount = calculateSupportForFPTreeNonSinglePath(bottomNode, item);
			if (supportCount < minNumTransactions || supportCount == 0)
			{
				hashHeaderTable->next(&it);
				continue;
			}

			int currentItemsetSize = 1 + suffixSize;
			int* currentItemset = (int*)memoryManagers[threadNr]->grow(sizeof(int) * currentItemsetSize);
			joinTwoArrays(&item, 1, suffix, suffixSize, currentItemset);
			if (addToFrequentItemsets(threadNr, currentItemset, currentItemsetSize, supportCount))
				return; // Itemset threshold reached!

			FPTree conditionalFPTree(memoryManagers[threadNr]);
			constructConditionalFpTree(hashHeaderTable->find(item)->value->node, &conditionalFPTree);
			mineFPTree(&conditionalFPTree, currentItemset, currentItemsetSize);

			hashHeaderTable->next(&it);

			memoryManagers[threadNr]->setToPreviousState(previousBufferNr, previousByteIndex);
		}
	}
}

/*
Return the size of the conditional pattern base.
*/
int FPGrowthMiner::constructConditionalFpTree(unsigned int initialItem, FPTree* fpTree)
{
	int threadNr = omp_get_thread_num();
	CustomVector<CustomVector<int>* > conditionalPatternBase(memoryManagers[threadNr]);
	CustomVector<int> supportCounts(memoryManagers[threadNr]);
	HashTable<int> itemsInPatternBase(memoryManagers[threadNr], DatasetPreprocessor::getLongestTransaction());
	itemsInPatternBase.insert(initialItem, initialItem);
	int sizeOfConditionalPatternBase = 0;

	unsigned int item = initialItem;
	unsigned long long byteIndex = cfpArray.getSubarrayStartIndices()[item];
	unsigned long long endByteIndex = item >= (unsigned int)cfpArray.getSubarrayCount() - 1 ?
		cfpArray.getArraySize() : cfpArray.getSubarrayStartIndices()[item + 1];

	while (byteIndex < endByteIndex) // for all leaf nodes in prefix tree
	{
		unsigned int deltaItem;
		long long deltaPos;
		unsigned int count;
		unsigned long long index = byteIndex;
		byteIndex = cfpArray.readElementFromArray(byteIndex, deltaItem, deltaPos, &count);

		CustomVector<int>* cv = (CustomVector<int>*)memoryManagers[threadNr]->grow(sizeof(CustomVector<int>));
		conditionalPatternBase.push_back(cv);
		cv->constructor(memoryManagers[threadNr]);
		conditionalPatternBase.back()->push_back(item);
		supportCounts.push_back(count);

		while (deltaItem != item) // traverse a path in the prefix tree
		{
			index = (index - cfpArray.getSubarrayStartIndices()[item]) - deltaPos;
			item -= deltaItem;
			index += cfpArray.getSubarrayStartIndices()[item]; // parent byte index*/
			cfpArray.readElementFromArray(index, deltaItem, deltaPos, NULL);

			conditionalPatternBase.back()->push_back(item);
			if (itemsInPatternBase.find(item) == NULL)
				itemsInPatternBase.insert(item, item);
		}
		item = initialItem; // reset item

		sizeOfConditionalPatternBase++;
	}

	// Reverse the arrays so they are sorted in frequency decreesing order.
	for (size_t i = 0; i < conditionalPatternBase.size(); i++)
	{
		int* array = &(conditionalPatternBase[i]->at(0));
		int size = conditionalPatternBase[i]->size() / 2;
		for (int j = 0; j < size; j++)
		{
			int temp = array[j];
			array[j] = array[conditionalPatternBase[i]->size() - j - 1];
			array[conditionalPatternBase[i]->size() - j - 1] = temp;
		}
	}

	// Create header table
	int numItems = (int)itemsInPatternBase.size();
	if (numItems > 0)
	{
		HashTable<HashHeaderTableElement*>* hashHeaderTable =
				(HashTable<HashHeaderTableElement*>*)memoryManagers[threadNr]->grow(sizeof(HashTable<HashHeaderTableElement*>));
		hashHeaderTable->constructor(memoryManagers[threadNr], numItems);
		HashTableIterator<int> it;
		itemsInPatternBase.begin(&it);
		while (it.value != NULL)
		{
			HashHeaderTableElement* element = (HashHeaderTableElement*)memoryManagers[threadNr]->grow(sizeof(HashHeaderTableElement));
			element->itemCount = 0;
			element->node = NULL;
			hashHeaderTable->insert(it.value->value, element);
			itemsInPatternBase.next(&it);
		}
		fpTree->setHashHeaderTable(hashHeaderTable);
	}
	else
		fpTree->setHashHeaderTable(NULL);

	// Create conditional FP-tree
	for (size_t i = 0; i < conditionalPatternBase.size(); i++)
	{
		if (conditionalPatternBase[i]->size() > 0)
		{
			fpTree->insertTransaction(
				&(conditionalPatternBase[i]->at(0)),
				(int)conditionalPatternBase[i]->size(),
				supportCounts[i]
			);
		}

		// No need created in custom memory manager
		//delete conditionalPatternBase[i];
	}

	return sizeOfConditionalPatternBase;
}

void FPGrowthMiner::constructConditionalFpTree(FPNode* firstNode, FPTree* fpTree)
{
	int threadNr = omp_get_thread_num();

	// Get the conditional patter base for the given item.
	CustomVector<CustomVector<int>* > conditionalPatternBase(memoryManagers[threadNr]);
	CustomVector<int> supportCounts(memoryManagers[threadNr]);
	HashTable<int> itemsInPatternBase(memoryManagers[threadNr], DatasetPreprocessor::getLongestTransaction());
	FPNode* node = firstNode;
	int i = 0;
	while (node != NULL)
	{
		CustomVector<int>* cv = (CustomVector<int>*)memoryManagers[threadNr]->grow(sizeof(CustomVector<int>));
		conditionalPatternBase.push_back(cv);
		cv->constructor(memoryManagers[threadNr]);

		conditionalPatternBase[i]->push_back(node->item);
		supportCounts.push_back(node->itemCount);
		if (itemsInPatternBase.find(node->item) == NULL)
			itemsInPatternBase.insert(node->item, node->item);

		FPNode* parent = node->parent;
		while (parent != NULL)
		{
			if (parent->item < 0)
				break; // This is root node

			conditionalPatternBase[i]->push_back(parent->item);
			if (itemsInPatternBase.find(parent->item) == NULL)
				itemsInPatternBase.insert(parent->item, parent->item);
			parent = parent->parent;
		}

		i++;
		node = node->nextEqualItem;
	}

	// Reverse the arrays so they are sorted in frequency decreesing order.
	for (size_t i = 0; i < conditionalPatternBase.size(); i++)
	{
		int* array = &(conditionalPatternBase[i]->at(0));
		int size = conditionalPatternBase[i]->size() / 2;
		for (int j = 0; j < size; j++)
		{
			int temp = array[j];
			array[j] = array[conditionalPatternBase[i]->size() - j - 1];
			array[conditionalPatternBase[i]->size() - j - 1] = temp;
		}
	}

	// Create header table
	int numItems = (int)itemsInPatternBase.size();
	if (numItems > 0)
	{
		HashTable<HashHeaderTableElement*>* hashHeaderTable =
				(HashTable<HashHeaderTableElement*>*)memoryManagers[threadNr]->grow(sizeof(HashTable<HashHeaderTableElement*>));
		hashHeaderTable->constructor(memoryManagers[threadNr], numItems);
		HashTableIterator<int> it;
		itemsInPatternBase.begin(&it);
		while (it.value != NULL)
		{
			HashHeaderTableElement* element = (HashHeaderTableElement*)memoryManagers[threadNr]->grow(sizeof(HashHeaderTableElement));
			element->itemCount = 0;
			element->node = NULL;
			hashHeaderTable->insert(it.value->value, element);
			itemsInPatternBase.next(&it);
		}
		fpTree->setHashHeaderTable(hashHeaderTable);
	}
	else
		fpTree->setHashHeaderTable(NULL);

	// Create conditional FP-tree
	for (size_t i = 0; i < conditionalPatternBase.size(); i++)
	{
		if (conditionalPatternBase[i]->size() > 0)
		{
			fpTree->insertTransaction(
				&(conditionalPatternBase[i]->at(0)),
				(int)conditionalPatternBase[i]->size(),
				supportCounts[i]
			);
		}

		// No need created in custom memory manager
		//delete conditionalPatternBase[i];
	}
}

void FPGrowthMiner::addRestOfItemsToFrequentItemsets(CustomVector<int> *items,
													int* currentItemset,
													int currentItemsetSize,
													int (FPGrowthMiner::*calcSupport)(CalculateSupportCountContext*),
													CalculateSupportCountContext* context)
{
	int threadNr = omp_get_thread_num();

	// Power set using counter
	unsigned int power = 1 << (unsigned int)items->size();
	for (unsigned int i = 1; i < power; i++)
	{
		unsigned int previousBufferNr = memoryManagers[threadNr]->getCurrentBufferNr();
		long long previousByteIndex = memoryManagers[threadNr]->getCurrentByteIndexInBuffer();

		unsigned int bitCounter = i;
		unsigned int setSize = __builtin_popcount(bitCounter);
		unsigned int outputSize = setSize + currentItemsetSize;
		int* output = (int*)memoryManagers[threadNr]->grow(sizeof(int) * outputSize);

		int outputIndex = 0;
		for (int j = 0; j < items->size() && j < sizeof(unsigned int) * 8; j++)
		{
			if (bitCounter & 1)
				output[outputIndex++] = items->at(j);
			bitCounter >>= 1;
		}
		for (int j = 0; j < currentItemsetSize; j++)
			output[j + setSize] = currentItemset[j];

		// calculate support
		context->bitCounter = bitCounter;
		context->itemCount = setSize;
		int support = (this->*calcSupport)(context);

		// add to frequent itemset
		if (addToFrequentItemsets(omp_get_thread_num(), output, outputSize, support))
			return; // Itemset threshold reached!

		memoryManagers[threadNr]->setToPreviousState(previousBufferNr, previousByteIndex);
	}
}

void FPGrowthMiner::joinTwoArrays(int* array1, int sizeArray1, int* array2, int sizeArray2, int* destination)
{
	std::copy(array1, array1 + sizeArray1, destination);
	std::copy(array2, array2 + sizeArray2, destination + sizeArray1);
}

int FPGrowthMiner::calculateSupportForFpTree(CalculateSupportCountContext* context)
{
	// Assuming only one path in three
	FPNode* node = context->hashHeaderTable->find(context->currentItem)->value->node;
	return node->itemCount;
}

int FPGrowthMiner::calculateSupportForFPTreeNonSinglePath(FPNode* bottomNode, int forItem)
{
	int supportCount = 0;

	FPNode* node = bottomNode;
	while (node != NULL)
	{
		FPNode* parent = node->parent;
		while (parent != NULL)
		{
			if (parent->item == forItem)
			{
				supportCount += node->itemCount;
				break;
			}
			parent = parent->parent;
		}
		node = node->nextEqualItem;
	}

	return supportCount;
}

