#include "FPTree.h"
#include "../FimLogger/Logger.h"

#include <omp.h>
#include <iostream>
using std::cout;
using std::endl;
#include <cstdlib>
#include <cmath>
using std::ceil;

FPTree::FPTree()
{
	hashHeaderTable = NULL;
	memoryManager = new StackMemoryManager();
	memoryManagerFromOutside = false;
	setRootNode(allocateNode());
	headerTable.itemCounts = NULL;
	headerTable.nodes = NULL;
}

FPTree::FPTree(StackMemoryManager* stackMemoryManager)
{
	hashHeaderTable = NULL;
	memoryManager = stackMemoryManager;
	memoryManagerFromOutside = true;
	setRootNode(allocateNode());
	headerTable.itemCounts = NULL;
	headerTable.nodes = NULL;
}

FPTree::~FPTree()
{
	if (hashHeaderTable != NULL)
	{
		// No need, created in own memory manager.
		//delete hashHeaderTable;
	}
	else
	{
		if (headerTable.itemCounts != NULL)
			delete[] headerTable.itemCounts;
		if (headerTable.nodes != NULL)
			delete[] headerTable.nodes;
	}
	if (!memoryManagerFromOutside)
		delete memoryManager;
}

/***** PUBLIC FUNCTIONS *****/

void FPTree::setHashHeaderTable(HashTable<HashHeaderTableElement*>* table)
{
	hashHeaderTable = table;
}

HashTable<HashHeaderTableElement*>* FPTree::getHashHeaderTable()
{
	return hashHeaderTable;
}

FPNode* FPTree::getRootNode()
{
	return rootNode;
}

HeaderTable* FPTree::getHeaderTable()
{
	return &headerTable;
}

/*
Can only be used for conditional pattern trees (i.e. when hashHeaderTable is not NULL).
*/
bool FPTree::containsSinglePathOnly(int item)
{
	FPNode* node = hashHeaderTable->find(item)->value->node;
	if (node != NULL && node->nextEqualItem != NULL)
		return false;
	return true;
}

void FPTree::insertTransaction(int* transaction, int itemCount, int supportCount)
{
	FPNode* currentNode = rootNode;
	for (int i = 0; i < itemCount; i++)
	{
		// see if node exists in children
		FPNode* child = currentNode->firstChild;
		while (child != NULL)
		{
			if (child->item == transaction[i])
			{
				currentNode = child;
				break;
			}
			else
				child = child->nextSibling;
		}

		if (child == NULL) // Item does not exist in the FPTree, hence add it and all its children.
		{
			for (int j = i; j < itemCount; j++)
			{
				FPNode* lastFirstChild = currentNode->firstChild;
				FPNode* newNode = allocateNode();
				newNode->item = transaction[j];
				if (supportCount != -1)
					newNode->itemCount = supportCount;
				else
					newNode->itemCount = 1;
				newNode->parent = currentNode;
				newNode->nextSibling = lastFirstChild;

				// Update header table
				if (hashHeaderTable == NULL)
				{
					FPNode* lastFirstNodeInHeader = headerTable.nodes[transaction[j]];
					headerTable.nodes[transaction[j]] = newNode;
					newNode->nextEqualItem = lastFirstNodeInHeader;
				}
				else
				{
					HashHeaderTableElement* element = hashHeaderTable->find(transaction[j])->value;
					FPNode* lastFirstNodeInHeader = element->node;
					element->node = newNode;
					newNode->nextEqualItem = lastFirstNodeInHeader;
					element->itemCount += supportCount;
				}

				currentNode->firstChild = newNode;
				currentNode = newNode;
			}

			return; // All of the items are added, hence return.
		}
		else
		{
			if (supportCount != -1)
				child->itemCount += supportCount;
			else
				child->itemCount++;
		}
	}
}

/***** PRIVATE FUNCTIONS *****/

void FPTree::setRootNode(FPNode* root)
{
	rootNode = root;
	root->item = -1;
}

FPNode* FPTree::allocateNode()
{
	FPNode* node = (FPNode*)memoryManager->grow(sizeof(FPNode));
	node->itemCount = 0;
	node->parent = NULL;
	node->firstChild = NULL;
	node->nextSibling = NULL;
	node->nextEqualItem = NULL;
	return node;
}

void FPTree::printFPTree(FPNode* node)
{
	if (node == NULL)
		return;

	cout << "id:" << node->item << " c:" << node->itemCount << " " << node << endl;

	printFPTree(node->firstChild);

	FPNode* sibling = node->nextSibling;
	while (sibling != NULL)
	{
		cout << endl;
		cout << "parent " << node->parent->item << endl;
		printFPTree(sibling);
		sibling = sibling->nextSibling;
	}
}
