#include "ShaFemMiner.h"
#include "FPTree.h"
#include "DatasetPreprocessor.h"

#include <omp.h>
#include <cmath>
#include <algorithm>
#include <iostream>
using std::cout;
using std::endl;

// This only works with 1 thread
//#define MEASURE_TIME

ShaFemMiner::ShaFemMiner(CfpArray* cfpArray, int numThreads, LogData* logData)
	: Miner(cfpArray, numThreads, logData), STEP(32), N(9)
{
	K = new int[numThreads];
	for (int i = 0; i < numThreads; i++)
		K[i] = 0;
	X = new int*[numThreads];
	for (int i = 0; i < numThreads; i++)
	{
		X[i] = new int[N];
		for (int j = 0; j < N; j++)
			X[i][j] = 0;
	}

	timeBitVectorMining = 0.0;
	memoryManagers = new StackMemoryManager*[numThreads];
	for (int i = 0; i < numThreads; i++)
	{
		memoryManagers[i] = new StackMemoryManager();
	}
}

ShaFemMiner::~ShaFemMiner()
{
	delete[] K;

	for (int i = 0; i < numThreads; i++)
	{
		delete[] X[i];
		delete memoryManagers[i];
	}
	delete[] X;
	delete[] memoryManagers;
}

/***** PUBLIC FUNCTIONS *****/

void ShaFemMiner::startMining()
{
	string startMessageMining = "Mining procedure started.\n";
	writeToLog(startMessageMining);
	double timeMining = omp_get_wtime();

	ParallelMinePattern();

	double totalTime = omp_get_wtime() - timeMining;
	logData->mineTime = totalTime;
	string endMessageMining = "Mining procedure finished. [" + toString(totalTime) + "s]\n\n";
	writeToLog(endMessageMining);

#ifdef MEASURE_TIME
	writeToLog("Bit vector time mining: " + toString(timeBitVectorMining) + "s\n");
	writeToLog("FPGrowth time mining: " + toString(totalTime - timeBitVectorMining) + "s\n\n");
#endif
}

/***** PRIVATE FUNCTIONS *****/

void ShaFemMiner::ParallelMinePattern()
{
	#pragma omp parallel for schedule(dynamic, 1) num_threads(numThreads)
	for (int item = cfpArray.getSubarrayCount(); item >= 1; --item) // Smallest item is 1 and not 0
	{
		int threadNr = omp_get_thread_num();

		int itemSupportCount = DatasetPreprocessor::supportCount[cfpArray.getRemapToInitialIdArray()[item]];
		if (itemSupportCount < minNumTransactions || itemSupportCount == 0) // itemSupportCount == 0 then continue, must be in at least one transaction
			continue; // current item is not frequent

		if (addToFrequentItemsets(threadNr, &item, 1, itemSupportCount))
			continue; // Itemset threshold reached!

		unsigned int previousBufferNr = memoryManagers[threadNr]->getCurrentBufferNr();
		long long previousByteIndex = memoryManagers[threadNr]->getCurrentByteIndexInBuffer();
		FPTree conditionalFPTree(memoryManagers[threadNr]);

		int numNewPatterns;
		int sizeOfConditionalPatternBase = constructConditionalFpTree(item, &conditionalFPTree, &numNewPatterns);

		UpdateK(numNewPatterns, sizeOfConditionalPatternBase, threadNr);
		if (sizeOfConditionalPatternBase > K[threadNr])
		{
			// MineFPTree
			MineFPTree(&conditionalFPTree, &item, 1);
		}
		else
		{
			// MineBitVector
			StartMineBitVector(conditionalFPTree.getHashHeaderTable()->find(item)->value->node, sizeOfConditionalPatternBase);
		}

		memoryManagers[threadNr]->setToPreviousState(previousBufferNr, previousByteIndex);
	}
}

void ShaFemMiner::MineFPTree(FPTree* tree, int* suffix, int suffixSize)
{
	int threadNr = omp_get_thread_num();
	if (tree->containsSinglePathOnly(suffix[0]))
	{
		CustomVector<int> path(memoryManagers[threadNr]);
		FPNode* node = tree->getRootNode()->firstChild;
		while (node != NULL)
		{
			if (node->item == suffix[0])
				break;
			path.push_back(node->item);
			node = node->firstChild;
		}

		CalculateSupportCountContext context;
		context.hashHeaderTable = tree->getHashHeaderTable();
		context.currentItem = suffix[0];
		addRestOfItemsToFrequentItemsets(
			&path, suffix, suffixSize,
			&ShaFemMiner::calculateSupportForFpTree,
			&context
		);
	}
	else
	{
		HashTable<HashHeaderTableElement*>* hashHeaderTable = tree->getHashHeaderTable();
		HashTableIterator<HashHeaderTableElement*> it;
		hashHeaderTable->begin(&it);
		while (it.value != NULL)
		{
			unsigned int previousBufferNr = memoryManagers[threadNr]->getCurrentBufferNr();
			long long previousByteIndex = memoryManagers[threadNr]->getCurrentByteIndexInBuffer();

			if (it.value->id == suffix[0]) // Element already added.
			{
				hashHeaderTable->next(&it);
				continue;
			}

			int item = it.value->id;
			FPNode* bottomNode = hashHeaderTable->find(suffix[0])->value->node;
			int supportCount = calculateSupportForFPTreeNonSinglePath(bottomNode, item);
			if (supportCount < minNumTransactions || supportCount == 0)
			{
				hashHeaderTable->next(&it);
				continue;
			}

			int currentItemsetSize = 1 + suffixSize;
			int* currentItemset = (int*)memoryManagers[threadNr]->grow(sizeof(int) * currentItemsetSize);
			joinTwoArrays(&item, 1, suffix, suffixSize, currentItemset);
			if (addToFrequentItemsets(threadNr, currentItemset, currentItemsetSize, supportCount))
				return; // Itemset threshold reached!

			int numNewPatterns;
			int sizeOfConditionalPatternBase = getSizeOfConditionalPatternBase(hashHeaderTable->find(item)->value->node, &numNewPatterns);

			UpdateK(numNewPatterns, sizeOfConditionalPatternBase, threadNr);
			if (sizeOfConditionalPatternBase > K[threadNr])
			{
				// MineFPTree
				FPTree conditionalFPTree(memoryManagers[threadNr]);
				constructConditionalFpTree(hashHeaderTable->find(item)->value->node, &conditionalFPTree);
				MineFPTree(&conditionalFPTree, currentItemset, currentItemsetSize);
			}
			else
			{
				// MineBitVector
				StartMineBitVector(hashHeaderTable->find(item)->value->node, sizeOfConditionalPatternBase);
			}

			hashHeaderTable->next(&it);

			memoryManagers[threadNr]->setToPreviousState(previousBufferNr, previousByteIndex);
		}
	}
}

void ShaFemMiner::StartMineBitVector(FPNode* node, int sizeOfConditionalPatternBase)
{
#ifdef MEASURE_TIME
	double startTimestamp = omp_get_wtime();
#endif

	int threadNr = omp_get_thread_num();
	unsigned int previousBufferNr = memoryManagers[threadNr]->getCurrentBufferNr();
	long long previousByteIndex = memoryManagers[threadNr]->getCurrentByteIndexInBuffer();

	BitsetVector bitsetVector;
	HashTable<int> unorderedItems(memoryManagers[threadNr], DatasetPreprocessor::getLongestTransaction());
	putAllItemsFromConditionalPatternBaseToSet(&unorderedItems, node);
	int numItemsInPatternBase = (int)unorderedItems.size();
	bitsetVector.items = (int*)memoryManagers[threadNr]->grow(sizeof(int) * numItemsInPatternBase);
	HashTableIterator<int> it;
	unorderedItems.begin(&it);
	int index = 0;
	while (it.value != NULL)
	{
		bitsetVector.items[index++] = it.value->value;
		unorderedItems.next(&it);
	}
	qsort(bitsetVector.items, numItemsInPatternBase, sizeof(int), &compareItem);

	bitsetVector.itemCount = numItemsInPatternBase;
	bitsetVector.bitsets = (Bitset*)memoryManagers[threadNr]->grow(sizeof(Bitset) * numItemsInPatternBase);
	int numBytes = ceil(sizeOfConditionalPatternBase / 8.0);
	for (int i = 0; i < numItemsInPatternBase; i++)
	{
		bitsetVector.bitsets[i].numBits = sizeOfConditionalPatternBase;
		bitsetVector.bitsets[i].bytes = (byte*)memoryManagers[threadNr]->grow(sizeof(byte) * bitsetVector.bitsets[i].numBits);
		for (int j = 0; j < numBytes; j++)
			bitsetVector.bitsets[i].bytes[j] = 0;
	}
	initializeBitsets(&bitsetVector, node);

	int* weights = (int*)memoryManagers[threadNr]->grow(sizeof(int) * sizeOfConditionalPatternBase);
	setWeights(weights, node);

	MineBitVector(&bitsetVector, weights, &node->item, 1);

	memoryManagers[threadNr]->setToPreviousState(previousBufferNr, previousByteIndex);

#ifdef MEASURE_TIME
	timeBitVectorMining += omp_get_wtime() - startTimestamp;
#endif
}

/*
Assuming that the values in bitVectors are sorted in support-descending order of their items.
*/
void ShaFemMiner::MineBitVector(BitsetVector* bitsetVector, int* weights, int* suffix, int suffixSize)
{
	int threadNr = omp_get_thread_num();
	CustomVector<int> items(memoryManagers[threadNr], bitsetVector->itemCount * bitsetVector->itemCount);
	CustomVector<Bitset> bitsets(memoryManagers[threadNr], bitsetVector->itemCount * bitsetVector->itemCount);
	unsigned int previousBufferNr = memoryManagers[threadNr]->getCurrentBufferNr();
	long long previousByteIndex = memoryManagers[threadNr]->getCurrentByteIndexInBuffer();

	for (int i = 0; i < bitsetVector->itemCount; i++)
	{
		int currentItemsetSize = 1 + suffixSize;
		int* currentItemset = (int*)memoryManagers[threadNr]->grow(sizeof(int) * currentItemsetSize);
		joinTwoArrays(&(bitsetVector->items[i]), 1, suffix, suffixSize, currentItemset);
		int supportCount = getSupport(&bitsetVector->bitsets[i], weights);

		if (supportCount >= minNumTransactions && supportCount != 0)
		{
			if (addToFrequentItemsets(threadNr, currentItemset, currentItemsetSize, supportCount))
				return; // Itemset threshold reached!
		}
		else
			continue;



		bool itemInAllItemsets = true;
		for (int j = 0; j < i; j++)
		{
			Bitset* result = (Bitset*)memoryManagers[threadNr]->grow(sizeof(Bitset));
			result->bytes = (byte*)memoryManagers[threadNr]->grow(sizeof(byte) * (int)ceil(bitsetVector->bitsets[i].numBits / 8.0));
			result->numBits = bitsetVector->bitsets[i].numBits;
			bitwiseAnd(bitsetVector->bitsets[i], bitsetVector->bitsets[j], result);
			supportCount = getSupport(result, weights);

			if (supportCount >= minNumTransactions && supportCount != 0)
			{
				items.push_back(bitsetVector->items[j]);
				bitsets.push_back(*result);

				if (!areEqual(*result, bitsetVector->bitsets[i]))
					itemInAllItemsets = false;
			}
		}

		if (itemInAllItemsets)
		{
			// For each combination x of the items output x UNION currentItemset

			CalculateSupportCountContext context;
			context.bitsetVector = bitsetVector;
			context.bitsets = &bitsets;
			context.weights = weights;
			addRestOfItemsToFrequentItemsets(
				&items, currentItemset, currentItemsetSize,
				&ShaFemMiner::calcuateSupportForBitsetVector,
				&context
			);
		}
		else if (!items.empty())
		{
			BitsetVector* newBitsetVector = (BitsetVector*)memoryManagers[threadNr]->grow(sizeof(BitsetVector));
			newBitsetVector->itemCount = (int)items.size();
			newBitsetVector->items = (int*)memoryManagers[threadNr]->grow(sizeof(int) * newBitsetVector->itemCount);
			newBitsetVector->bitsets = (Bitset*)memoryManagers[threadNr]->grow(sizeof(Bitset) * newBitsetVector->itemCount);

			unsigned int index = 0;
			for (int j = 0; j < i; j++)
			{
				if (items.size() <= index)
					break;

				if (items.at(index) == bitsetVector->items[j])
				{
					newBitsetVector->items[index] = items.at(index);
					newBitsetVector->bitsets[index].bytes = bitsets.at(index).bytes;
					newBitsetVector->bitsets[index].numBits = bitsets.at(index).numBits;
					index++;
				}
			}

			MineBitVector(newBitsetVector, weights, currentItemset, currentItemsetSize);
		}

		items.clear();
		bitsets.clear();
	}
	memoryManagers[threadNr]->setToPreviousState(previousBufferNr, previousByteIndex);
}

void ShaFemMiner::UpdateK(int numNewPatterns, int size, int threadNr)
{
	for (int i = 0; i < N; i++)
	{
		if (size > i * STEP)
			X[threadNr][i] += numNewPatterns;
		else
			break;
	}

	K[threadNr] = 0;
	for (int i = N - 1; i >= 1; i--)
	{
		if (X[threadNr][i - 1] >= 2 * X[threadNr][i])
		{
			K[threadNr] = (i + 1) * STEP;
			break;
		}
	}
}

int ShaFemMiner::getSizeOfConditionalPatternBase(FPNode* bottomNode, int* numUniqueItems)
{
	FPNode* node = bottomNode;
	if (node == NULL)
		return 0;

	HashTable<int> uniqueItems(memoryManagers[omp_get_thread_num()], DatasetPreprocessor::getLongestTransaction());
	uniqueItems.insert(node->item, node->item);
	int sizeOfConditionalPatternBase = 1;
	while (node->nextEqualItem != NULL)
	{
		FPNode* parent = node->parent;
		while (parent != NULL)
		{
			if (parent->item < 0)
				break; // This is root node.

			if (uniqueItems.find(parent->item) == NULL)
				uniqueItems.insert(parent->item, parent->item);
			parent = parent->parent;
		}

		node = node->nextEqualItem;
		sizeOfConditionalPatternBase++;
	}

	*numUniqueItems = (int)uniqueItems.size();

	return sizeOfConditionalPatternBase;
}

void ShaFemMiner::joinTwoArrays(int* array1, int sizeArray1, int* array2, int sizeArray2, int* destination)
{
	std::copy(array1, array1 + sizeArray1, destination);
	std::copy(array2, array2 + sizeArray2, destination + sizeArray1);
}

int ShaFemMiner::getSupport(Bitset* bitset, int* weights)
{
	int supportCount = 0;
	unsigned int numBytes = (unsigned int)ceil(bitset->numBits / 8.0);
	for (unsigned int i = 0; i < numBytes; i++)
	{
		unsigned char byteValue = bitset->bytes[i];
		unsigned int bitNr = i * 8;
		while (byteValue != 0)
		{
			if (byteValue & 128)
				supportCount += weights[bitNr];
			byteValue <<= 1;
			bitNr++;
		}
	}
	return supportCount;
}

void ShaFemMiner::putAllItemsFromConditionalPatternBaseToSet(HashTable<int> *set, FPNode* node)
{
	if (node == NULL)
		return;

	// set->insert(node->item); // Don't add current item, only the ones above.
	while (node != NULL)
	{
		FPNode* parent = node->parent;
		while (parent != NULL)
		{
			if (parent->item < 0)
				break; // This is root node

			if (set->find(parent->item) == NULL)
				set->insert(parent->item, parent->item);
			parent = parent->parent;
		}

		node = node->nextEqualItem;
	}
}

void ShaFemMiner::initializeBitsets(BitsetVector* bitsetVector, FPNode* node)
{
	if (node == NULL)
		return;

	int currentTransactionBit = 0;
	int itemIndex = 0;

	// Don't take current item into consideration, only the ones above.
	/*while (bitsetVector->items[itemIndex] != node->item)
		itemIndex++;
	setBitAt(bitsetVector->bitsets[itemIndex], currentTransactionBit, true);*/

	while (node != NULL)
	{
		FPNode* parent = node->parent;
		while (parent != NULL)
		{
			if (parent->item < 0)
				break; // This is root node

			itemIndex = 0;
			while (bitsetVector->items[itemIndex] != parent->item)
				itemIndex++;
			setBitAt(bitsetVector->bitsets[itemIndex], currentTransactionBit, true);

			parent = parent->parent;
		}

		node = node->nextEqualItem;
		currentTransactionBit++;
	}
}

void ShaFemMiner::setWeights(int* weights, FPNode* node)
{
	if (node == NULL)
		return;

	int currentIndex = 0;
	while (node != NULL)
	{
		weights[currentIndex++] = node->itemCount;
		node = node->nextEqualItem;
	}
}

/*
Return the size of the conditional pattern base.
*/
int ShaFemMiner::constructConditionalFpTree(unsigned int initialItem, FPTree* fpTree, int *numUniqueItems)
{
	int threadNr = omp_get_thread_num();
	CustomVector<CustomVector<int>* > conditionalPatternBase(memoryManagers[threadNr]);
	CustomVector<int> supportCounts(memoryManagers[threadNr]);
	HashTable<int> itemsInPatternBase(memoryManagers[threadNr], DatasetPreprocessor::getLongestTransaction());
	itemsInPatternBase.insert(initialItem, initialItem);
	int sizeOfConditionalPatternBase = 0;

	unsigned int item = initialItem;
	unsigned long long byteIndex = cfpArray.getSubarrayStartIndices()[item];
	unsigned long long endByteIndex = item >= (unsigned int)cfpArray.getSubarrayCount() - 1 ?
		cfpArray.getArraySize() : cfpArray.getSubarrayStartIndices()[item + 1];

	while (byteIndex < endByteIndex) // for all leaf nodes in prefix tree
	{
		unsigned int deltaItem;
		long long deltaPos;
		unsigned int count;
		unsigned long long index = byteIndex;
		byteIndex = cfpArray.readElementFromArray(byteIndex, deltaItem, deltaPos, &count);

		CustomVector<int>* cv = (CustomVector<int>*)memoryManagers[threadNr]->grow(sizeof(CustomVector<int>));
		conditionalPatternBase.push_back(cv);
		cv->constructor(memoryManagers[threadNr]);
		conditionalPatternBase.back()->push_back(item);
		supportCounts.push_back(count);

		while (deltaItem != item) // traverse a path in the prefix tree
		{
			index = (index - cfpArray.getSubarrayStartIndices()[item]) - deltaPos;
			item -= deltaItem;
			index += cfpArray.getSubarrayStartIndices()[item]; // parent byte index*/
			cfpArray.readElementFromArray(index, deltaItem, deltaPos, NULL);

			conditionalPatternBase.back()->push_back(item);
			if (itemsInPatternBase.find(item) == NULL)
				itemsInPatternBase.insert(item, item);
		}
		item = initialItem; // reset item

		sizeOfConditionalPatternBase++;
	}

	// Reverse the arrays so they are sorted in frequency decreesing order.
	for (size_t i = 0; i < conditionalPatternBase.size(); i++)
	{
		int* array = &(conditionalPatternBase[i]->at(0));
		int size = conditionalPatternBase[i]->size() / 2;
		for (int j = 0; j < size; j++)
		{
			int temp = array[j];
			array[j] = array[conditionalPatternBase[i]->size() - j - 1];
			array[conditionalPatternBase[i]->size() - j - 1] = temp;
		}
	}

	*numUniqueItems = itemsInPatternBase.size();

	// Create header table
	int numItems = (int)itemsInPatternBase.size();
	if (numItems > 0)
	{
		HashTable<HashHeaderTableElement*>* hashHeaderTable =
				(HashTable<HashHeaderTableElement*>*)memoryManagers[threadNr]->grow(sizeof(HashTable<HashHeaderTableElement*>));
		hashHeaderTable->constructor(memoryManagers[threadNr], numItems);
		HashTableIterator<int> it;
		itemsInPatternBase.begin(&it);
		while (it.value != NULL)
		{
			HashHeaderTableElement* element = (HashHeaderTableElement*)memoryManagers[threadNr]->grow(sizeof(HashHeaderTableElement));
			element->itemCount = 0;
			element->node = NULL;
			hashHeaderTable->insert(it.value->value, element);
			itemsInPatternBase.next(&it);
		}
		fpTree->setHashHeaderTable(hashHeaderTable);
	}
	else
		fpTree->setHashHeaderTable(NULL);

	// Create conditional FP-tree
	for (size_t i = 0; i < conditionalPatternBase.size(); i++)
	{
		if (conditionalPatternBase[i]->size() > 0)
		{
			fpTree->insertTransaction(
				&(conditionalPatternBase[i]->at(0)),
				(int)conditionalPatternBase[i]->size(),
				supportCounts[i]
			);
		}

		// No need created in custom memory manager
		//delete conditionalPatternBase[i];
	}

	return sizeOfConditionalPatternBase;
}

void ShaFemMiner::constructConditionalFpTree(FPNode* firstNode, FPTree* fpTree)
{
	int threadNr = omp_get_thread_num();

	// Get the conditional patter base for the given item.
	CustomVector<CustomVector<int>* > conditionalPatternBase(memoryManagers[threadNr]);
	CustomVector<int> supportCounts(memoryManagers[threadNr]);
	HashTable<int> itemsInPatternBase(memoryManagers[threadNr], DatasetPreprocessor::getLongestTransaction());
	FPNode* node = firstNode;
	int i = 0;
	while (node != NULL)
	{
		CustomVector<int>* cv = (CustomVector<int>*)memoryManagers[threadNr]->grow(sizeof(CustomVector<int>));
		conditionalPatternBase.push_back(cv);
		cv->constructor(memoryManagers[threadNr]);

		conditionalPatternBase[i]->push_back(node->item);
		supportCounts.push_back(node->itemCount);
		if (itemsInPatternBase.find(node->item) == NULL)
			itemsInPatternBase.insert(node->item, node->item);

		FPNode* parent = node->parent;
		while (parent != NULL)
		{
			if (parent->item < 0)
				break; // This is root node

			conditionalPatternBase[i]->push_back(parent->item);
			if (itemsInPatternBase.find(parent->item) == NULL)
				itemsInPatternBase.insert(parent->item, parent->item);
			parent = parent->parent;
		}

		i++;
		node = node->nextEqualItem;
	}

	// Reverse the arrays so they are sorted in frequency decreesing order.
	for (size_t i = 0; i < conditionalPatternBase.size(); i++)
	{
		int* array = &(conditionalPatternBase[i]->at(0));
		int size = conditionalPatternBase[i]->size() / 2;
		for (int j = 0; j < size; j++)
		{
			int temp = array[j];
			array[j] = array[conditionalPatternBase[i]->size() - j - 1];
			array[conditionalPatternBase[i]->size() - j - 1] = temp;
		}
	}

	// Create header table
	int numItems = (int)itemsInPatternBase.size();
	if (numItems > 0)
	{
		HashTable<HashHeaderTableElement*>* hashHeaderTable =
				(HashTable<HashHeaderTableElement*>*)memoryManagers[threadNr]->grow(sizeof(HashTable<HashHeaderTableElement*>));
		hashHeaderTable->constructor(memoryManagers[threadNr], numItems);
		HashTableIterator<int> it;
		itemsInPatternBase.begin(&it);
		while (it.value != NULL)
		{
			HashHeaderTableElement* element = (HashHeaderTableElement*)memoryManagers[threadNr]->grow(sizeof(HashHeaderTableElement));
			element->itemCount = 0;
			element->node = NULL;
			hashHeaderTable->insert(it.value->value, element);
			itemsInPatternBase.next(&it);
		}
		fpTree->setHashHeaderTable(hashHeaderTable);
	}
	else
		fpTree->setHashHeaderTable(NULL);

	// Create conditional FP-tree
	for (size_t i = 0; i < conditionalPatternBase.size(); i++)
	{
		if (conditionalPatternBase[i]->size() > 0)
		{
			fpTree->insertTransaction(
				&(conditionalPatternBase[i]->at(0)),
				(int)conditionalPatternBase[i]->size(),
				supportCounts[i]
			);
		}

		// No need created in custom memory manager
		//delete conditionalPatternBase[i];
	}
}

void ShaFemMiner::addRestOfItemsToFrequentItemsets(CustomVector<int>* items,
													int* currentItemset,
													int currentItemsetSize,
													int (ShaFemMiner::*calcSupport)(CalculateSupportCountContext*),
													CalculateSupportCountContext* context)
{
	int threadNr = omp_get_thread_num();

	// Power set using counter
	unsigned int power = 1 << (unsigned int)items->size();
	for (unsigned int i = 1; i < power; i++)
	{
		unsigned int previousBufferNr = memoryManagers[threadNr]->getCurrentBufferNr();
		long long previousByteIndex = memoryManagers[threadNr]->getCurrentByteIndexInBuffer();

		unsigned int bitCounter = i;
		unsigned int setSize = __builtin_popcount(bitCounter);
		unsigned int outputSize = setSize + currentItemsetSize;
		int* output = (int*)memoryManagers[threadNr]->grow(sizeof(int) * outputSize);

		int outputIndex = 0;
		for (int j = 0; j < items->size() && j < sizeof(unsigned int) * 8; j++)
		{
			if (bitCounter & 1)
				output[outputIndex++] = items->at(j);
			bitCounter >>= 1;
		}
		for (int j = 0; j < currentItemsetSize; j++)
			output[j + setSize] = currentItemset[j];

		// calculate support
		context->bitCounter = bitCounter;
		context->itemCount = setSize;
		int support = (this->*calcSupport)(context);

		// add to frequent itemset
		if (addToFrequentItemsets(omp_get_thread_num(), output, outputSize, support))
			return; // Itemset threshold reached!

		memoryManagers[threadNr]->setToPreviousState(previousBufferNr, previousByteIndex);
	}
}

int ShaFemMiner::calcuateSupportForBitsetVector(CalculateSupportCountContext* context)
{
	int threadNr = omp_get_thread_num();
	Bitset* finalBitset = (Bitset*)memoryManagers[threadNr]->grow(sizeof(Bitset));
	finalBitset->bytes = (byte*)memoryManagers[threadNr]->grow(
		sizeof(byte) * (int)ceil(context->bitsetVector->bitsets[0].numBits / 8.0)
	);
	finalBitset->numBits = context->bitsetVector->bitsets[0].numBits;
	setBitsetValues(*finalBitset, true);
	int index = 0;
	for (int j = 0; j < sizeof(unsigned int) * 8; j++)
	{
		if (context->bitCounter & 1)
			bitwiseAnd(*finalBitset, context->bitsets->at(index++), finalBitset);
		context->bitCounter >>= 1;
	}
	int support = getSupport(finalBitset, context->weights);

	return support;
}

int ShaFemMiner::calculateSupportForFpTree(CalculateSupportCountContext* context)
{
	// Assuming only one path in three
	FPNode* node = context->hashHeaderTable->find(context->currentItem)->value->node;
	return node->itemCount;
}

int ShaFemMiner::calculateSupportForFPTreeNonSinglePath(FPNode* bottomNode, int forItem)
{
	int supportCount = 0;

	FPNode* node = bottomNode;
	while (node != NULL)
	{
		FPNode* parent = node->parent;
		while (parent != NULL)
		{
			if (parent->item == forItem)
			{
				supportCount += node->itemCount;
				break;
			}
			parent = parent->parent;
		}
		node = node->nextEqualItem;
	}

	return supportCount;
}
