#define _USE_MATH_DEFINES

#include "DatasetGenerator.h"

#include <cstdlib>
#include <vector>
using std::vector;

#include <iostream>
using std::cout;
using std::endl;

#include <stdlib.h>
#include <time.h>
#include <algorithm>
#include <cmath>
#include <random>

#include <omp.h>

#include <fstream>
using std::ofstream;

DatasetGenerator::DatasetGenerator(InputData& inputData)
	: inputData(inputData)
{
	averageSizeOfMegatransactions = ceil(0.3 * inputData.averageTransactionLength);
	numberOfMegatransactions = ceil(0.02 * inputData.numTransactions);
	poissonDistrMeanForMegatransactions = averageSizeOfMegatransactions;
	poissonDistrMeanForTransactions = inputData.averageTransactionLength;
	expDistrRateForMegatransactions = 2.0;
	expDistrRateForTransactions = 1.0;
	normalDistrMean = 0.5;
	normalDistrVariance = 0.1;

	generators = new std::default_random_engine*[inputData.numThreads];
	for (int i = 0; i < inputData.numThreads; i++)
		generators[i] = new std::default_random_engine(std::random_device{}());

	setAndRandomizeItems();
	createMegatransactions();
	createMegatransactionWeights();
	setTransactionLengths();
	assignMegatransactionsToTransactions();
	assignNonAssignedItems();
}

DatasetGenerator::~DatasetGenerator()
{
	for (int i = 0; i < megatransactions.size(); i++)
		delete megatransactions[i];

	delete[] transactionLengths;
	delete[] items;
	delete[] itemCount;
	delete[] megatransactionWeights;

	for (int i = 0; i < inputData.numTransactions; i++)
		delete[] transactions[i];
	delete[] transactions;

	for (int i = 0; i < inputData.numThreads; i++)
		delete generators[i];
	delete[] generators;
}

void DatasetGenerator::setAndRandomizeItems()
{
	cout << "Setting and shuffling items. ";
	cout.flush();
	items = new int[inputData.numUniqueItems];
	for (int i = 0; i < inputData.numUniqueItems; i++)
		items[i] = i + 1;
	std::random_shuffle(&items[0], &items[inputData.numUniqueItems]);
	cout << "[finished]" << endl;
}

void DatasetGenerator::createMegatransactions()
{
	cout << "Creating potential megatransactions. ";
	cout.flush();
	std::poisson_distribution<int> poissonDistribution(poissonDistrMeanForMegatransactions);
	std::exponential_distribution<double> expDistribution(expDistrRateForMegatransactions);

	for (int i = 0; i < numberOfMegatransactions; i++)
	{
		int threadNr = omp_get_thread_num();
		int size = poissonDistribution(*generators[threadNr]);
		megatransactions.push_back(new vector<ExtendedItem>);

		double fractionFromPrevious = expDistribution(*generators[threadNr]);
		fractionFromPrevious = fractionFromPrevious > 1.0 ? 1.0 : fractionFromPrevious;

		vector<ExtendedItem> availableItems;
		for (int j = 0; j < inputData.numUniqueItems * inputData.maxspan; j++)
		{
			ExtendedItem item;
			item.item = j;
			item.dat = 0;
			while (item.item >= inputData.numUniqueItems)
			{
				item.item -= inputData.numUniqueItems;
				item.dat++;
			}

			availableItems.push_back(item);
		}

		vector<ExtendedItem> copyOfLastItemset;
		if (i > 0)
		{
			for (int j = 0; j < megatransactions[i - 1]->size(); j++)
				copyOfLastItemset.push_back(megatransactions[i - 1]->at(j));
		}

		for (int j = 0; j < size; j++)
		{
			if (i > 0 && (j / (double)megatransactions[i - 1]->size()) < fractionFromPrevious && copyOfLastItemset.size() > 0)
			{ // take from previous itemset
				int index = getRandomInt(0, copyOfLastItemset.size() - 1);
				megatransactions[i]->push_back(copyOfLastItemset[index]);
				copyOfLastItemset.erase(copyOfLastItemset.begin() + index);
				availableItems.erase(std::remove(availableItems.begin(),
												 availableItems.end(),
												 megatransactions[i]->back()),
									 availableItems.end());
			}
			else
			{
				int index = getRandomInt(0, availableItems.size() - 1);
				megatransactions[i]->push_back(availableItems[index]);
				availableItems.erase(availableItems.begin() + index);
			}
		}

		unsigned int smallestDat = -1;
		for (int j = 0; j < megatransactions[i]->size(); j++)
		{
			if (megatransactions[i]->at(j).dat < smallestDat)
				smallestDat = megatransactions[i]->at(j).dat;
		}
		for (int j = 0; j < megatransactions[i]->size(); j++)
		{
			megatransactions[i]->at(j).dat -= smallestDat;
		}

		// print megatransactions
		/*
		cout << endl;
		for (int j = 0; j < megatransactions.back()->size(); j++)
		{
			cout << megatransactions.back()->at(j).item << ":";
			cout << megatransactions.back()->at(j).dat << " ";
		}
		*/
	}
	cout << "[finished]" << endl;
}

void DatasetGenerator::createMegatransactionWeights()
{
	cout << "Assigning megatransaction weights. ";
	cout.flush();
	std::exponential_distribution<double> expDistribution(expDistrRateForTransactions);

	megatransactionWeights = new double[megatransactions.size()];
	double sum = 0;
	for (int i = 0; i < megatransactions.size(); i++)
	{
		int threadNr = omp_get_thread_num();
		megatransactionWeights[i] = expDistribution(*generators[threadNr]);
		sum += megatransactionWeights[i];
	}

	// normalize
	for (int i = 0; i < megatransactions.size(); i++)
		megatransactionWeights[i] /= sum;

	// set probability intervals
	for (int i = 1; i < megatransactions.size(); i++)
		megatransactionWeights[i] += megatransactionWeights[i-1];

	cout << "[finished]" << endl;
}

void DatasetGenerator::setTransactionLengths()
{
	cout << "Setting transaction lengths. ";
	cout.flush();
	std::poisson_distribution<int> poissonDistribution(poissonDistrMeanForTransactions);

	int totalAmountOfItems = inputData.numTransactions * inputData.averageTransactionLength;
	transactionLengths = new int[inputData.numTransactions];

	for (int i = 0; i < inputData.numTransactions; i++)
	{
		int threadNr = omp_get_thread_num();
		transactionLengths[i] = poissonDistribution(*generators[threadNr]);
		if (transactionLengths[i] <= 0)
			transactionLengths[i] = 1;
		totalAmountOfItems -= transactionLengths[i];
	}

	if (totalAmountOfItems < 0)
	{
		for (int i = 0; i < inputData.numTransactions; i++)
		{
			int add = inputData.averageTransactionLength - transactionLengths[i];
			if (totalAmountOfItems + add > 0)
				add = -totalAmountOfItems;

			if (add > 0)
			{
				transactionLengths[i] += add;
				totalAmountOfItems += add;
			}

			if (totalAmountOfItems == 0)
				break;
		}
	}
	else if (totalAmountOfItems > 0)
	{
		for (int i = 0; i < inputData.numTransactions; i++)
		{
			int sub = transactionLengths[i] - inputData.averageTransactionLength;
			if (totalAmountOfItems - sub < 0)
				sub = totalAmountOfItems;

			if (sub > 0)
			{
				transactionLengths[i] -= sub;
				totalAmountOfItems -= sub;
			}

			if (totalAmountOfItems == 0)
				break;
		}
	}

	cout << "[finished]" << endl;
}

void DatasetGenerator::assignMegatransactionsToTransactions()
{
	cout << "Creating transactions. [0%]";
	cout.flush();

	transactions = new int*[inputData.numTransactions];
	for (int i = 0; i < inputData.numTransactions; i++)
		transactions[i] = new int[transactionLengths[i]];
	itemCount = new int[inputData.numUniqueItems];
	for (int i = 0; i < inputData.numUniqueItems; i++)
		itemCount[i] = 0;

	int* currentTransactionLength = new int[inputData.numTransactions];
	for (int i = 0; i < inputData.numTransactions; i++)
		currentTransactionLength[i] = 0;

	std::normal_distribution<double> distribution(normalDistrMean, sqrt(normalDistrVariance));

	double lastPercent = 0.0;
	// Assign potential megatransactions to each transaction
	for (int i = 0; i < inputData.numTransactions; i++)
	{
		// Print status
		#pragma omp critical
		{
			if (i / (double)inputData.numTransactions >= lastPercent + 0.01)
			{
				lastPercent += 0.01;
				if (lastPercent < 0.1)
					cout << "\b\b\b";
				else
					cout << "\b\b\b\b";
				cout << (int)(100 * lastPercent) << "%]";
				cout.flush();
			}
		}

		while (currentTransactionLength[i] < transactionLengths[i])
		{
			vector<ExtendedItem>* megatransaction = megatransactions[getNextMegatransactionIndex()];
			for (int j = 0; j < megatransaction->size(); j++)
			{
				int item = megatransaction->at(j).item;
				int dat = megatransaction->at(j).dat;
				bool isCorrupted = distribution(*generators[omp_get_thread_num()]) > 1.0;
				if (isCorrupted)
				{
					item = getRandomInt(0, inputData.numUniqueItems - 1);
					dat = 0;
				}

				if (i + dat >= inputData.numTransactions)
					continue;

				if (currentTransactionLength[i + dat] < transactionLengths[i + dat])
				{
					bool isItemInTransaction = false;
					for (int k = 0; k < currentTransactionLength[i + dat]; k++)
					{
						if (transactions[i + dat][k] == item)
						{
							isItemInTransaction = true;
							break;
						}
					}

					if (!isItemInTransaction)
					{
						transactions[i + dat][currentTransactionLength[i + dat]] = item;
						currentTransactionLength[i + dat]++;
						itemCount[item]++;
					}
				}
			}
		}
	}

	delete[] currentTransactionLength;

	cout << "\b\b\b\b\b" << "[finished]" << endl;
}

void DatasetGenerator::assignNonAssignedItems()
{
	cout << "Assigning non assigneed items. [0%]";
	cout.flush();

	double lastPercent = 0.0;
	// Assign non assigned items
	for (int i = 0; i < inputData.numUniqueItems; i++)
	{
		// Print status
		#pragma omp critical
		{
			if (i / (double)inputData.numTransactions >= lastPercent + 0.01)
			{
				lastPercent += 0.01;
				if (lastPercent < 0.1)
					cout << "\b\b\b";
				else
					cout << "\b\b\b\b";
				cout << (int)(100 * lastPercent) << "%]";
				cout.flush();
			}
		}

		if (itemCount[i] > 0)
			continue;

		while (true)
		{
			int transaction = getRandomInt(0, inputData.numTransactions - 1);
			int index = getRandomInt(0, transactionLengths[transaction] - 1);
			if (itemCount[transactions[transaction][index]] <= 1)
				continue;
			else
			{
				itemCount[transactions[transaction][index]]--;
				transactions[transaction][index] = i;
				itemCount[i]++;
				break;
			}
		}
	}

	cout << "\b\b\b\b\b" << "[finished]" << endl;
}

int DatasetGenerator::getNextMegatransactionIndex()
{
	int threadNr = omp_get_thread_num();
	std::uniform_real_distribution<double> distribution(0, 1);
	double randomValue = distribution(*generators[threadNr]);

	for (int i = 0; i < megatransactions.size(); i++)
	{
		if (randomValue <= megatransactionWeights[i])
			return i;
	}

	return -1; // This should never be the case
}

/*
Returns random value between min and max.
(Includes min and max)
*/
int DatasetGenerator::getRandomInt(int min, int max)
{
	if (min == max)
		return min;

	std::uniform_int_distribution<int> distribution(min, max);
	int random_int = distribution(*generators[omp_get_thread_num()]);
	return random_int;
}

void DatasetGenerator::writeToPath(string path)
{
	cout << "Writing to file: ";
	cout << path << " ";
	cout.flush();

	ofstream file;
	file.open(path.c_str(), std::ofstream::binary);

	for (int i = 0; i < inputData.numTransactions; i++)
	{
		for (int j = 0; j < transactionLengths[i]; j++)
		{
			file << transactions[i][j] << ' ';
		}
		file << char(10); // new line ASCII
	}

	file.close();

	cout << "[finished]" << endl;
}
