#include "BitTable.h"
#include "BitsetUtility.h"
#include "../FimLogger/Logger.h"

#include <omp.h>

BitTable::BitTable(MultiBuffer** multiBuffers,
				   InputData& inputData,
				   DatasetPreprocessor& datasetPreprocessor,
				   StackMemoryManager** memoryManagers,
				   LogData& logData)
	: multiBuffers(multiBuffers),
	  inputData(inputData),
	  datasetPreprocessor(datasetPreprocessor),
	  memoryManagers(memoryManagers),
	  logData(logData)
{
	numRowsForItemTable = datasetPreprocessor.getNumFrequentItems();
	numColumnsForItemTable = datasetPreprocessor.getNumTransactionsWithFrequentItems();
	numBytesPerRowForItemTable = getNumBytes(numColumnsForItemTable);

	numRowsForTransactionTable = numColumnsForItemTable;
	numColumnsForTransactionTable = numRowsForItemTable;
	numBytesPerRowForTransactionTable = getNumBytes(numColumnsForTransactionTable);

	allocateMemory();
	createBitTable();
}

BitTable::~BitTable()
{
}

/***** PUBLIC FUNCTIONS *****/

int BitTable::getNumRowsForItemTable() const
{
	return numRowsForItemTable;
}

int BitTable::getNumColumnsForItemTable() const
{
	return numColumnsForItemTable;
}

int BitTable::getNumBytesPerRowForItemTable() const
{
	return numBytesPerRowForItemTable;
}

int BitTable::getNumRowsForTransactionTable() const
{
	return numRowsForTransactionTable;
}

int BitTable::getNumColumnsForTransactionTable() const
{
	return numColumnsForTransactionTable;
}

int BitTable::getNumBytesPerRowForTransactionTable() const
{
	return numBytesPerRowForTransactionTable;
}

void BitTable::itemIdToExtendedItem(unsigned int itemId,
									unsigned int* itemOut,
									unsigned char* dimensionalAttributeOut) const
{
	*itemOut = datasetPreprocessor.getRemapToInitialIdArray()[itemId];
	*dimensionalAttributeOut = 0;
	static const int itemsPerDimAttr = (datasetPreprocessor.getSupportBufferSize() / (inputData.maxspan + 1));
	while ((int)*itemOut - itemsPerDimAttr > 0)
	{
		*itemOut -= itemsPerDimAttr;
		(*dimensionalAttributeOut)++;
	}
}

void BitTable::extendedItemToItemId(unsigned int item,
									unsigned char dimensionalAttribute,
									unsigned int* itemIdOut) const
{
	static const int itemsPerDimAttr = (datasetPreprocessor.getSupportBufferSize() / (inputData.maxspan + 1));
	*itemIdOut = item + dimensionalAttribute * itemsPerDimAttr;
	*itemIdOut = datasetPreprocessor.getInitialToRemapIdArray()[*itemIdOut];
}

void BitTable::getItemsFromSubsumbeBitVector(unsigned char* subsumeBitVector,
											 unsigned int* subsumeItemsOut,
											 unsigned int itemToIgnore) const
{
	int index = 0;
	for (unsigned int i = 0; i < numBytesPerRowForTransactionTable; i++)
	{
		unsigned char byteValue = subsumeBitVector[i];
		unsigned int bitNr = i * 8;
		while (byteValue != 0)
		{
			if (byteValue & 128 && bitNr != itemToIgnore)
				subsumeItemsOut[index++] = bitNr;
			byteValue <<= 1;
			bitNr++;
		}
	}
}

bool BitTable::isItemInTransaction(int row, int column) const
{
	return getBitAt(itemTable[row], column);
}

unsigned char* BitTable::getItemBitVector(unsigned int item) const
{
	return itemTable[item];
}

unsigned char* BitTable::getTransactionBitVector(unsigned int transactionNr) const
{
	return transactionTable[transactionNr];
}

unsigned long long BitTable::getMemoryUsageInBytes() const
{
	unsigned long long totalMemoryUsage = 0;
	for (int i = 0; i < inputData.numThreads; i++)
		totalMemoryUsage += memoryManagers[i]->getMemoryUsageInBytes();
	return totalMemoryUsage;
}

/***** PRIVATE FUNCTIONS *****/

void BitTable::allocateMemory()
{
	string startMessage = "Allocating BitTable memory.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	itemTable = (unsigned char**)memoryManagers[0]->grow(sizeof(unsigned char*) * numRowsForItemTable);
	#pragma omp parallel for num_threads(inputData.numThreads)
	for (int i = 0; i < numRowsForItemTable; i++)
	{
		itemTable[i] = memoryManagers[omp_get_thread_num()]->grow(numBytesPerRowForItemTable);
		zeroBytes(itemTable[i], numBytesPerRowForItemTable);
	}

	transactionTable = (unsigned char**)memoryManagers[0]->grow(sizeof(unsigned char*) * numRowsForTransactionTable);
	#pragma omp parallel for num_threads(inputData.numThreads)
	for (int i = 0; i < numRowsForTransactionTable; i++)
	{
		transactionTable[i] = memoryManagers[omp_get_thread_num()]->grow(numBytesPerRowForTransactionTable);
		zeroBytes(transactionTable[i], numBytesPerRowForTransactionTable);
	}

	logData.preprocessTime += omp_get_wtime() - time;
	string endMessage = "Memory allocation finished. [" + toString(omp_get_wtime() - time) + "s]\n";
	writeToLog(endMessage);
	double memoryUsageInMB = getMemoryUsageInBytes() / (double)(1024.0 * 1024.0);
	writeToLog("BitTable memory usage: " + toString(memoryUsageInMB) + "MB\n\n");
}

void BitTable::createBitTable()
{
	string startMessage = "Creating BitTable.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	#pragma omp parallel num_threads(inputData.numThreads)
	{
		int threadNr = omp_get_thread_num();

		bool containsFrequentItem = false;
		int firstTransactionNr = getFirstTransactionNr(threadNr);
		int transactionNr = firstTransactionNr;
		int item = -1;
		for (int i = 0; i < multiBuffers[threadNr]->getNumBytes(); i++)
		{
			char byte = multiBuffers[threadNr]->getByteAndDelete(&i);

			if (isdigit(byte))
			{
				if (item == -1)
					item = 0;
				else
					item *= 10;

				item += byte - '0';
			}
			else
			{
				if (item != -1 && datasetPreprocessor.isItemFrequent(item))
				{
					containsFrequentItem = true;
					int transactionByteNr = transactionNr / 8;
					int firstTransactionByteNr = firstTransactionNr / 8;
					int lastTransactionByteNr = getLastTransactionNr(threadNr) / 8;
					bool isSharedByte = transactionByteNr == firstTransactionByteNr;
					isSharedByte = isSharedByte || transactionByteNr == lastTransactionByteNr;
					if (isSharedByte)
					{
						#pragma omp critical
						{
							setBitsToTrue(item, transactionNr);
						}
					}
					else
						setBitsToTrue(item, transactionNr);
				}
				item = -1;

				if (byte == '\n' && containsFrequentItem)
				{
					transactionNr++;
					containsFrequentItem = false;
				}
			}
		}
	}

	logData.preprocessTime += omp_get_wtime() - time;
	string endMessage = "BitTable created. [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);
}

void BitTable::setBitsToTrue(int item, int transactionNr)
{
	for (int dim = 0; dim < inputData.maxspan + 1; dim++)
	{
		if (transactionNr - dim < 0)
			break;

		int extendedItem = item + dim * (datasetPreprocessor.getSupportBufferSize() / (inputData.maxspan + 1));
		int index = datasetPreprocessor.getInitialToRemapIdArray()[extendedItem];
		int transactionIndex = transactionNr - dim;
		if (!datasetPreprocessor.isItemFrequent(extendedItem) || transactionIndex < 0)
			continue;

		setBitAt(itemTable[index], transactionIndex, true);
		setBitAt(transactionTable[transactionIndex], index, true);
	}
}

int BitTable::getFirstTransactionNr(int threadNr) const
{
	if (threadNr <= 0)
		return 0;

	int sum = 0;
	for (int i = 0; i < threadNr; i++)
		sum += datasetPreprocessor.getNumTransactionsForThread(i);

	return sum;
}

int BitTable::getLastTransactionNr(int threadNr) const
{
	if (threadNr < 0)
		return 0;

	int sum = 0;
	for (int i = 0; i <= threadNr; i++)
		sum += datasetPreprocessor.getNumTransactionsForThread(i);
	sum = sum == 0 ? 0 : sum - 1;

	return sum;
}
