#include "DatasetPreprocessor.h"
#include "../FimLogger/Logger.h"

#include <omp.h>
#include <cmath>

int* DatasetPreprocessor::supportCount = NULL;

DatasetPreprocessor::DatasetPreprocessor(MultiBuffer** multiBuffers, InputData &inputData)
	: multiBuffers(multiBuffers), inputData(inputData)
{
	numTransactionsForThread = new int[inputData.numThreads];
	for (int i = 0; i < inputData.numThreads; i++)
		numTransactionsForThread[i] = 0;

	countSupport();
	countTransactionsWithFrequentItems();
	countNumberOfFrequentItems();
	remapItemsByFrequency();
}

DatasetPreprocessor::~DatasetPreprocessor()
{
	delete[] numTransactionsForThread;
	delete[] supportCount;
	delete[] initialToRemapId;
	delete[] remapToInitialId;
}

/***** PUBLIC FUNCTIONS *****/

int DatasetPreprocessor::getNumTransactions() const
{
	return transactionCount;
}

int DatasetPreprocessor::getSupportBufferSize() const
{
	return supportBufferSize;
}

int DatasetPreprocessor::getSupportCount(unsigned int item, unsigned int dimensionalAttribute) const
{
	static const int dimensionSize = supportBufferSize / (inputData.maxspan + 1);
	return supportCount[item + dimensionalAttribute * dimensionSize];
}

bool DatasetPreprocessor::isItemFrequent(unsigned int extendedItemIndex) const
{
	static const int minNumTransactions = getMinNumTransactions();
	int support = supportCount[extendedItemIndex];
	return support >= minNumTransactions;
}

/*
The minimum amount of transactions an item must be within in order to be frequent.
*/
int DatasetPreprocessor::getMinNumTransactions() const
{
	int minNumTransactions = (int)ceil(inputData.minSup * transactionCount);
	if (minNumTransactions == 0)
		minNumTransactions = 1; // The item must occur in at least 1 transaction.
	return minNumTransactions;
}

/*
Returns the amount of transactions in the given buffer for the thread.
*/
int DatasetPreprocessor::getNumTransactionsForThread(int threadNr) const
{
	return numTransactionsForThread[threadNr];
}

int DatasetPreprocessor::getNumTransactionsWithFrequentItems() const
{
	return numTransactionsWithFrequentItems;
}

int DatasetPreprocessor::getNumFrequentItems() const
{
	return numFrequentItems;
}

int* DatasetPreprocessor::getRemapToInitialIdArray() const
{
	return remapToInitialId;
}

int* DatasetPreprocessor::getInitialToRemapIdArray() const
{
	return initialToRemapId;
}

/***** PRIVATE FUNCTIONS *****/

/*
Counts the support for each individual item, the length of the longest transaction and
transaction count.
*/
void DatasetPreprocessor::countSupport()
{
	string startMessage = "Calculating support count buffer size and the length of the longest transaction.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();
	getMaxIdAndLongestTransaction(&supportBufferSize, &longestTransaction);
	supportBufferSize += 1; // +1 in order to support items with ID = 0

	supportBufferSize *= (inputData.maxspan + 1);

	string message2 = "Buffer size: " + std::to_string(supportBufferSize) + " * 4 B = " + std::to_string((int)ceil((supportBufferSize * 4) / (double)1024)) +
		 " kB. Longest transaction: " + std::to_string(longestTransaction) + ". Number of transactions: " + std::to_string(transactionCount) +
		 ". [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(message2);

	string message3 = "Counting support for each item.\n";
	writeToLog(message3);
	time = omp_get_wtime();

	supportCount = new int[supportBufferSize];
	for (int i = 0; i < supportBufferSize; i++)
		supportCount[i] = 0; // Initialize support counts to 0.

	int transactionNr = 0;
	int item = -1;
	for (int k = 0; k < inputData.numThreads; k++)
	{
		for (int i = 0; i < multiBuffers[k]->getNumBytes(); i++)
		{
			char byte = multiBuffers[k]->getByte(&i);

			if (isdigit(byte))
			{
				if (item == -1)
					item = 0;
				else
					item *= 10;

				item += byte - '0';
			}
			else
			{
				if (item != -1) // This is the case if the last byte was a space (in case multiple spaces occur after each other).
				{
					for (int dim = 0; dim < inputData.maxspan + 1; dim++)
					{
						if (transactionNr - dim < 0)
							break;

						int index = item + dim * (supportBufferSize / (inputData.maxspan + 1));
						supportCount[index]++;
					}
				}
				item = -1;

				if (byte == '\n')
				{
					transactionNr++;
				}
			}
		}
	}

	string endMessage = "Support counting finished. [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);
}

/*
Finds and returns the item with highest ID, and returns the length of the longest transaction.
This also calucates the transaction count.
*/
void DatasetPreprocessor::getMaxIdAndLongestTransaction(int* maxId, int* maxTransactionLength)
{
	int item = -1;
	int transactionLength = 0;
	*maxId = -1;
	*maxTransactionLength = -1;
	for (int k = 0; k < inputData.numThreads; k++)
	{
		for (int i = 0; i < multiBuffers[k]->getNumBytes(); i++)
		{
			char byte = multiBuffers[k]->getByte(&i);

			if (isdigit(byte))
			{
				if (item == -1)
					item = 0;
				else
					item *= 10;

				item += byte - '0';
			}
			else
			{
				if (item != -1) // This is the case if the last byte was a space (in case multiple spaces occur after each other).
				{
					if (item > *maxId)
						*maxId = item;
					transactionLength++;
				}
				item = -1;

				if (byte == '\n')
				{
					if (transactionLength > *maxTransactionLength)
						*maxTransactionLength = transactionLength;
					transactionLength = 0;
					transactionCount++;
				}
			}
		}
	}
}

void DatasetPreprocessor::countTransactionsWithFrequentItems()
{
	string startMessage = "Counting number of transactions with frequent items. \n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	int minNumTransactions = getMinNumTransactions();
	bool containsFrequentItem = false;
	int item = -1;
	for (int k = 0; k < inputData.numThreads; k++)
	{
		for (int i = 0; i < multiBuffers[k]->getNumBytes(); i++)
		{
			char byte = multiBuffers[k]->getByte(&i);

			if (isdigit(byte))
			{
				if (item == -1)
					item = 0;
				else
					item *= 10;

				item += byte - '0';
			}
			else
			{
				if (item != -1) // This is the case if the last byte was a space (in case multiple spaces occur after each other).
				{
					if (supportCount[item] >= minNumTransactions)
						containsFrequentItem = true;
				}
				item = -1;

				if (byte == '\n' && containsFrequentItem)
				{
					numTransactionsWithFrequentItems++;
					numTransactionsForThread[k]++;
					containsFrequentItem = false;
				}
			}
		}
	}

	string endMessage = "Transaction counting finished. [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);
}

void DatasetPreprocessor::countNumberOfFrequentItems()
{
	string startMessage = "Counting number of frequent items. \n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	for (int i = 0; i < supportBufferSize; i++)
	{
		if (isItemFrequent(i))
			numFrequentItems++;
	}

	string endMessage = "Number of frequent items: " + toString(numFrequentItems) + " [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);
}

/*
Identifiers are assign in decreasing order of item frequency.
They are stored in an item mapping array.
*/
void DatasetPreprocessor::remapItemsByFrequency()
{
	string startMessage = "Remapping items by frequency.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	remapToInitialId = new int[supportBufferSize];
	for (int i = 0; i < supportBufferSize; i++)
		remapToInitialId[i] = i;
	qsort(remapToInitialId, supportBufferSize, sizeof(int), compareItem);

	initialToRemapId = new int[supportBufferSize];
	for (int i = 0; i < supportBufferSize; i++)
		initialToRemapId[remapToInitialId[i]] = i;

	string endMessage = "Remapping finished. [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);
}

/***** FUNCTIONS OUTSIDE CLASS *****/

int compareItem(const void* itemIdA, const void* itemIdB)
{
	int value = DatasetPreprocessor::supportCount[*(int*)itemIdB] - DatasetPreprocessor::supportCount[*(int*)itemIdA];
	if (value != 0)
		return value;
	else // sort by ID
		return *(int*)itemIdB - *(int*)itemIdA;
}
