#include "EmbarrassinglyParallelMiner.h"
#include "BitsetUtility.h"

#include <omp.h>

EmbarrassinglyParallelMiner::EmbarrassinglyParallelMiner(StackMemoryManager** memoryManagers,
														 DatasetPreprocessor& datasetPreprocessor,
														 BitTable& bitTable,
														 IndexArray& indexArray,
														 InputData &inputData,
														 LogData& logData)
	: Miner(memoryManagers, datasetPreprocessor, bitTable, indexArray, inputData, logData)
{
}

EmbarrassinglyParallelMiner::~EmbarrassinglyParallelMiner()
{
}

void EmbarrassinglyParallelMiner::indexBTFITI()
{
	IndexArrayElement* elements = indexArray.getElements();

	#pragma omp parallel for schedule(dynamic, 1) num_threads(inputData.numThreads)
	for (int i = 0; i < indexArray.getNumElements(); i++)
	{
		indexBTFITIForItem(i, elements, omp_get_thread_num());
	}
}

void EmbarrassinglyParallelMiner::depthFirst(ExtendedItemset& itemset, ExtendedItemset& tail, int threadNr)
{
	if (tail.itemCount == 0)
		return;

	ExtendedItemset newItemset;
	newItemset.itemCount = itemset.itemCount + 1;
	long long sizeofItems = newItemset.itemCount * sizeof(unsigned int*);
	newItemset.items = (unsigned int*)memoryManagers[threadNr]->grow(sizeofItems);
	long long sizeofDats = newItemset.itemCount * sizeof(unsigned char*);
	newItemset.dimensionalAttributes = memoryManagers[threadNr]->grow(sizeofDats);
	std::copy(itemset.items, itemset.items + itemset.itemCount, newItemset.items);
	std::copy(itemset.dimensionalAttributes, itemset.dimensionalAttributes + itemset.itemCount, newItemset.dimensionalAttributes);

	ExtendedItemset newTail;
	newTail.itemCount = tail.itemCount - 1;
	long long sizeOfTailItems = newTail.itemCount * sizeof(unsigned int*);
	newTail.items = (unsigned int*)memoryManagers[threadNr]->grow(sizeOfTailItems);
	long long sizeOfTailDats = newTail.itemCount * sizeof(unsigned char*);
	newTail.dimensionalAttributes = memoryManagers[threadNr]->grow(sizeOfTailDats);

	for (int i = 0; i < tail.itemCount; i++)
	{
		depthFirstForItem(i, itemset, tail, newItemset, newTail, threadNr);
	}

	memoryManagers[threadNr]->shrink(sizeofItems + sizeofDats + sizeOfTailItems + sizeOfTailDats);
}

void EmbarrassinglyParallelMiner::addAllSubsumeSets(unsigned int* items,
							  unsigned int itemCount,
							  unsigned int* subsume,
							  unsigned int subsumeCount,
							  unsigned char* dimensionalAttributes,
							  double support,
							  int threadNr)
{
	ExtendedItemset output;
	output.itemCount = itemCount + subsumeCount;
	output.items = (unsigned int*)memoryManagers[threadNr]->grow(output.itemCount * sizeof(unsigned int));
	output.dimensionalAttributes = memoryManagers[threadNr]->grow(output.itemCount * sizeof(unsigned char));
	std::copy(items, items + itemCount, output.items);
	std::copy(dimensionalAttributes, dimensionalAttributes + itemCount, output.dimensionalAttributes);

	// Power set using counter
	unsigned int power = 1 << (unsigned int)subsumeCount;
	for (unsigned int i = 1; i < power; i++)
	{
		unsigned int bitCounter = i;

		int outputIndex = 0;
		for (int j = 0; j < subsumeCount; j++)
		{
			if (bitCounter & 1)
			{
				output.items[itemCount + outputIndex] = subsume[j];
				output.dimensionalAttributes[itemCount + outputIndex] = dimensionalAttributes[j + itemCount];
				outputIndex++;
			}
			bitCounter >>= 1;
		}
		output.itemCount = outputIndex + itemCount;

		if (support < inputData.minSup)
			continue;

		addToFrequentItemsets(output, support, threadNr);
	}

	memoryManagers[threadNr]->shrink(itemCount + subsumeCount);
}

void EmbarrassinglyParallelMiner::combineItemsSubsumeAndTail(unsigned int* items,
									   unsigned int itemCount,
									   unsigned int* subsume,
									   unsigned int subsumeCount,
									   unsigned char* dimensionalAttributes,
									   ExtendedItemset& tail,
									   int threadNr)
{
	ExtendedItemset itemset;
	itemset.itemCount = itemCount + subsumeCount;
	itemset.items = (unsigned int*)memoryManagers[threadNr]->grow(itemset.itemCount * sizeof(unsigned int));
	itemset.dimensionalAttributes = memoryManagers[threadNr]->grow(itemset.itemCount * sizeof(unsigned char));
	std::copy(items, items + itemCount, itemset.items);
	std::copy(dimensionalAttributes, dimensionalAttributes + itemCount, itemset.dimensionalAttributes);

	// Power set using counter
	unsigned int power = 1 << (unsigned int)subsumeCount;
	for (unsigned int i = 1; i < power; i++)
	{
		unsigned int bitCounter = i;

		int itemsetIndex = 0;
		for (int j = 0; j < subsumeCount; j++)
		{
			if (bitCounter & 1)
			{
				itemset.items[itemCount + itemsetIndex] = subsume[j];
				itemset.dimensionalAttributes[itemCount + itemsetIndex] = dimensionalAttributes[j + itemCount];
				itemsetIndex++;
			}
			bitCounter >>= 1;
		}
		itemset.itemCount = itemsetIndex + itemCount;

		depthFirst(itemset, tail, threadNr);
	}

	memoryManagers[threadNr]->shrink(itemCount + subsumeCount);
}
