#include "IndexArray.h"
#include "BitsetUtility.h"
#include "../FimLogger/Logger.h"

#include <omp.h>

IndexArray::IndexArray(StackMemoryManager** memoryManagers,
					   BitTable& bitTable,
					   InputData &inputData,
					   DatasetPreprocessor& datasetPreprocessor,
					   LogData& logData)
	: memoryManagers(memoryManagers),
	  bitTable(bitTable),
	  inputData(inputData),
	  datasetPreprocessor(datasetPreprocessor),
	  logData(logData)
{
	createIndexArray();
}

IndexArray::~IndexArray()
{
}

/***** PUBLIC FUNCTIONS *****/

unsigned long long IndexArray::getMemoryUsageInBytes() const
{
	unsigned long long totalMemoryUsage = 0;
	for (int i = 0; i < inputData.numThreads; i++)
		totalMemoryUsage += memoryManagers[i]->getMemoryUsageInBytes();
	return totalMemoryUsage;
}

int IndexArray::getNumElements() const
{
	return datasetPreprocessor.getNumFrequentItems();
}

IndexArrayElement* IndexArray::getElements() const
{
	return elements;
}

/***** PRIVATE FUNCTIONS *****/

void IndexArray::createIndexArray()
{
	string startMessage = "Creating IndexArray.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();
	double oldMemoryUsageInMB = getMemoryUsageInBytes() / (double)(1024.0 * 1024.0);

	int numFrequentItems = datasetPreprocessor.getNumFrequentItems();
	elements = (IndexArrayElement*)memoryManagers[0]->grow(sizeof(IndexArrayElement) * numFrequentItems);

	static const int numBytes = bitTable.getNumBytesPerRowForTransactionTable();
	unsigned char** subsumeBitVectors = new unsigned char*[inputData.numThreads];
	for (int i = 0; i < inputData.numThreads; i++)
		subsumeBitVectors[i] = new unsigned char[numBytes];

	#pragma omp parallel for schedule(dynamic, 1) num_threads(inputData.numThreads)
	for (int i = 0; i < datasetPreprocessor.getNumFrequentItems(); i++)
	{
		int threadNr = omp_get_thread_num();
		int transactionNr = 0;
		bool isSubsumeSet = false;

		elements[i].item = i;

		while (transactionNr < bitTable.getNumRowsForTransactionTable()) // for each transaction
		{
			if (bitTable.isItemInTransaction(i, transactionNr))
			{
				unsigned char* transactionBitVector = bitTable.getTransactionBitVector(transactionNr);
				if (isSubsumeSet)
					intersectBitsets(subsumeBitVectors[threadNr], transactionBitVector, numBytes, subsumeBitVectors[threadNr]);
				else
				{
					std::copy(transactionBitVector, transactionBitVector + numBytes, subsumeBitVectors[threadNr]);
					isSubsumeSet = true;
				}
			}

			transactionNr++;
		}

		elements[i].itemCount = popcount(subsumeBitVectors[threadNr], bitTable.getNumColumnsForTransactionTable()) - 1;
		elements[i].subsume = (unsigned int*)memoryManagers[threadNr]->grow(elements[i].itemCount * sizeof(unsigned int));
		bitTable.getItemsFromSubsumbeBitVector(subsumeBitVectors[threadNr], elements[i].subsume, elements[i].item);
	}

	for (int i = 0; i < inputData.numThreads; i++)
		delete[] subsumeBitVectors[i];
	delete[] subsumeBitVectors;

	logData.preprocessTime += omp_get_wtime() - time;
	string endMessage = "IndexArray created. [" + toString(omp_get_wtime() - time) + "s]\n";
	writeToLog(endMessage);
	double newMemoryUsageInMB = getMemoryUsageInBytes() / (double)(1024.0 * 1024.0);
	logData.memoryUsageInMB = newMemoryUsageInMB;
	writeToLog("IndexArray memory usage: " + toString(newMemoryUsageInMB - oldMemoryUsageInMB) + "MB\n\n");
}
