#include "Miner.h"
#include "BitsetUtility.h"
#include "../FimLogger/Logger.h"

#include <omp.h>
#include <fstream>
using std::ofstream;
#include <cmath>
#include <cstring>
#include <iostream>
using std::cout;
using std::endl;

Miner::Miner(StackMemoryManager** memoryManagers,
			 DatasetPreprocessor& datasetPreprocessor,
			 BitTable& bitTable,
			 IndexArray& indexArray,
			 InputData &inputData,
			 LogData& logData)
	: memoryManagers(memoryManagers),
	  datasetPreprocessor(datasetPreprocessor),
	  bitTable(bitTable),
	  indexArray(indexArray),
	  inputData(inputData),
	  logData(logData)
{
	extendedItemsetStacks = new ExtendedItemsetStack*[inputData.numThreads];
	for (int i = 0; i < inputData.numThreads; i++)
		extendedItemsetStacks[i] = new ExtendedItemsetStack;
}

Miner::~Miner()
{
	for (int i = 0; i < inputData.numThreads; i++)
		delete extendedItemsetStacks[i];
	delete[] extendedItemsetStacks;
}

/***** PUBLIC FUNCTIONS *****/

void Miner::startMining()
{
	string startMessage = "Mining procedure started.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	indexBTFITI();

	double totalTime = omp_get_wtime() - time;
	logData.mineTime = totalTime;
	string endMessage = "Mining procedure finished. [" + toString(totalTime) + "s]\n\n";
	writeToLog(endMessage);
}

void Miner::writeResults(string path)
{
	reportNumberOfFrequentItemsetsFound();
	string startMessage = "Writing file to: " + path + "\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	ofstream file;
	file.open(path);

	for (int k = 0; k < inputData.numThreads; k++)
	{
		for (int i = (int)extendedItemsetStacks[k]->size() - 1; i >= 0; i--)
		{
			unsigned int* itemCount;
			double* support;
			unsigned char* dimensionalAttributes;
			unsigned int* items;
			extendedItemsetStacks[k]->top(&itemCount, &support, &dimensionalAttributes, &items);
			for (unsigned int j = 0; j < *itemCount; j++)
				file << items[j] << ":" << (unsigned short)dimensionalAttributes[j] << " ";
			file << "(" << *support << ")" << std::endl;
			extendedItemsetStacks[k]->pop();
		}
	}

	file.close();

	string endMessage = "Finished writing. [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);

	std::ifstream in(path, std::ios::binary | std::ios::ate);
	writeToLog("Created file size: " + std::to_string((int)ceil(in.tellg() / 1024.0)) + "kB\n\n");
}

void Miner::reportNumberOfFrequentItemsetsFound()
{
	int numFrequentItemsets = 0;
	for (int i = 0; i < inputData.numThreads; i++)
		numFrequentItemsets += (int)extendedItemsetStacks[i]->size();
	writeToLog("Number of frequent itemsets found: " + std::to_string(numFrequentItemsets) + "\n");
}

/***** PROTECTED FUNCTIONS *****/

void Miner::indexBTFITIForItem(int i, IndexArrayElement* elements, int threadNr)
{
	long long sizeOfItemAndSubsume = (elements[i].itemCount + 1) * sizeof(unsigned int);
	long long sizeOfDimensionalAttributes = (elements[i].itemCount + 1) * sizeof(unsigned char);
	unsigned int* itemAndSubsume = (unsigned int*)memoryManagers[threadNr]->grow(sizeOfItemAndSubsume);
	unsigned char* dats = memoryManagers[threadNr]->grow(sizeOfDimensionalAttributes);

	bitTable.itemIdToExtendedItem(elements[i].item, itemAndSubsume, dats);
	for (int j = 0; j < elements[i].itemCount; j++)
		bitTable.itemIdToExtendedItem(elements[i].subsume[j], &(itemAndSubsume[j + 1]), &(dats[j + 1]));

	if (dats[0] == 0)
	{
		indexBitTableFI(itemAndSubsume, 1, &(itemAndSubsume[1]), elements[i].itemCount, dats, threadNr);
	}
	else
	{
		long long sizeOfNewItemsAndSubsume = (elements[i].itemCount + 1) * sizeof(unsigned int);
		long long sizeOfNewDimensionalAttributes = (elements[i].itemCount + 1) * sizeof(unsigned char);
		unsigned int* newItemsAndSubsume = (unsigned int*)memoryManagers[threadNr]->grow(sizeOfNewItemsAndSubsume);
		unsigned char* newDats = memoryManagers[threadNr]->grow(sizeOfNewDimensionalAttributes);
		newItemsAndSubsume[0] = itemAndSubsume[0];
		newDats[0] = dats[0];

		for (int j = 1; j < elements[i].itemCount + 1; j++) // iterate through all items in subsume
		{
			if (dats[j] == 0)
			{
				newItemsAndSubsume[1] = itemAndSubsume[j];
				newDats[1] = dats[j];
				for (int k = 2; k < elements[i].itemCount + 1; k++)
				{
					if (k > j)
					{
						newItemsAndSubsume[k] = itemAndSubsume[k];
						newDats[k] = dats[k];
					}
					else
					{
						newItemsAndSubsume[k] = itemAndSubsume[k - 1];
						newDats[k] = dats[k - 1];
					}
				}

				indexBitTableFI(newItemsAndSubsume, 2, &(newItemsAndSubsume[2]), elements[i].itemCount - 1, newDats, threadNr);
			}
		}

		memoryManagers[threadNr]->shrink(sizeOfNewItemsAndSubsume + sizeOfNewDimensionalAttributes);
	}

	memoryManagers[threadNr]->shrink(sizeOfItemAndSubsume + sizeOfDimensionalAttributes);
}

void Miner::indexBitTableFI(unsigned int* items,
							unsigned int itemCount,
							unsigned int* subsume,
							unsigned int subsumeCount,
							unsigned char* dimensionalAttributes,
							int threadNr)
{
	ExtendedItemset itemset { itemCount, items, dimensionalAttributes };
	double support = getSupport(itemset, threadNr);
	addToFrequentItemsets(itemset, support, threadNr);

	ExtendedItemset tail;
	unsigned int itemId = getHighestItemIdWithEqualOrGreaterSupport(items, dimensionalAttributes, itemCount, threadNr);
	tail.itemCount = itemId;
	long long sizeOfTailItems = tail.itemCount * sizeof(unsigned int);
	long long sizeOfTailDats = tail.itemCount * sizeof(unsigned char);
	tail.items = (unsigned int*)memoryManagers[threadNr]->grow(sizeOfTailItems);
	tail.dimensionalAttributes = memoryManagers[threadNr]->grow(sizeOfTailDats);

	if (subsumeCount == 0)
	{
		int it = 0;
		for (int i = 0; i < tail.itemCount; i++)
		{
			bitTable.itemIdToExtendedItem(it, &tail.items[i], &tail.dimensionalAttributes[i]);

			// Don't add this item to tail if it is contained in items
			for (int j = 0; j < itemCount; j++)
			{
				if (tail.items[i] == items[j] && tail.dimensionalAttributes[i] == dimensionalAttributes[j])
				{
					tail.itemCount--;
					i--;
					break;
				}
			}

			it++;
		}

		depthFirst(itemset, tail, threadNr);

		memoryManagers[threadNr]->shrink(sizeOfTailItems + sizeOfTailDats);
		return;
	}

	addAllSubsumeSets(items, itemCount, subsume, subsumeCount, dimensionalAttributes, support, threadNr);

	// Create tail with items more frequent than itemId, and that doesn't occur in subsume or in items.
	int it = 0;
	for (int i = 0; i < tail.itemCount; i++)
	{
		bool continueLoop = false;
		bitTable.itemIdToExtendedItem(it, &tail.items[i], &tail.dimensionalAttributes[i]);

		// Don't add this item to tail if it is contained in items
		for (int j = 0; j < itemCount; j++)
		{
			if (tail.items[i] == items[j] && tail.dimensionalAttributes[i] == dimensionalAttributes[j])
			{
				continueLoop = true;
				tail.itemCount--;
				i--;
				break;
			}
		}
		if (continueLoop)
		{
			it++;
			continue;
		}

		// Don't add this item to tail if it is contained in subsume
		for (int j = 0; j < subsumeCount; j++)
		{
			if (tail.items[i] == subsume[j] && tail.dimensionalAttributes[i] == dimensionalAttributes[itemCount + j])
			{
				tail.itemCount--;
				i--;
				break;
			}
		}

		it++;
	}
	depthFirst(itemset, tail, threadNr);

	if (tail.itemCount != 0)
		combineItemsSubsumeAndTail(items, itemCount, subsume, subsumeCount, dimensionalAttributes, tail, threadNr);

	memoryManagers[threadNr]->shrink(sizeOfTailItems + sizeOfTailDats);
}

void Miner::depthFirstForItem(int i,
							  ExtendedItemset& itemset,
							  ExtendedItemset& tail,
							  ExtendedItemset& newItemset,
							  ExtendedItemset& newTail,
							  int threadNr)
{
	newItemset.items[itemset.itemCount] = tail.items[i];
	newItemset.dimensionalAttributes[itemset.itemCount] = tail.dimensionalAttributes[i];
	double support = getSupport(newItemset, threadNr);
	if (support >= inputData.minSup)
	{
		addToFrequentItemsets(newItemset, support, threadNr);

		std::copy(&tail.items[i + 1], &tail.items[i + 1] + tail.itemCount - i, newTail.items);
		std::copy(&tail.dimensionalAttributes[i + 1], &tail.dimensionalAttributes[i + 1] + tail.itemCount - i, newTail.dimensionalAttributes);
		newTail.itemCount = tail.itemCount - i - 1;

		depthFirst(newItemset, newTail, threadNr);
	}
}

void Miner::addToFrequentItemsets(ExtendedItemset& extendedItemset, double support, int threadNr)
{
	unsigned int* itemsOut;
	unsigned char* datsOut;
	extendedItemsetStacks[threadNr]->push(extendedItemset.itemCount, support * 100, &datsOut, &itemsOut);
	std::copy(extendedItemset.items, extendedItemset.items + extendedItemset.itemCount, itemsOut);
	std::copy(extendedItemset.dimensionalAttributes, extendedItemset.dimensionalAttributes + extendedItemset.itemCount, datsOut);
}

double Miner::getSupport(ExtendedItemset& extendedItemset, int threadNr)
{
	return getSupportCount(extendedItemset, threadNr) / (double)datasetPreprocessor.getNumTransactions();
}

unsigned int Miner::getSupportCount(ExtendedItemset& extendedItemset, int threadNr)
{
	int numBytes = bitTable.getNumBytesPerRowForItemTable();
	unsigned char* itemBitVector = memoryManagers[threadNr]->grow(numBytes);
	for (int i = 0; i < extendedItemset.itemCount; i++)
	{
		unsigned int itemId;
		bitTable.extendedItemToItemId(extendedItemset.items[i], extendedItemset.dimensionalAttributes[i], &itemId);
		unsigned char* currentItemBitVector = bitTable.getItemBitVector(itemId);

		if (i == 0)
			std::copy(currentItemBitVector, currentItemBitVector + numBytes, itemBitVector);
		else
			intersectBitsets(currentItemBitVector, itemBitVector, numBytes, itemBitVector);
	}
	int supportCount = popcount(itemBitVector, bitTable.getNumColumnsForItemTable());
	memoryManagers[threadNr]->shrink(numBytes);

	return supportCount;
}


unsigned int Miner::getHighestItemIdWithEqualOrGreaterSupport(unsigned int* items,
															  unsigned char* dats,
															  unsigned int itemCount,
															  int threadNr)
{
	unsigned int itemId;

	if (itemCount == 1)
	{
		bitTable.extendedItemToItemId(items[0], dats[0], &itemId);
	}
	else
	{
		ExtendedItemset itemset { itemCount, items, dats };
		double supportCount = getSupportCount(itemset, threadNr);

		// TODO: implement binary search
		for (int i = 0; i < datasetPreprocessor.getSupportBufferSize(); i++)
		{
			int index = datasetPreprocessor.getRemapToInitialIdArray()[i];

			if (datasetPreprocessor.supportCount[index] < supportCount)
			{
				itemId = i - 1;
				break;
			}
		}
	}

	return itemId;
}
