#include "DatasetPreprocessor.h"
#include "../FimLogger/Logger.h"

#include <omp.h>
#include <cmath>

int* DatasetPreprocessor::supportCount = NULL;

DatasetPreprocessor::DatasetPreprocessor(MultiBuffer **multiBuffers, double minSup, int numThreads)
{
	this->multiBuffers = multiBuffers;
	this->minSup = minSup;
	this->numThreads = numThreads;
	numTransactionsForThread = new int[numThreads];
	for (int i = 0; i < numThreads; i++)
		numTransactionsForThread[i] = 0;
	countSupport();
	countTransactionsWithFrequentItems();
}

DatasetPreprocessor::~DatasetPreprocessor()
{
	delete[] supportCount;
	delete[] numTransactionsForThread;
}

/***** PUBLIC FUNCTIONS *****/

int DatasetPreprocessor::getSupportBufferSize()
{
	return supportBufferSize;
}

/*
The minimum amount of transactions an item must be within in order to be frequent.
*/
int DatasetPreprocessor::getMinNumTransactions()
{
	int minNumTransactions = (int)ceil(minSup * transactionCount);
	if (minNumTransactions == 0)
		minNumTransactions = 1; // The item must occur in at least 1 transaction.
	return minNumTransactions;
}

/*
Returns the number of items that have greater or equal support to minSup.
*/
int DatasetPreprocessor::getNumItems()
{
	int minNumTransactions = getMinNumTransactions();

	string startMessage = "Calculating number of items in dataset that occur in at least " +
		std::to_string(minNumTransactions) + " transaction(s).\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	int itemCount = 0;
	for (int i = 0; i < supportBufferSize; i++)
	{
		if (supportCount[i] >= minNumTransactions)
			itemCount++;
	}

	string endMessage = "Item count: " + std::to_string(itemCount) +
		". [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);

	return itemCount;
}

int DatasetPreprocessor::getTransactionCount()
{
	return transactionCount;
}

int DatasetPreprocessor::getLongestTransaction()
{
	return longestTransaction;
}

double DatasetPreprocessor::getMinSup()
{
	return minSup;
}

int DatasetPreprocessor::getNumThreads()
{
	return numThreads;
}

/*
Returns the amount of transactions in the given buffer for the thread.
*/
int DatasetPreprocessor::getNumTransactionsForThread(int threadNr)
{
	return numTransactionsForThread[threadNr];
}

int DatasetPreprocessor::getNumTransactionsWithFrequentItems()
{
	return numTransactionsWithFrequentItems;
}

/***** PRIVATE FUNCTIONS *****/

/*
Counts the support for each individual item, the length of the longest transaction and
transaction count.
*/
void DatasetPreprocessor::countSupport()
{
	string startMessage = "Calculating support count buffer size and the length of the longest transaction.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();
	getMaxIdAndLongestTransaction(&supportBufferSize, &longestTransaction);
	supportBufferSize += 1; // +1 in order to support items with ID = 0
	supportBufferSize += 1; // +1 again so smallest remap ID will be 1 and not 0.

	string message2 = "Buffer size: " + std::to_string(supportBufferSize) + " * 4 B = " + std::to_string((int)ceil((supportBufferSize * 4) / (double)1024)) +
		 " kB. Longest transaction: " + std::to_string(longestTransaction) + ". Number of transactions: " + std::to_string(transactionCount) +
		 ". [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(message2);

	string message3 = "Counting support for each item.\n";
	writeToLog(message3);
	time = omp_get_wtime();

	supportCount = new int[supportBufferSize];
	for (int i = 0; i < supportBufferSize; i++)
		supportCount[i] = 0; // Initialize support counts to 0.

	int item = -1;
	for (int k = 0; k < numThreads; k++)
	{
		for (int i = 0; i < multiBuffers[k]->getNumBytes(); i++)
		{
			char byte = multiBuffers[k]->getByte(&i);

			if (isdigit(byte))
			{
				if (item == -1)
					item = 0;
				else
					item *= 10;

				item += byte - '0';
			}
			else
			{
				if (item != -1) // This is the case if the last byte was a space (in case multiple spaces occur after each other).
				{
					supportCount[item]++;
				}
				item = -1;
			}
		}
	}

	string endMessage = "Support counting finished. [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);
}

/*
Finds and returns the item with highest ID, and returns the length of the longest transaction.
This also calucates the transaction count.
*/
void DatasetPreprocessor::getMaxIdAndLongestTransaction(int* maxId, int* maxTransactionLength)
{
	int item = -1;
	int transactionLength = 0;
	*maxId = -1;
	*maxTransactionLength = -1;
	for (int k = 0; k < numThreads; k++)
	{
		for (int i = 0; i < multiBuffers[k]->getNumBytes(); i++)
		{
			char byte = multiBuffers[k]->getByte(&i);

			if (isdigit(byte))
			{
				if (item == -1)
					item = 0;
				else
					item *= 10;

				item += byte - '0';
			}
			else
			{
				if (item != -1) // This is the case if the last byte was a space (in case multiple spaces occur after each other).
				{
					if (item > *maxId)
						*maxId = item;
					transactionLength++;
				}
				item = -1;

				if (byte == '\n')
				{
					if (transactionLength > *maxTransactionLength)
						*maxTransactionLength = transactionLength;
					transactionLength = 0;
					transactionCount++;
				}
			}
		}
	}
}

void DatasetPreprocessor::countTransactionsWithFrequentItems()
{
	string startMessage = "Counting number of transactions iwth frequent items. \n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	int minNumTransactions = getMinNumTransactions();
	bool containsFrequentItem = false;
	int item = -1;
	for (int k = 0; k < numThreads; k++)
	{
		for (int i = 0; i < multiBuffers[k]->getNumBytes(); i++)
		{
			char byte = multiBuffers[k]->getByte(&i);

			if (isdigit(byte))
			{
				if (item == -1)
					item = 0;
				else
					item *= 10;

				item += byte - '0';
			}
			else
			{
				if (item != -1) // This is the case if the last byte was a space (in case multiple spaces occur after each other).
				{
					if (supportCount[item] >= minNumTransactions)
						containsFrequentItem = true;
				}
				item = -1;

				if (byte == '\n' && containsFrequentItem)
				{
					numTransactionsWithFrequentItems++;
					numTransactionsForThread[k]++;
					containsFrequentItem = false;
				}
			}
		}
	}

	string endMessage = "Transaction counting finished. [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);
}

/***** FUNCTIONS OUTSIDE CLASS *****/

int compareItem(const void* itemIdA, const void* itemIdB)
{
	int value = DatasetPreprocessor::supportCount[*(int*)itemIdB] - DatasetPreprocessor::supportCount[*(int*)itemIdA];
	if (value != 0)
		return -value;
	else // sort by ID
		return *(int*)itemIdA - *(int*)itemIdB;
}
