#include "EclatMiner.h"
#include "BitsetUtility.h"
#include "../FimLogger/Logger.h"

#include <fstream>
#include <omp.h>
#include <cmath>

EclatMiner::EclatMiner(VerticalDataLayout& verticalDataLayout,
					   DatasetPreprocessor& datasetPreprocessor,
					   LogData& logData)
	: verticalDataLayout(verticalDataLayout),
	  datasetPreprocessor(datasetPreprocessor),
	  logData(logData)
{
	numThreads = datasetPreprocessor.getNumThreads();
	itemsetStack = new ItemsetStack*[numThreads];
	equivalenceClassStack = new EquivalenceClassStack*[numThreads];
	for (int i = 0; i < numThreads; i++)
	{
		itemsetStack[i] = new ItemsetStack();
		equivalenceClassStack[i] = new EquivalenceClassStack(
			getNumBytes(verticalDataLayout.getNumBitsPerBitmap())
		);
	}
}

EclatMiner::~EclatMiner()
{
	for (int i = 0; i < numThreads; i++)
	{
		delete itemsetStack[i];
		delete equivalenceClassStack[i];
	}
	delete[] itemsetStack;
	delete[] equivalenceClassStack;
}

/***** PUBLIC FUNCTIONS *****/

void EclatMiner::writeResults(string path) const
{
	reportNumberOfFrequentItemsetsFound();
	string startMessage = "Writing file to: " + path + "\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	std::ofstream file;
	file.open(path.c_str());

	for (int k = 0; k < numThreads; k++)
	{
		for (int i = (int)itemsetStack[k]->size() - 1; i >= 0; i--)
		{
			unsigned int* itemCount;
			double* support;
			unsigned int* items;
			itemsetStack[k]->top(&itemCount, &support, &items);
			for (unsigned int j = 0; j < *itemCount; j++)
				file << items[j] << " ";
			file << "(" << *support * 100 << ")" << std::endl;
			itemsetStack[k]->pop();
		}
	}

	file.close();

	string endMessage = "Finished writing. [" + toString(omp_get_wtime() - time) + "s]\n";
	writeToLog(endMessage);

	std::ifstream in(path.c_str(), std::ios::binary | std::ios::ate);
	writeToLog("Created file size: " + toString((int)ceil(in.tellg() / 1024.0)) + "kB\n\n");
}

void EclatMiner::reportNumberOfFrequentItemsetsFound() const
{
	int numFrequentItemsets = 0;
	for (int i = 0; i < numThreads; i++)
		numFrequentItemsets += (int)itemsetStack[i]->size();
	writeToLog("Number of frequent itemsets found: " + toString(numFrequentItemsets) + "\n");
}

/***** PROTECTED FUNCTIONS *****/

void EclatMiner::addToFrequentItemsets(int itemCount, unsigned int* items, double support)
{
	int threadNr = omp_get_thread_num();
	unsigned int* itemAddress = itemsetStack[threadNr]->push(itemCount, support);
	for (int i = 0; i < itemCount; i++)
		itemAddress[i] = items[i];
}
