#include "HybridClassMiner.h"
#include "BitsetUtility.h"
#include "../FimLogger/Logger.h"
#include "Utility.h"

#include <omp.h>
#include <cmath>
#include <algorithm>

HybridClassMiner::HybridClassMiner(VerticalDataLayout& verticalDataLayout,
								   DatasetPreprocessor& datasetPreprocessor,
								   LogData &logData)
	: EclatMiner(verticalDataLayout, datasetPreprocessor, logData)
{
	omp_init_lock(&masterLock);
}

HybridClassMiner::~HybridClassMiner()
{
	omp_destroy_lock(&masterLock);
}

/***** PUBLIC FUNCTIONS *****/

void HybridClassMiner::startMining()
{
	string startMessage = "Started hybrid class mining.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	vector<unsigned int>* frequentItems = verticalDataLayout.getFrequentItems();
	omp_set_nested(1);

	#pragma omp parallel num_threads(numThreads)
	{
		int threadNr = omp_get_thread_num();

		int start = threadNr * frequentItems->size() / numThreads;
		int end = threadNr - 1 == numThreads ? frequentItems->size() :
				  (threadNr + 1) * frequentItems->size() / numThreads;

		for (size_t i = start; i < end; i++)
		{
			unsigned int item = frequentItems->at(i);

			double support = calculateSupport(
				datasetPreprocessor.supportCount[item],
				datasetPreprocessor.getTransactionCount()
			);
			addToFrequentItemsets(1, &item, support);

			equivalenceClassStack[threadNr]->push(1, &item, verticalDataLayout.getTidBitmaps()[item]);
			mine(i + 1, threadNr);
			equivalenceClassStack[threadNr]->pop();
		}

		addFreeThread(threadNr);
	}

	double elapsedTime = omp_get_wtime() - time;
	string endMessage = "Mining finished. [" + toString(elapsedTime) + "s]\n\n";
	writeToLog(endMessage);
	logData.mineTime = elapsedTime;
}

/***** PRIVATE FUNCTIONS *****/

void HybridClassMiner::mine(int nextIndex, int threadNr)
{
	static const int MIN_NUM_TRANSACTIONS = datasetPreprocessor.getMinNumTransactions();

	unsigned int* itemCount;
	unsigned int* items;
	unsigned char* tidBitmap;
	equivalenceClassStack[threadNr]->top(&itemCount, &items, &tidBitmap);

	unsigned long long numBits = verticalDataLayout.getNumBitsPerBitmap();
	unsigned long long numBytes = getNumBytes(numBits);
	vector<unsigned int>* frequentItems = verticalDataLayout.getFrequentItems();

	int threads = 1;
	omp_set_lock(&masterLock);
	threads += freeThreads.size();
	int* availableThreads = NULL;
	if (threads > 1)
	{
		availableThreads = new int[threads];
		availableThreads[0] = threadNr;
		for (size_t i = 0; i < freeThreads.size(); i++)
			availableThreads[i + 1] = freeThreads[i];
		freeThreads.clear();
	}
	omp_unset_lock(&masterLock);

	// TODO: refactor
	if (threads > 1)
	{
		#pragma omp parallel for schedule(dynamic, 1) num_threads(threads)
		for (int i = nextIndex; i < frequentItems->size(); i++)
		{
			int nestedThreadNr = availableThreads[omp_get_thread_num()];
			unsigned int item = frequentItems->at(i);
			unsigned int* newItems;
			unsigned char* newBitmap;
			equivalenceClassStack[nestedThreadNr]->pushEmpty(*itemCount + 1, &newItems, &newBitmap);

			intersectBitsets(
				tidBitmap,
				verticalDataLayout.getTidBitmaps()[item],
				numBytes,
				newBitmap
			);

			unsigned long long newSupportCount = popcount(newBitmap, numBits);
			if (newSupportCount >= MIN_NUM_TRANSACTIONS)
			{
				arrayCopy(items, newItems, *itemCount);
				newItems[*itemCount] = item;

				unsigned int* fitems = itemsetStack[nestedThreadNr]->push(
					*itemCount + 1,
					calculateSupport(newSupportCount, verticalDataLayout.getNumBitsPerBitmap())
				);
				arrayCopy(newItems, fitems, *itemCount + 1);

				if (i + 1 < frequentItems->size())
					mine(i + 1, nestedThreadNr);
			}
			equivalenceClassStack[nestedThreadNr]->pop();
		}

		omp_set_lock(&masterLock);
		for (int i = 1; i < threads; i++)
			freeThreads.push_back(availableThreads[i]);
		delete[] availableThreads;
		omp_unset_lock(&masterLock);
	}
	else
	{
		for (int i = nextIndex; i < frequentItems->size(); i++)
		{
			unsigned int item = frequentItems->at(i);
			unsigned int* newItems;
			unsigned char* newBitmap;
			equivalenceClassStack[threadNr]->pushEmpty(*itemCount + 1, &newItems, &newBitmap);

			intersectBitsets(
				tidBitmap,
				verticalDataLayout.getTidBitmaps()[item],
				numBytes,
				newBitmap
			);

			unsigned long long newSupportCount = popcount(newBitmap, numBits);
			if (newSupportCount >= MIN_NUM_TRANSACTIONS)
			{
				arrayCopy(items, newItems, *itemCount);
				newItems[*itemCount] = item;

				unsigned int* fitems = itemsetStack[threadNr]->push(
					*itemCount + 1,
					calculateSupport(newSupportCount, verticalDataLayout.getNumBitsPerBitmap())
				);
				arrayCopy(newItems, fitems, *itemCount + 1);

				if (i + 1 < frequentItems->size())
					mine(i + 1, threadNr);
			}
			equivalenceClassStack[threadNr]->pop();
		}
	}
}

void HybridClassMiner::addFreeThread(int threadNr)
{
	omp_set_lock(&masterLock);
	freeThreads.push_back(threadNr);
	omp_unset_lock(&masterLock);
}

void HybridClassMiner::removeFreeThread(int threadNr)
{
	omp_set_lock(&masterLock);
	freeThreads.erase(
		std::remove(freeThreads.begin(), freeThreads.end(), threadNr), freeThreads.end()
	);
	omp_unset_lock(&masterLock);
}
