#include "IndependentClassMiner.h"
#include "BitsetUtility.h"
#include "../FimLogger/Logger.h"
#include "Utility.h"

#include <omp.h>
#include <fstream>
#include <cmath>

IndependentClassMiner::IndependentClassMiner(VerticalDataLayout& verticalDataLayout,
											 DatasetPreprocessor& datasetPreprocessor,
											 LogData& logData)
	: EclatMiner(verticalDataLayout, datasetPreprocessor, logData)
{
}

/***** PUBLIC FUNCTIONS *****/

void IndependentClassMiner::startMining()
{
	string startMessage = "Started independent class mining.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	vector<unsigned int>* frequentItems = verticalDataLayout.getFrequentItems();

	#pragma omp parallel for schedule(dynamic, 1) num_threads(numThreads)
	for (size_t i = 0; i < frequentItems->size(); i++)
	{
		unsigned int item = frequentItems->at(i);

		double support = calculateSupport(
			datasetPreprocessor.supportCount[item],
			datasetPreprocessor.getTransactionCount()
		);
		addToFrequentItemsets(1, &item, support);

		int threadNr = omp_get_thread_num();
		equivalenceClassStack[threadNr]->push(1, &item, verticalDataLayout.getTidBitmaps()[item]);
		mine(i + 1);
		equivalenceClassStack[threadNr]->pop();
	}

	double elapsedTime = omp_get_wtime() - time;
	string endMessage = "Mining finished. [" + toString(elapsedTime) + "s]\n\n";
	writeToLog(endMessage);
	logData.mineTime = elapsedTime;
}

/***** PRIVATE FUNCTIONS *****/

void IndependentClassMiner::mine(int nextIndex)
{
	static const int MIN_NUM_TRANSACTIONS = datasetPreprocessor.getMinNumTransactions();

	unsigned int* itemCount;
	unsigned int* items;
	unsigned char* tidBitmap;
	equivalenceClassStack[omp_get_thread_num()]->top(&itemCount, &items, &tidBitmap);

	unsigned long long numBits = verticalDataLayout.getNumBitsPerBitmap();
	unsigned long long numBytes = getNumBytes(numBits);
	vector<unsigned int>* frequentItems = verticalDataLayout.getFrequentItems();
	for (int i = nextIndex; i < frequentItems->size(); i++)
	{
		int threadNr = omp_get_thread_num();
		unsigned int item = frequentItems->at(i);
		unsigned int* newItems;
		unsigned char* newBitmap;
		equivalenceClassStack[threadNr]->pushEmpty(*itemCount + 1, &newItems, &newBitmap);

		intersectBitsets(
			tidBitmap,
			verticalDataLayout.getTidBitmaps()[item],
			numBytes,
			newBitmap
		);

		unsigned long long newSupportCount = popcount(newBitmap, numBits);
		if (newSupportCount >= MIN_NUM_TRANSACTIONS)
		{
			arrayCopy(items, newItems, *itemCount);
			newItems[*itemCount] = item;

			unsigned int* fitems = itemsetStack[threadNr]->push(
				*itemCount + 1,
				calculateSupport(newSupportCount, verticalDataLayout.getNumBitsPerBitmap())
			);
			arrayCopy(newItems, fitems, *itemCount + 1);

			if (i + 1 < frequentItems->size())
				mine(i + 1);
		}
		equivalenceClassStack[threadNr]->pop();
	}
}
