#include "VerticalDataLayout.h"
#include "BitsetUtility.h"
#include "../FimLogger/Logger.h"
#include "Utility.h"

#include <omp.h>
#include <cctype>
#include <string>
using std::string;
#include <algorithm>

VerticalDataLayout::VerticalDataLayout(DatasetPreprocessor& datasetPreprocessor,
									   MultiBuffer** multiBuffers, LogData &logData)
	: datasetPreprocessor(datasetPreprocessor),
	  multiBuffers(multiBuffers),
	  logData(logData)
{
	string startMessage = "Creating vertical data layout.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	numThreads = datasetPreprocessor.getNumThreads();
	numRows = datasetPreprocessor.getSupportBufferSize();
	numColumns = datasetPreprocessor.getNumTransactionsWithFrequentItems();
	stackMemoryManager = new StackMemoryManager*[numThreads];
	for (int i = 0; i < numThreads; i++)
		stackMemoryManager[i] = new StackMemoryManager();

	setFrequentItems();
	allocateMemory();
	createVerticalDataLayout();

	double elapsedTime = omp_get_wtime() - time;
	string endMessage = "Vertical data layout created. [" + toString(elapsedTime) + "s]\n";
	writeToLog(endMessage);
	double memoryUsageInMB = getMemoryUsageInBytes() / (double)(1024.0 * 1024.0);
	writeToLog("Total memory usage: " + toString(memoryUsageInMB) + "MB\n\n");
	logData.memoryUsageInMB = memoryUsageInMB;
	logData.preprocessTime = elapsedTime;
}

VerticalDataLayout::~VerticalDataLayout()
{
	delete[] tidBitmaps;

	for (int i = 0; i < numThreads; i++)
		delete stackMemoryManager[i];
	delete[] stackMemoryManager;
}

/***** PUBLIC FUNCTIONS *****/

vector<unsigned int>* VerticalDataLayout::getFrequentItems()
{
	return &frequentItems;
}

unsigned long long VerticalDataLayout::getNumBitsPerBitmap()
{
	return numColumns;
}

unsigned char** VerticalDataLayout::getTidBitmaps() const
{
	return tidBitmaps;
}

unsigned long long VerticalDataLayout::getMemoryUsageInBytes() const
{
	unsigned long long totalMemoryUsage = 0;
	for (int i = 0; i < numThreads; i++)
		totalMemoryUsage += stackMemoryManager[i]->getMemoryUsageInBytes();
	return totalMemoryUsage;
}

/***** PRIVATE FUNCTIONS *****/

void VerticalDataLayout::setFrequentItems()
{
	for (unsigned long long i = 0; i < numRows; i++)
	{
		if (DatasetPreprocessor::supportCount[i] >= datasetPreprocessor.getMinNumTransactions())
		{
			frequentItems.push_back(i);
		}
	}

	// Sort in frequency descending order
	// In this way the most frequent items will be mined first
	qsort(&frequentItems[0], frequentItems.size(), sizeof(int), compareItem);
}

void VerticalDataLayout::allocateMemory()
{
	writeToLog("Allocating memory.\n");
	tidBitmaps = new unsigned char*[numRows];
	#pragma omp parallel for num_threads(numThreads)
	for (unsigned long long i = 0; i < numRows; i++)
	{
		int threadNr = omp_get_thread_num();
		if (DatasetPreprocessor::supportCount[i] >= datasetPreprocessor.getMinNumTransactions())
		{
			int numBytes = getNumBytes(numColumns);
			tidBitmaps[i] = stackMemoryManager[threadNr]->grow(numBytes);
			zeroBytes(tidBitmaps[i], numBytes);
		}
		else
			tidBitmaps[i] = NULL;
	}
	writeToLog("Memory allocated.\n");
}

void VerticalDataLayout::createVerticalDataLayout()
{
	#pragma omp parallel num_threads(numThreads)
	{
		int threadNr = omp_get_thread_num();

		bool containsFrequentItem = false;
		int firstTransactionNr = getFirstTransactionNr(threadNr);
		int transactionNr = firstTransactionNr;
		int item = -1;
		for (int i = 0; i < multiBuffers[threadNr]->getNumBytes(); i++)
		{
			char byte = multiBuffers[threadNr]->getByteAndDelete(&i);

			if (isdigit(byte))
			{
				if (item == -1)
					item = 0;
				else
					item *= 10;

				item += byte - '0';
			}
			else
			{
				if (item != -1) // This is the case if the last byte was a space
				{				// (in case multiple spaces occur after each other).
					if (tidBitmaps[item] != NULL)
					{
						containsFrequentItem = true;
						int transactionByteNr = transactionNr / 8;
						int firstTransactionByteNr = firstTransactionNr / 8;
						int lastTransactionByteNr = getLastTransactionNr(threadNr) / 8;
						bool isSharedByte = transactionByteNr == firstTransactionByteNr;
						isSharedByte = isSharedByte || transactionByteNr == lastTransactionByteNr;
						if (isSharedByte)
						{
							#pragma omp critical
							{
								// This byte might be shared between two threads.
								setBitAt(tidBitmaps[item], transactionNr, true);
							}
						}
						else
							setBitAt(tidBitmaps[item], transactionNr, true);
					}
				}
				item = -1;

				if (byte == '\n' && containsFrequentItem)
				{
					transactionNr++;
					containsFrequentItem = false;
				}
			}
		}
	}
	for (int i = 0; i < datasetPreprocessor.getNumThreads(); i++)
		delete multiBuffers[i];
	delete[] multiBuffers;
}

int VerticalDataLayout::getFirstTransactionNr(int threadNr)
{
	if (threadNr <= 0)
		return 0;

	int sum = 0;
	for (int i = 0; i < threadNr; i++)
		sum += datasetPreprocessor.getNumTransactionsForThread(i);

	return sum;
}

int VerticalDataLayout::getLastTransactionNr(int threadNr)
{
	if (threadNr < 0)
		return 0;

	int sum = 0;
	for (int i = 0; i <= threadNr; i++)
		sum += datasetPreprocessor.getNumTransactionsForThread(i);
	sum = sum == 0 ? 0 : sum - 1;

	return sum;
}
