#include "FPTree.h"
#include "../FimLogger/Logger.h"
#include "PreProcessor.h"

#include <omp.h>
#include <iostream>
using std::cout;
using std::endl;
#include <cstdlib>
using std::qsort;
#include <cmath>
using std::ceil;

FPTree::FPTree(MultiBuffer* multiBuffer,
			   double minSup,
			   int numTransactions,
			   int longestTransaction)
	: multiBuffer(multiBuffer),
	  minSup(minSup),
	  numTransactions(numTransactions),
	  longestTransaction(longestTransaction)
{
	memoryManagerFromOutside = false;
	hashHeaderTable = NULL;
	memoryManager = new StackMemoryManager();
	setRootNode(allocateNode());
	initHeaderTable();
	buildFPTree();

	//printFPTree(rootNode);
}

/*
This is used whenever the tree is constructured outside this class.
For example conditional FP Trees.
*/
FPTree::FPTree()
{
	memoryManagerFromOutside = false;
	hashHeaderTable = NULL;
	memoryManager = new StackMemoryManager();
	setRootNode(allocateNode());
	headerTable.itemCounts = NULL;
	headerTable.nodes = NULL;
}

FPTree::FPTree(StackMemoryManager* stackMemoryManager)
{
	memoryManagerFromOutside = true;
	hashHeaderTable = NULL;
	memoryManager = stackMemoryManager;
	setRootNode(allocateNode());
	headerTable.itemCounts = NULL;
	headerTable.nodes = NULL;
}

FPTree::~FPTree()
{
	if (hashHeaderTable != NULL)
	{
		// No need, created in own memory manager.
		//delete hashHeaderTable;
	}
	else
	{
		if (headerTable.itemCounts != NULL)
			delete[] headerTable.itemCounts;
		if (headerTable.nodes != NULL)
			delete[] headerTable.nodes;
	}
	if (!memoryManagerFromOutside)
		delete memoryManager;
}

/***** PUBLIC FUNCTIONS *****/

void FPTree::setHashHeaderTable(HashTable<HashHeaderTableElement*>* table)
{
	hashHeaderTable = table;
}

HashTable<HashHeaderTableElement*>* FPTree::getHashHeaderTable()
{
	return hashHeaderTable;
}

FPNode* FPTree::getRootNode()
{
	return rootNode;
}

HeaderTable* FPTree::getHeaderTable()
{
	return &headerTable;
}

unsigned long long FPTree::getMemoryUsage()
{
	return memoryManager->getMemoryUsage();
}

/*
Can only be used for conditional pattern trees (i.e. when hashHeaderTable is not NULL).
*/
bool FPTree::containsSinglePathOnly(int item)
{
	FPNode* node = hashHeaderTable->find(item)->value->node;
	if (node != NULL && node->nextEqualItem != NULL)
		return false;
	return true;
}

void FPTree::insertTransaction(int* transaction, int itemCount, int supportCount)
{
	FPNode* currentNode = rootNode;
	for (int i = 0; i < itemCount; i++)
	{
		// see if node exists in children
		FPNode* child = currentNode->firstChild;
		while (child != NULL)
		{
			if (child->item == transaction[i])
			{
				currentNode = child;
				break;
			}
			else
				child = child->nextSibling;
		}

		if (child == NULL) // Item does not exist in the FPTree, hence add it and all its children.
		{
			for (int j = i; j < itemCount; j++)
			{
				FPNode* lastFirstChild = currentNode->firstChild;
				FPNode* newNode = allocateNode();
				newNode->item = transaction[j];
				if (supportCount != -1)
					newNode->itemCount = supportCount;
				else
					newNode->itemCount = 1;
				newNode->parent = currentNode;
				newNode->nextSibling = lastFirstChild;

				// Update header table
				if (hashHeaderTable == NULL)
				{
					FPNode* lastFirstNodeInHeader = headerTable.nodes[transaction[j]];
					headerTable.nodes[transaction[j]] = newNode;
					newNode->nextEqualItem = lastFirstNodeInHeader;
				}
				else
				{
					HashHeaderTableElement* element = hashHeaderTable->find(transaction[j])->value;
					FPNode* lastFirstNodeInHeader = element->node;
					element->node = newNode;
					newNode->nextEqualItem = lastFirstNodeInHeader;
					element->itemCount += supportCount;
				}

				currentNode->firstChild = newNode;
				currentNode = newNode;
			}

			return; // All of the items are added, hence return.
		}
		else
		{
			if (supportCount != -1)
				child->itemCount += supportCount;
			else
				child->itemCount++;
		}
	}
}

/***** PRIVATE FUNCTIONS *****/

void FPTree::setRootNode(FPNode* root)
{
	rootNode = root;
	root->item = -1;
}

void FPTree::initHeaderTable()
{
	int threadNr = omp_get_thread_num();
	headerTable.numItems = PreProcessor::getCountListBufferSize();
	headerTable.itemCounts = new int[headerTable.numItems];
	for (int i = 0; i < headerTable.numItems; i++)
		headerTable.itemCounts[PreProcessor::getInitialToRemapIdArray()[i]] = PreProcessor::getLocalCountList(threadNr)[i];

	headerTable.nodes = new FPNode*[headerTable.numItems];
	for (int i = 0; i < headerTable.numItems; i++)
		headerTable.nodes[i] = NULL;
}

void FPTree::buildFPTree()
{
	// An item must be in at least minNumTransactions transactions in order to be added to the tree.
	int minNumTransactions = (int)ceil(minSup * numTransactions);
	int* items = new int[longestTransaction];
	int itemIndex = 0;
	int item = -1;
	for (int i = 0; i < multiBuffer->getNumBytes(); i++)
	{
		char byte = multiBuffer->getByteAndDelete(&i);

		if (isdigit(byte))
		{
			if (item == -1)
				item = 0;
			else
				item *= 10;

			item += byte - '0';
		}
		else
		{
			if (item != -1) // This is the case if the last byte was a space (in case multiple spaces occur after each other).
				items[itemIndex++] = item;
			item = -1;

			if (byte == '\n')
			{
				qsort(items, itemIndex, sizeof(int),  &PreProcessor::compareItem);

				// Remove infrequent items and remap items.
				for (int j = 0; j < itemIndex; j++)
				{
					if (PreProcessor::getGlobalCountList()[items[j]] < minNumTransactions)
					{
						itemIndex = j;
						break;
					}
					items[j] = PreProcessor::getInitialToRemapIdArray()[items[j]];
				}

				if (itemIndex != 0)
				{
					insertTransaction(items, itemIndex);
					itemIndex = 0;
				}
			}
		}
	}
	delete[] items;
}

FPNode* FPTree::allocateNode()
{
	FPNode* node = (FPNode*)memoryManager->grow(sizeof(FPNode));
	node->itemCount = 0;
	node->parent = NULL;
	node->firstChild = NULL;
	node->nextSibling = NULL;
	node->nextEqualItem = NULL;
	return node;
}

void FPTree::printFPTree(FPNode* node)
{
	if (node == NULL)
		return;

	cout << "id:" << node->item << " c:" << node->itemCount << " " << node << endl;

	printFPTree(node->firstChild);

	FPNode* sibling = node->nextSibling;
	while (sibling != NULL)
	{
		cout << endl;
		cout << "parent " << node->parent->item << endl;
		printFPTree(sibling);
		sibling = sibling->nextSibling;
	}
}
