#ifndef HASH_TABLE_H_
#define HASH_TABLE_H_

#include "StackMemoryManager.h"

#include <cstdlib>

const int BUCKET_SIZE = 3;

template <typename T>
struct HashTableElement
{
	int id;
	T value;
};

template <typename T>
struct Bucket
{
	HashTableElement<T>** values;
	Bucket* nextBucket;
};

template <typename T>
struct HashTableIterator
{
	HashTableElement<T>* value;
	Bucket<T>* currentBucket;
	int bucketNr;
	int bucketIndex;
};

template <typename T>
class HashTable
{
	private:
		StackMemoryManager* memoryManager;
		Bucket<T>** buckets;
		int initialBucketCount;
		int itemCount;

		int hashFunction(int id)
		{
			// Knuth's Multiplicative Hashing
			return id * 2654435761 % initialBucketCount;
		}

		void createBucket(Bucket<T>* forBucket)
		{
			forBucket->values = (HashTableElement<T>**)allocateMemory(sizeof(HashTableElement<T>*) * BUCKET_SIZE);
			for (int j = 0; j < BUCKET_SIZE; j++)
				forBucket->values[j] = NULL;
			forBucket->nextBucket = NULL;
		}

		unsigned char* allocateMemory(long long byteCount)
		{
			return memoryManager->grow(byteCount);
		}
	public:
		HashTable(StackMemoryManager* memoryManager, int initialBucketCount)
		{
			constructor(memoryManager, initialBucketCount);
		}

		void constructor(StackMemoryManager* memoryManager, int initialBucketCount)
		{
			itemCount = 0;
			this->memoryManager = memoryManager;
			this->initialBucketCount = initialBucketCount;
			buckets = (Bucket<T>**)allocateMemory(sizeof(Bucket<T>*) * initialBucketCount);
			for (int i = 0; i < initialBucketCount; i++)
			{
				buckets[i] = (Bucket<T>*)allocateMemory(sizeof(Bucket<T>));
				createBucket(buckets[i]);
			}
		}
		
		~HashTable()
		{
			// WARNING: This method does not delete the memory that was allocated.
		}

		HashTableElement<T>* find(int id)
		{
			Bucket<T>* bucket = buckets[hashFunction(id)];

			while (bucket != NULL)
			{
				for (int i = 0; i < BUCKET_SIZE; i++)
				{
					if (bucket->values[i] != NULL && bucket->values[i]->id == id)
						return bucket->values[i];
				}
				bucket = bucket->nextBucket;
			}

			return NULL;
		}

		void insert(int id, T value)
		{
			Bucket<T>* bucket = buckets[hashFunction(id)];
			itemCount++;

			while (true)
			{
				for (int i = 0; i < BUCKET_SIZE; i++)
				{
					if (bucket->values[i] == NULL)
					{
						bucket->values[i] = (HashTableElement<T>*)allocateMemory(sizeof(HashTableElement<T>));
						bucket->values[i]->id = id;
						bucket->values[i]->value = value;
						return;
					}
				}

				if (bucket->nextBucket == NULL)
				{
					bucket->nextBucket = (Bucket<T>*)allocateMemory(sizeof(Bucket<T>));
					createBucket(bucket->nextBucket);
					bucket->nextBucket->values[0] = (HashTableElement<T>*)allocateMemory(sizeof(HashTableElement<T>));
					bucket->nextBucket->values[0]->id = id;
					bucket->nextBucket->values[0]->value = value;
					return;
				}

				bucket = bucket->nextBucket;
			}
		}

		void begin(HashTableIterator<T>* it)
		{
			if (itemCount == 0 || initialBucketCount == 0)
			{
				it->value = NULL;
				return;
			}

			it->bucketIndex = 0;
			it->bucketNr = 0;
			it->currentBucket = buckets[0];
			it->value = it->currentBucket->values[it->bucketIndex];

			// Find first element
			while (it->value == NULL)
				next(it);
		}

		void next(HashTableIterator<T>* it)
		{
			it->bucketIndex++;
			if (it->bucketIndex >= BUCKET_SIZE || it->currentBucket->values[it->bucketIndex] == NULL)
			{
				if (it->bucketIndex >= BUCKET_SIZE)
				{
					it->bucketIndex = 0;
					it->currentBucket = it->currentBucket->nextBucket;

					if (it->currentBucket != NULL)
					{
						it->value = it->currentBucket->values[it->bucketIndex];
						return;
					}
					else
					{
						it->bucketNr++;
						if (it->bucketNr >= initialBucketCount)
						{
							it->value = NULL; // No more elements
							return;
						}
						else
							it->currentBucket = buckets[it->bucketNr];
					}
				}

				while (it->currentBucket->values[it->bucketIndex] == NULL)
				{
					it->bucketIndex = 0;
					it->bucketNr++;
					if (it->bucketNr >= initialBucketCount)
					{
						it->value = NULL; // No more elements
						return;
					}
					else
						it->currentBucket = buckets[it->bucketNr];
				}
			}

			it->value = it->currentBucket->values[it->bucketIndex];
		}

		unsigned int size()
		{
			return itemCount;
		}
};

#endif // HASH_TABLE_H_
