#include "PreProcessor.h"
#include "../FimLogger/Logger.h"

#include <iostream>
#include <omp.h>
#include <cmath>
using std::ceil;
#include <algorithm>
using std::qsort;

int PreProcessor::countListBufferSize;
int** PreProcessor::localCountLists;
int* PreProcessor::globalCountList;
int* PreProcessor::initialToRemapId;
int* PreProcessor::remapToInitialId;
int PreProcessor::maxTransactionLength;

PreProcessor::PreProcessor(MultiBuffer** multiBuffers, int numThreads)
	: multiBuffers(multiBuffers), NUM_THREADS(numThreads)
{
	getDatasetInformation(&countListBufferSize, &maxTransactionLength, &numberOfTransactions);

	localCountLists = new int*[NUM_THREADS];
	for (int i = 0; i < NUM_THREADS; i++)
		localCountLists[i] = new int[countListBufferSize];
	globalCountList = new int[countListBufferSize];

	computeLocalCountLists();
	computeGlobalCountList();
	remapItemsByFrequency();
}

PreProcessor::~PreProcessor()
{
	for (int i = 0; i < NUM_THREADS; i++)
		delete[] localCountLists[i];
	delete[] localCountLists;
	delete[] globalCountList;
	delete[] initialToRemapId;
	delete[] remapToInitialId;
}

/***** PUBLIC FUNCTIONS *****/

int PreProcessor::compareItem(const void* itemIdA, const void* itemIdB)
{
	int value = globalCountList[*(int*)itemIdB] - globalCountList[*(int*)itemIdA];
	if (value != 0)
		return value;
	else // sort by ID
		return *(int*)itemIdB - *(int*)itemIdA;
}

int* PreProcessor::getGlobalCountList()
{
	return globalCountList;
}

int* PreProcessor::getLocalCountList(int threadNr)
{
	return localCountLists[threadNr];
}

int PreProcessor::getCountListBufferSize()
{
	return countListBufferSize;
}

int* PreProcessor::getRemapToInitialIdArray()
{
	return remapToInitialId;
}

int* PreProcessor::getInitialToRemapIdArray()
{
	return initialToRemapId;
}

int PreProcessor::getNumberOfTransactions()
{
	return numberOfTransactions;
}

int PreProcessor::getMaxTransactionLength()
{
	return maxTransactionLength;
}

/***** PRIVATE FUNCTIONS *****/

/*
Finds and returns the item with highest ID + 1 (+1 in order to support ID=0), and returns the length of the longest transaction.
This also calucates the transaction count.
*/
void PreProcessor::getDatasetInformation(int* maxId, int* maxTransactionLength, int* transactionCount) const
{
	string startMessage = "Calculating support count buffer size, the length of the longest transaction and transaction count.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	int* localMaxId = new int[NUM_THREADS];
	int* localMaxTransactionLength = new int[NUM_THREADS];
	int* localTransactionCount = new int[NUM_THREADS];

	#pragma omp parallel num_threads(NUM_THREADS)
	{
		int threadNr = omp_get_thread_num();
		int item = -1;
		int transactionLength = 0;
		localMaxId[threadNr] = -1;
		localMaxTransactionLength[threadNr] = -1;
		localTransactionCount[threadNr] = 0;
		for (int i = 0; i < multiBuffers[threadNr]->getNumBytes(); i++)
		{
			char byte = multiBuffers[threadNr]->getByte(&i);

			if (isdigit(byte))
			{
				if (item == -1)
					item = 0;
				else
					item *= 10;

				item += byte - '0';
			}
			else
			{
				if (item != -1) // This is the case if the last byte was a space (in case multiple spaces occur after each other).
				{
					if (item > localMaxId[threadNr])
						localMaxId[threadNr] = item;
					transactionLength++;
				}
				item = -1;

				if (byte == '\n')
				{
					if (transactionLength > localMaxTransactionLength[threadNr])
						localMaxTransactionLength[threadNr] = transactionLength;
					transactionLength = 0;
					localTransactionCount[threadNr]++;
				}
			}
		}
	}

	*maxId = -1;
	*maxTransactionLength = -1;
	*transactionCount = 0;
	for (int i = 0; i < NUM_THREADS; i++)
	{
		if (localMaxId[i] > *maxId)
			*maxId = localMaxId[i];

		if (localMaxTransactionLength[i] > *maxTransactionLength)
			*maxTransactionLength = localMaxTransactionLength[i];

		*transactionCount += localTransactionCount[i];
	}
	(*maxId)++; // In order to support values with ID = 0

	delete[] localMaxId;
	delete[] localMaxTransactionLength;
	delete[] localTransactionCount;

	string endMessage = "Buffer size: " + toString(*maxId) + " * (NUM_THREADS + 2) * sizeof(int) B = " + toString((int)ceil((*maxId * (NUM_THREADS + 2) * sizeof(int)) / (double)1024)) +
		" kB. Longest transaction: " + toString(*maxTransactionLength) + ". Number of transactions: " + toString(*transactionCount) +
		 ". [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);
}

void PreProcessor::computeLocalCountLists()
{
	string startMessage = "Computing local count lists.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	#pragma omp parallel num_threads(NUM_THREADS)
	{
		int threadNr = omp_get_thread_num();

		for (int i = 0; i < countListBufferSize; i++)
			localCountLists[threadNr][i] = 0; // Initialize support counts to 0.
	
		int item = -1;
		for (int i = 0; i < multiBuffers[threadNr]->getNumBytes(); i++)
		{
			char byte = multiBuffers[threadNr]->getByte(&i);

			if (isdigit(byte))
			{
				if (item == -1)
					item = 0;
				else
					item *= 10;

				item += byte - '0';
			}
			else
			{
				if (item != -1) // This is the case if the last byte was a space (in case multiple spaces occur after each other).
					localCountLists[threadNr][item]++;
				item = -1;
			}
		}
	}

	string endMessage = "Local count lists computed. [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);
}

void PreProcessor::computeGlobalCountList()
{
	string startMessage = "Computing global count list.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	int partitionSize = countListBufferSize / NUM_THREADS;

	#pragma omp parallel num_threads(NUM_THREADS)
	{
		int threadNr = omp_get_thread_num();

		int start = threadNr * partitionSize;
		int end = threadNr == NUM_THREADS - 1 ? countListBufferSize : (threadNr + 1) * partitionSize;

		for (int i = start; i < end; i++)
			globalCountList[i] = 0; // Initialize support counts to 0.

		for (int i = start; i < end; i++)
		{
			for (int j = 0; j < NUM_THREADS; j++)
				globalCountList[i] += localCountLists[j][i];
		}
	}

	string endMessage = "Global count list computed. [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);
}

/*
Identifiers are assign in increasing order of item frequency.
They are stored in an item mapping array.
*/
void PreProcessor::remapItemsByFrequency()
{
	string startMessage = "Remapping items by frequency.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	remapToInitialId = new int[countListBufferSize];
	for (int i = 0; i < countListBufferSize; i++)
		remapToInitialId[i] = i;
	qsort(remapToInitialId, countListBufferSize, sizeof(int), &PreProcessor::compareItem);

	initialToRemapId = new int[countListBufferSize];
	for (int i = 0; i < countListBufferSize; i++)
		initialToRemapId[remapToInitialId[i]] = i;

	string endMessage = "Remapping finished. [" + toString(omp_get_wtime() - time) + "s]\n\n";
	writeToLog(endMessage);
}
