#include "ShaFem.h"
#include "FileLoader.h"
#include "PreProcessor.h"
#include "FPTree.h"
#include "../FimLogger/Logger.h"
#include "XfpTree.h"
#include "XfpTreeMiner.h"

#include <omp.h>
#include <ctime>
#include <iostream>
using namespace std;

ShaFem::ShaFem(InputData& inputData, LogData& logData)
	: inputData(inputData), logData(logData)
{
	multiBuffers = new MultiBuffer*[inputData.numThreads];
}

ShaFem::~ShaFem()
{
	delete[] multiBuffers;
}

/***** PUBLIC FUNCTIONS *****/

void ShaFem::startAlgorithm()
{
	FileLoader fileLoader;
	fileLoader.loadFileToBuffers(inputData.datasetPath, &multiBuffers, inputData.numThreads);
	PreProcessor preProcessor(multiBuffers, inputData.numThreads);

	int numberOfTransactions = preProcessor.getNumberOfTransactions();
	int maxTransactionLength = preProcessor.getMaxTransactionLength();

	string startMessage = "Creating local FP-trees.\n";
	writeToLog(startMessage);
	double time = omp_get_wtime();

	FPTree** fpTrees = new FPTree*[inputData.numThreads];
	#pragma omp parallel num_threads(inputData.numThreads)
	{
		int threadNr = omp_get_thread_num();
		fpTrees[threadNr] = new FPTree(multiBuffers[threadNr], inputData.minSup, numberOfTransactions, maxTransactionLength);
		delete multiBuffers[threadNr];
	}		

	double totalMemoryConsumption = 0.0;
	double fpTreesTime = omp_get_wtime() - time;
	for (int i = 0; i < inputData.numThreads; i++)
		totalMemoryConsumption += fpTrees[i]->getMemoryUsage() / (double)(1024 * 1024);;
	string endMessage = "Local FP-trees created (" + toString(totalMemoryConsumption) +
		" MB). [" + toString(fpTreesTime) + "s]\n\n";
	writeToLog(endMessage);
	logData.memoryUsageInMB = totalMemoryConsumption;
	logData.preprocessTime += fpTreesTime;

	string startMessageXfpTree = "Creating XFP-tree.\n";
	writeToLog(startMessageXfpTree);
	double timeXfpTree = omp_get_wtime();

	XfpTree xfpTree(fpTrees, inputData.numThreads);

	double xfpTreeTotalTime = omp_get_wtime() - timeXfpTree;
	logData.preprocessTime += xfpTreeTotalTime;
	string endMessageXfpTree = "XFP-tree created. [" + toString(xfpTreeTotalTime) + "s]\n\n";
	writeToLog(endMessageXfpTree);

	XfpTreeMiner xfpTreeMiner(&xfpTree, inputData.numThreads, inputData.minSup, numberOfTransactions, logData);
	xfpTreeMiner.startMiningProcedure();

	if (inputData.writeOutput)
		xfpTreeMiner.writeResults(algorithmName + "_output.txt");
	else
		xfpTreeMiner.reportNumberOfFrequentItemsetsFound();

	// delete fp-trees
	for (int i = 0; i < inputData.numThreads; i++)
		delete fpTrees[i];
	delete[] fpTrees;
}
